from collections import OrderedDict
from contextlib import contextmanager
from infinishell.infishell import commandgroup, render
from infinishell.infinibox.commands import EntityCommands, set_renderer_query_fields
from .interface import MTU, Mode
from infinishell.infinibox.commands.infiniguard.lib.dde import get_dde_url
from .nettools import NetworkCommands, DdeNetworkDevice, DDEBondMember, DDEBondPort
from infinishell.infinibox.commands.infiniguard import ig_command, DDEApp
from .nettools import RebootChoice
from ..options import QueryOptions


ETHERNET_PORT_FIELDS = OrderedDict(
    device_name='Port',
    patch_panel='Patch panel',
    members='Members',
    mode='Bonding mode',
    maximum_speed='Link speed',
    connection='Link status',
    dde_node='DDE Node',
    mtu='MTU',
)


@contextmanager
def patch_dde_entity(cls, dde: DDEApp):
    class _DdeSpecificEntity(cls.ENTITY_TYPE):
        URI = get_dde_url(dde.dde_instance_id,
                          DdeNetworkDevice.URI)
    original = cls.ENTITY_TYPE.DDE_INSTANCE_ID
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = dde.dde_instance_id
    yield
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = original


def _filter_bond_params(params, context):
    interface_type = context(param_name='interface_type', raw=False)

    if interface_type != 'bond':
        return [arg for arg in params if arg.name not in ('mode', 'members')]
    else:
        return params


def _convert_port_fields(ports, dde):
    updated_ports = []
    for port in ports:
        if port['type'].lower() in ('slave', 'port'):
            port['members'] = None
            port['mode'] = None
        elif port['type'].lower() == 'bond':
            port['patch_panel'] = None
        port['dde_node'] = dde.node_id
        updated_ports.append(port)
    return updated_ports


@commandgroup(prefix="port", parent=NetworkCommands)
class NetworkInterfaceCommands(EntityCommands):
    ENTITY_TYPE = DdeNetworkDevice

    @staticmethod
    def _filter_by_dde(dde, devices):
        return next((device for device in devices if device.dde_instance_id == dde.dde_instance_id), None)

    @ig_command(
        description="Display ethernet ports",
        types=(DDEApp,),
        options=QueryOptions(fields=ETHERNET_PORT_FIELDS,
                             substitute=ETHERNET_PORT_FIELDS, detailed=True)
    )
    @render(fields=ETHERNET_PORT_FIELDS,
            substitute=ETHERNET_PORT_FIELDS,
            detailed_fields=ETHERNET_PORT_FIELDS,
            detailed_individually=True)
    def query(self, _api, dde):
        set_renderer_query_fields()
        return _convert_port_fields(DdeNetworkDevice.get_dde(_api, dde), dde)

    @ig_command(
        description="Delete bond",
        types=(DDEApp, DDEBondPort, RebootChoice)
    )
    def delete(self, _api, dde, bond, reboot):
        with patch_dde_entity(self.__class__, dde):
            bond = self._filter_by_dde(dde, bond)
            url = f'{self.ENTITY_TYPE.get_uri()}/{bond.device_name}'
            _api.delete(url, data=dict(reboot=reboot))

    @ig_command(
        description="Modify MTU, Bond Mode, Bond Members",
        types=(DDEApp, DdeNetworkDevice, RebootChoice, MTU, DDEBondMember.List, Mode)
    )
    def modify(self, _api, dde, port, reboot, mtu=None, members=None, mode=None):
        port = self._filter_by_dde(dde, port)
        with patch_dde_entity(self.__class__, dde):
            url = f'{self.ENTITY_TYPE.get_uri()}/{port.device_name}'
            data = dict(mtu=mtu, mode=mode, members=members and [member.name for member in members],
                        reboot=reboot)
            data = {k: v for k, v in data.items() if v is not None}
            _api.put(url, data=data)
