from contextlib import contextmanager
from functools import partial
from infinishell.infinibox.commands import EntityCommands
from infinishell.infinibox.commands.infiniguard import ig_command, DDEApp
from infinishell.infinibox.commands.options import QueryOptions
from infinishell.infishell import commandgroup
from infinishell.infishell.types import EnumType, Netmask, IPAddress, MappedEnumType, Range
from infinishell.infishell.utils import strlist
from .lib.context_entity import filtered
from .nettools import NetworkCommands, DdeNetworkDevice, DDEBondMember, DdeNetworkInterface, RebootChoice
from .. import render

QUERY_FIELDS = strlist("devname intfname ext_dde_ip ip_address netmask gateway nat traffic_type")
SUBSTITUTE_FIELDS = dict(
    devname='Device Name',
    intfname='Interface',
    ext_dde_ip='DDE External IP',
    ip_address='Interface IP',
    nat='NAT IP',
    ddeapp='DDE App'
)
DETAILED_QUERY_FIELDS = strlist("ddeapp devname intfname ext_dde_ip ip_address netmask gateway nat traffic_type")


APP_ID_TO_NAME = {
    1: 'a',
    2: 'b',
    3: 'standby'
}


class InterfaceType(EnumType):
    def _get_values(self):
        return 'port', 'bond'


class Mode(MappedEnumType):
    def _get_mapping(self):
        return {
            'Round Robin': 'RR',
            'Active Backup': 'AB',
            'LACP': 'LACP'
        }


class YesChoice(MappedEnumType):
    def _get_mapping(self):
        return {'YES': True}


class MTU(Range):
    MIN = 576
    MAX = 9000


class VLAN(Range):
    MIN = 2
    MAX = 4094


class TrafficType(EnumType):
    def _get_values(self):
        return 'any', 'data', 'replication'


@contextmanager
def patch_dde_entity(cls, dde: DDEApp):
    original = cls.ENTITY_TYPE.DDE_INSTANCE_ID
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = dde.dde_instance_id
    yield
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = original


def _filter_bond_params(params, context):
    interface_type = context(param_name='interface_type', raw=False)

    if interface_type != 'bond':
        return [arg for arg in params if arg.name not in ('mode', 'members')]
    else:
        return params


@commandgroup(prefix="interface", parent=NetworkCommands)
class NetworkInterfaceCommands(EntityCommands):
    ENTITY_TYPE = DdeNetworkInterface

    @staticmethod
    def _filter_by_dde(dde, devices):
        """
        Even though the autocomplete is performed after filtering on the dde instance,
        the selected devname from all ddes are passed to the command as a list,
         and must be filtered again for the devname from the correct DDE.
        """
        return devices and next((device for device in devices if device.dde_instance_id == dde.dde_instance_id), None)

    @ig_command(
        description="Create an interface on a port or bond device",
        types=(DDEApp, InterfaceType, DdeNetworkDevice, IPAddress, Netmask, IPAddress, RebootChoice, DDEBondMember.List,
               Mode, VLAN, TrafficType.List, IPAddress, YesChoice, MTU)
    )
    @filtered(
        filter_params=_filter_bond_params
    )
    def create(self, _api, dde, interface_type, devname, ip_address, netmask, gateway, reboot, members=None, mode=None,
               vlan=None, traffic_type=None, nat=None, ext_dde_ip=None, mtu=None):
        devname = self._filter_by_dde(dde, devname)
        with patch_dde_entity(self.__class__, dde):
            return self._create(_api,
                                devname=devname.name,
                                interface_type=interface_type,
                                ip_address=ip_address.ip,
                                netmask=netmask.netmask,
                                gateway=gateway.ip,
                                members=members and [member.name for member in members],
                                reboot=reboot,
                                mode=mode,
                                vlan=vlan,
                                traffic_type=traffic_type,
                                nat=nat and nat.ip,
                                ext_dde_ip=ext_dde_ip,
                                mtu=mtu,
                                )

    @ig_command(
        description="Delete an interface on a port or bond device",
        types=(DDEApp, DdeNetworkInterface, RebootChoice)
    )
    def delete(self, _api, dde, interface, reboot):
        interface = self._filter_by_dde(dde, interface)
        with patch_dde_entity(self.__class__, dde):
            self._delete(_api, interface, data=dict(reboot=reboot))

    @ig_command(
        description="Modify NAT of an interface on a port or bond device",
        types=(DDEApp, DdeNetworkInterface, IPAddress, RebootChoice)
    )
    def modify(self, _api, dde, interface, nat, reboot):
        interface = self._filter_by_dde(dde, interface)
        with patch_dde_entity(self.__class__, dde):
            self._update(_api,
                         interface,
                         reboot=reboot,
                         nat=nat and nat.ip)

    @ig_command(
        description="Set an interface to be the external IP of the DDE",
        types=(DDEApp, DdeNetworkInterface, RebootChoice)
    )
    def set_ext_dde_ip(self, _api, dde, interface, reboot):
        interface = self._filter_by_dde(dde, interface)
        with patch_dde_entity(self.__class__, dde):
            self._update(_api,
                         interface,
                         reboot=reboot,
                         ext_dde_ip=True)

    def _query_hook(self, intfs):
        for intf in intfs:
            intf.ddeapp = APP_ID_TO_NAME[intf.dde_instance_id]
            if 'nat' not in intf:
                intf.nat = ''
            yield intf

    @ig_command(
        description="List interfaces",
        types=(DDEApp, DdeNetworkInterface, DdeNetworkDevice),
        options=QueryOptions(fields=QUERY_FIELDS, substitute=SUBSTITUTE_FIELDS, detailed=True)
    )
    @render(fields=QUERY_FIELDS,
            substitute=SUBSTITUTE_FIELDS,
            detailed_fields=DETAILED_QUERY_FIELDS,
            detailed_individually=True)
    def query(self, _api, dde, interface=None, devname=None):
        interface = self._filter_by_dde(dde, interface)
        device = self._filter_by_dde(dde, devname)
        with patch_dde_entity(self.__class__, dde):
            return self._query(_api, _hook=partial(self._query_hook),
                               intfname=interface and interface.intfname,
                               devname=device and device.name,
                               dde_instance_id=dde.dde_instance_id)
