from . import DDEApp, DDEActiveApp, ig_command
from .nodes import DdeAppCommands
from ..types import ComponentNestedEntity
from infinishell.infinibox.exceptions import (ManagementException,
                                              UnexpectedAPIResponse,
                                              NoSuchEntity)
from infinishell.infishell import commandgroup
from ..options import QueryOptions
from .lib.renderer import render_multi_fields
from infinishell.infinibox.renderer.values import ValueTypeRenderer, StorageCapacityRenderer
from .. import render, set_renderer_query_fields
from collections import OrderedDict
from infinishell.infinibox.commands.infiniguard.lib.dde import allddeget, ddeget
import re
from munch import Munch


class ReplicationShare(ComponentNestedEntity):
    CONTAINER_TYPE = DDEApp
    CONTAINER_SUBCOMPONENTS_FIELD = "stats/replication"
    SUBCOMPONENT_CONTAINER_ID_FIELD = "ddeapp"
    NAME_PATTERN = re.compile(r"^(\w[\w-]*)$", re.I)
    ENTITY_NAME_FORMATTER = "{}"
    ENTITY_NAME_PART_FIELDS = ("_name",)
    HAS_NAME = True
    URI = "../dde"

    @classmethod
    def get_all(cls, api, *args, **kwargs):
        pluck = api.create_pluck(cls.CONTAINER_TYPE.ID_FIELD, cls.CONTAINER_TYPE.NAME_FIELD)
        # Components do not support filtering, so filter on client side
        container_id = kwargs.get("container_id")
        all_nodes = {dde[cls.CONTAINER_TYPE.ID_FIELD]: dde for dde in cls.CONTAINER_TYPE.iter_all(api, pluck=pluck)}
        if container_id is not None:
            try:
                all_shares = {container_id: ddeget(api, container_id, cls.CONTAINER_SUBCOMPONENTS_FIELD)}
            except Exception:
                all_shares = {container_id: None}
        else:
            all_shares = allddeget(api, cls.CONTAINER_SUBCOMPONENTS_FIELD)
        all_shares = {ddeid: shares for ddeid, shares in all_shares.items() if shares is not None}
        entities = [cls._entity_hook(cls._attach_container(entity, all_nodes[ddeid]))
                    for ddeid, ents in all_shares.items()
                    for entity in ents if entity.get('replication_type') == 'FILEDIR' and cls._entity_predicate(entity)]
        if kwargs.get("hook"):
            entities = kwargs["hook"](entities)
        return entities

    @classmethod
    def _attach_container(cls, entity, container):
        entity[cls.CONTAINER_TYPE.NAME_FIELD] = container[cls.CONTAINER_TYPE.NAME_FIELD]
        return entity

    @classmethod
    def _get_entity_by_ids(cls, api, entity_id):
        entities = [e for e in cls.get_all(api) if e.get('name') == entity_id]
        if len(entities) == 0:
            raise ManagementException(404, None)
        return entities

    @classmethod
    def get_by_name(cls, api, name):
        match = cls.NAME_PATTERN.match(name)
        if not match:
            raise NoSuchEntity('"{}" is not a valid {} name'.format(name, cls.get_entity_name()))
        ids = match.groups()
        try:
            return cls._get_entity_by_ids(api, *ids)
        except ManagementException as e:
            if e.http_code != 404:
                raise
            raise NoSuchEntity('No such {}: "{}"'.format(cls.get_entity_name(), name))


class ReplicationRateRenderer(ValueTypeRenderer):
    _rates = ((1000**4, "TB", "TB/sec"),
              (1000**3, "GB", "GB/sec"),
              (1000**2, "MB", "MB/sec"),
              (1000,    "KB", "KB/sec"),
              (1,       "B",   "B/sec"))

    def render(self, value, empty_value, **properties):
        if value == 0 or value is None:
            return "-"
        _, wanted_unit = properties.get("capacity_unit", (None, None))
        for rate, unit, txt in self._rates:
            if (wanted_unit and unit == wanted_unit) or (not wanted_unit and value >= rate):
                return "{:.2f} {}".format(round(value / float(rate)), txt)


REPORT_URI = "stats/replication"
REPORT_FIELDS = OrderedDict(
    dde_id='dde_app',
    name='name',
    replication_state='state',
    replicated_time='last_replicated',
    replication_start='start',
    replication_completion='end',
    average_transfer_rate='avg_thru',
    estimated_duration='time_to_comp',
    estimated_completion_time='est_comp_time',
    remaining='remaining',
    original_data_size='size',
)
SHORT_REPORT_FIELDS = [field for field in REPORT_FIELDS.keys() if field != 'original_data_size']

INFO_SUBSTITUTES = OrderedDict(
    targetName='Target_dedup_engines',
    originalDataSize='Original_data_size',
    actualDataSent='Actual_data_sent',
    averageSendRate='Average_send_rate',
    sourceName='Source_dedup_engines',
    actualDataReceived='Actual_data_received',
    averageReceiveRate='Average_receive_rate'
)
INFO_FIELDS = dict(
    target=list(INFO_SUBSTITUTES)[:4],
    source=list(INFO_SUBSTITUTES)[4:]
)
REPLICATION_RENDERERS = dict(
    actualDataSent=StorageCapacityRenderer(),
    actualDataReceived=StorageCapacityRenderer(),
    averageSendRate=ReplicationRateRenderer(),
    averageReceiveRate=ReplicationRateRenderer(),
)


@commandgroup(prefix="replication", parent=DdeAppCommands)
class ReplicationCommands(object):
    @ig_command(
        description="",
        types=(DDEActiveApp,),
    )
    @render_multi_fields(
        value_renderers=REPLICATION_RENDERERS,
        fields=list(INFO_SUBSTITUTES),
        detailed_individually=True,
        substitute=INFO_SUBSTITUTES,
        individually=True,
    )
    def info(self, _api, dde):
        data = []
        for host_kind in list(INFO_FIELDS):
            host_info = ddeget(_api, dde.dde_instance_id, f'rest/{host_kind.capitalize()}Stats', unwrap_result=False)
            accumulator = Munch()
            for host in host_info.get(f'{host_kind}Stat', [Munch({fn: "none" if fn == 'hostName' else 0
                                                                  for fn in ['hostName']+INFO_FIELDS[host_kind][1:]})]):
                for field, value in host.items():
                    if field == 'hostName':
                        accumulator[f'{host_kind}Name'] = accumulator.get(f'{host_kind}Name', [])+[value]
                    else:
                        accumulator[field] = accumulator.get(field, 0)+value
            data = data+[Munch(
                fields=INFO_FIELDS[host_kind],
                title=f'{"Send" if host_kind == "target" else "Receive"} settings and cumulative statistics',
                data=[accumulator]
            )]

        control = {
            "allow_json": True,
            "allow_csv": False,
            "additional_fields": [],
            "fields": list(INFO_SUBSTITUTES)
        }
        return {"control": control, "data": data}

    @ig_command(
        description="Display the DDE APP replication report and last time sync",
        types=(DDEActiveApp, ReplicationShare,),
        options=QueryOptions(fields=REPORT_FIELDS.keys(),
                             substitute=REPORT_FIELDS,
                             detailed=True, ),
    )
    @render(
        fields=SHORT_REPORT_FIELDS,
        detailed_fields=REPORT_FIELDS.keys(),
        substitute=REPORT_FIELDS,
        detailed_individually=True,
    )
    def report(self, _api, dde, sharename=None):
        set_renderer_query_fields()
        if sharename:
            sharename = sharename[0].name
        result = []
        try:
            # this request may be slow. extend timeout but stay below 30 seconds of mgmt proxy timeout
            replications = ddeget(_api, dde.dde_instance_id, 'stats/replication', 29)
            for rep in replications:
                if rep.replication_type != 'FILEDIR':
                    continue
                if sharename is not None and rep.name != sharename:
                    continue
                rep.dde_id = dde.dde_instance_name
                if rep.replication_state != 'Disabled':
                    last_replicated = rep.get('last_transferred', "Unsupported")
                    if last_replicated != "Unsupported":
                        rep.replicated_time = None if last_replicated == "0" else int(last_replicated)
                    else:
                        rep.replicated_time = 'Never'
                    rep.remaining = rep.get('remaining_data_to_transfer', 0)
                    if rep.remaining == 0 and rep.replication_state == 'Enabled':
                        rep.remaining = 'Synchronized'
                    if 'average_transfer_Rate' not in rep:
                        rep.average_transfer_rate = None
                    if 'estimated_duration' not in rep:
                        rep.estimated_duration = 0
                result.append(rep)
        except (ManagementException, UnexpectedAPIResponse, TypeError):
            pass
        return result
