from infinishell.infishell import commandgroup
from .nettools import NetworkCommands, RebootChoice
from .interface import DdeNetworkInterface
from infinishell.infinibox.commands.infiniguard import ig_command, DDEApp
from infinishell.infinibox.commands.infiniguard.lib.dde import get_dde_url
from .. import render
from infinishell.infishell.types import Netmask, MappedChoiceType, IPAddress
from infinishell.infinibox.commands.options import QueryOptions
from infinishell.infinibox.commands.types import Entity
from contextlib import contextmanager
from infinishell.infinibox.commands import EntityCommands, ENTITY,\
    set_renderer_query_fields
from collections import OrderedDict
from itertools import chain


STATIC_ROUTE_FIELDS = OrderedDict(
    devname='Interface',
    network='Destination',
    mask='Netmask',
    gateway='Gateway'
)


class StaticRoute(Entity):
    ENTITY_NAME = 'Static Route'
    PAGED = True
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = True
    NAME_FIELD = "name"
    SORTABLE = False
    URI = 'api/network/static_routes'
    DDE_INSTANCE_ID = None

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)


@contextmanager
def patch_dde_entity(cls, dde: DDEApp):
    class _DdeSpecificEntity(cls.ENTITY_TYPE):
        URI = get_dde_url(dde.dde_instance_id,
                          StaticRoute.URI)
    original = cls.ENTITY_TYPE.DDE_INSTANCE_ID
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = dde.dde_instance_id
    yield
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = original


@commandgroup(prefix="route", parent=NetworkCommands)
class NetworkRouteCommands(EntityCommands):
    ENTITY_TYPE = StaticRoute

    @staticmethod
    def _filter_by_dde(dde, devices):
        """
        Even though the autocomplete is performed after filtering on the dde instance,
        the selected devname from all ddes are passed to the command as a list,
         and must be filtered again for the devname from the correct DDE.
        """
        return next((device for device in devices if device.dde_instance_id == dde.dde_instance_id), None)

    @ig_command(
        description="Display Routing table",
        types=(DDEApp, DdeNetworkInterface),
        options=QueryOptions(fields=STATIC_ROUTE_FIELDS,
                             substitute=STATIC_ROUTE_FIELDS, detailed=True)
    )
    @render(fields=STATIC_ROUTE_FIELDS,
            substitute=STATIC_ROUTE_FIELDS,
            detailed_fields=STATIC_ROUTE_FIELDS,
            detailed_individually=True)
    def query(self, _api, dde, interface=None):
        def with_interface(interface):
            routes = interface.get('routes', [])
            for r in routes:
                r.devname = interface.intfname
            return routes

        set_renderer_query_fields()

        if interface is not None:
            interface = self._filter_by_dde(dde, interface)
            return with_interface(interface)
        else:
            interfaces = DdeNetworkInterface.get_dde(_api, dde)
            return chain.from_iterable(with_interface(i) for i in interfaces)

    @ig_command(
        description="Add a static route to Interface using the specified network IP address, netmask, and gateway",
        types=(DDEApp, DdeNetworkInterface, IPAddress, Netmask, IPAddress, RebootChoice)
    )
    @render(fields=STATIC_ROUTE_FIELDS,
            substitute=STATIC_ROUTE_FIELDS)
    def add(self, _api, dde, interface, network, netmask, gateway, reboot):
        with patch_dde_entity(self.__class__, dde):
            interface = self._filter_by_dde(dde, interface)
            return self._create(_api, returns=ENTITY,
                                devname=interface.intfname,
                                network=network.ip,
                                mask=netmask.netmask,
                                gateway=gateway.ip,
                                reboot=reboot)

    @ig_command(
        description='Delete static route from interface routing table',
        types=(DDEApp, DdeNetworkInterface, IPAddress, RebootChoice)
    )
    def delete(self, _api, dde, interface, network, reboot):
        with patch_dde_entity(self.__class__, dde):
            interface = self._filter_by_dde(dde, interface)
            url = self.ENTITY_TYPE.get_uri(network.ip, 'devices', interface.intfname)
            _api.delete(url,
                        data=dict(reboot=reboot))
