from __future__ import annotations

import re
from functools import partial

from munch import Munch

from infinishell.infinibox.commands.infiniguard.lib.dde import AnyDdeApi, DdeEntityCommands
from infinishell.infishell.types import Range, RegexType, YesNo, Timestamp, EnumType
from infinishell.infishell.utils import strlist
from . import DDEActiveApp, ig_command
from .infinisafe import InfinisafeCommands, infinisafecommandgroup
from .lib.renderer import renderfields
from .lib.tasksupport import report_started_task
from .. import render, roles
from ..options import QueryOptions
from ..types import Entity
from ...renderer.values import ValueTypeRenderer, MillisecondsTimestampRenderer
from ...exceptions import ManagementException, NoSuchEntity
from http import HTTPStatus

QUERY_FIELDS = strlist("snapid ddeapp created_at lock_state expires_at originated_by label")
DETAILED_QUERY_FIELDS = strlist("snapid ddeapp created_at retention lock_state expires_at originated_by label version")
SUBSTITUTE = dict(
    snapid='id',
    ddeapp='dde_app',
    created_at='created',
    expires_at='expiration',
    originated_by='originated',
    lock_state='lock_state',
    label='label',
    retention='retention',
    version='version'
)
POLICY_DICT = Munch({
    'system policy': 'System',
    'manual': 'Manual',
    'pre-recovery': 'Recover'
})


class Snapshot(Entity):
    ENTITY_NAME = "Snapshot"
    PAGED = True
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "snapid"
    UNIQUE_NAME = True
    NAME_FIELD = "snapid"
    URI = 'cr/ddesnapshots'
    PUT_UNWRAPPED = True

    def get_description(self):
        return "InfiniSafe Snapshot"

    @classmethod
    def paged_query(cls, api, params=None, filter=None, sort=None,  # @ReservedAssignment
                    pluck=None, hook=None, **kwargs):
        return super().paged_query(AnyDdeApi(api), params, filter, sort, pluck, hook, **kwargs)

    @classmethod
    def get_by_name(cls, api, name: str):
        try:
            return super().get_by_name(api, name)
        except ManagementException as e:
            if e.http_code == HTTPStatus.BAD_REQUEST and \
                    e.error.code == 'INVALID_QUERY_PARAMETERS':
                raise NoSuchEntity('Command rejected: No such {}: "{}"'.format(cls.get_entity_name(), name))
            else:
                raise

    @classmethod
    def get_by_id(cls, api, id):  # @ReservedAssignment
        return super().get_by_id(AnyDdeApi(api), id)


class Retention(Range):
    MIN = 4
    MAX = 14

    DEFAULT = 7

    def get_description(self):
        return f"The number of days, between {self.MIN} and {self.MAX}, " \
               f"that the snapshot will be retained before it is deleted. The default is {self.DEFAULT}."


class Immutable(YesNo):
    DEFAULT = False

    def get_description(self):
        reverse_mapping = {v: k for k, v in self._get_mapping().items()}
        return super().get_description() + f". The default is {reverse_mapping[self.DEFAULT]}."


class InfinidatRetention(Range):
    MIN = 1
    MAX = 30


class Label(RegexType):
    MAX_LENGTH = 100
    _ALLOWED_SYMBOLS = "^&'@()[]$=!-#{}%.+~_ "

    def _get_pattern(self):
        return r"^[\w{}]*$".format(re.escape(self._ALLOWED_SYMBOLS))

    def _get_flags(self):
        return re.UNICODE

    def _verify_raw(self, value):
        if self.MAX_LENGTH and len(value) > self.MAX_LENGTH:
            return "length must be up to {} characters".format(self.MAX_LENGTH)
        return super(Label, self)._verify_raw(value)

    def _get_error(self):
        return "label contains illegal characters"

    def get_description(self):
        return "A maximum of {} characters, numbers, spaces, and the following symbols: " \
               "\"{}\" (excluding quotation marks). Leading and trailing whitespace characters " \
               "are stripped. Empty label \"\" will remove any existing label".format(self.MAX_LENGTH,
                                                                                      self._ALLOWED_SYMBOLS.strip())


class Expiration(Timestamp):
    def get_description(self):
        return "A timestamp, e.g: '2011-04-13', '2011-04-13 18:30:12'. The date and time when this snapshot will " \
               "expire and be deleted. Expiration must be in the future and equal to or later than the current" \
               " expiration time."


class LockState(EnumType):
    def _get_values(self):
        return 'LOCKED', 'UNLOCKED', 'EXPIRED'


class OriginatedBy(EnumType):
    INVERSE_POLICY = {v: k for k, v in POLICY_DICT.items()}

    def _get_values(self):
        return self.INVERSE_POLICY.keys()

    def _get_actual(self, value):
        try:
            actual_value = self.INVERSE_POLICY[value]
        except KeyError:
            return False, "Invalid origin"

        return True, actual_value


class DaysRenderer(ValueTypeRenderer):
    def render(self, value, empty_value, **properties):  # @UnusedVariable
        return f'{int(value)} days' if isinstance(value, (int, float)) else value


@infinisafecommandgroup(prefix="snapshot", parent=InfinisafeCommands)
class SnapCommands(DdeEntityCommands):
    ENTITY_TYPE = Snapshot

    @ig_command(
        description="List snapshots",
        types=(DDEActiveApp, Snapshot, OriginatedBy, LockState, Label, Timestamp, Timestamp, Timestamp, Timestamp),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, substitute=SUBSTITUTE, detailed=True,
                             sortable=QUERY_FIELDS)
    )
    @renderfields(
        value_renderers=dict(retention=DaysRenderer(), created_at=MillisecondsTimestampRenderer(),
                             expires_at=MillisecondsTimestampRenderer()),
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        substitute=SUBSTITUTE,
        detailed_individually=True
    )
    def query(self, _api, dde=None, snapid=None, originated=None, lock_state=None, label=None,
              created_from=None, created_to=None, expiration_from=None, expiration_to=None):
        snap_filter = _api.create_filter()
        snap_filter = self._filter_range(snap_filter, "created_at", created_from, created_to)
        snap_filter = self._filter_range(snap_filter, "expires_at", expiration_from, expiration_to)
        return self._query(_api, snapid, _hook=partial(self._query_hook), _filter=snap_filter,
                           ddeapp=dde.dde_instance_name if dde else None,
                           originated_by=originated, lock_state=lock_state,
                           label=label)

    def _query_hook(self, snaps):
        for snap in snaps:
            snap.retention = snap.origin.retention
            snap.originated_by = POLICY_DICT.get(snap.originated_by, snap.originated_by)
            if 'label' not in snap:
                snap.label = ''
            yield snap

    @staticmethod
    def _filter_range(snap_filter, field, from_value, to_value):
        if from_value is not None:
            if to_value is not None:
                return snap_filter.between(field, from_value, to_value)
            return snap_filter.ge(field, from_value)
        elif to_value is not None:
            return snap_filter.le(field, to_value)

        return snap_filter

    @ig_command(
        description="Create a snapshot",
        types=(DDEActiveApp, Label, Retention, Immutable)
    )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, dde, label="", retention=Retention.DEFAULT, immutable=Immutable.DEFAULT):
        return self._create(_api,
                            ddeapp=dde.dde_instance_name,
                            label=label,
                            retention=retention,
                            immutable=immutable)

    @ig_command(
        description="Set snapshot label",
        types=(Snapshot, Label,)
    )
    @roles.INFINIDAT.ADMIN
    def setlabel(self, _api, snapid, label):
        return self._update(_api, snapid, label=label)

    @ig_command(
        description="Delete snapshot",
        types=(Snapshot,),
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, snapid):
        return self._delete(_api, snapid)

    def _start_task(self, _api, description, snapshot, operation_name, ddes):
        _api = AnyDdeApi(_api, ddes)
        dderesp = self._postex(_api, snapshot, operation_name)
        ddeid, started_task = (dderesp.ddeid, dderesp.response)

        report_started_task(ddeid, started_task, description)

    @ig_command(
        description="Attach a snapshot to the DDE Node 3",
        types=(Snapshot,),
    )
    @render(disable_renderer_override=True)
    @roles.INFINIDAT.ADMIN
    def attach(self, _api, snapid):
        self._start_task(_api, f'Attaching snapshot {snapid.snapid} to node 3',
                         snapid, 'attach', (1, 2))

    @ig_command(
        description="Recover App from a snapshot",
        types=(Snapshot,),
    )
    @render(disable_renderer_override=True)
    @roles.INFINIDAT.ADMIN
    def recover(self, _api, snapid):
        snapshot_ddeid = {'a': 1, 'b': 2}[snapid.ddeapp]
        ddes = {3, 1, 2} - {snapshot_ddeid}
        self._start_task(_api, f'Recover from snapshot {snapid.snapid}',
                         snapid, 'recover', ddes)

    @ig_command(
        description="Detach snapshot from DDE node 3",
    )
    @render(disable_renderer_override=True)
    @roles.INFINIDAT.ADMIN
    def detach(self, _api):
        self._start_task(_api, f'Detaching snapshot from DDE node 3',
                         None, 'detach', (1, 2))

    @ig_command(
        description="Lock an unlocked snapshot",
        types=(Snapshot, Expiration)
    )
    @roles.INFINIDAT.ADMIN
    def lock(self, _api, snapid, expiration=None):
        if not expiration:
            immutable = True
        else:
            immutable = None if snapid.lock_state == 'LOCKED' else True

        self._update(_api, snapid, immutable=immutable,
                     expires_at=expiration, _multi_field_put=True)
        new_snap = Snapshot().get_by_id(_api, snapid.snapid)
        return f'Snapshot "{snapid.snapid}" is now locked and will expire at ' \
               f'{MillisecondsTimestampRenderer().render(new_snap.expires_at, "-")}'
