from __future__ import annotations

import concurrent.futures
import logging
from collections import namedtuple
from concurrent.futures import Future
from concurrent.futures import ThreadPoolExecutor as Executor
from http import HTTPStatus
from typing import Callable, Any, List, Dict, Tuple

from infinishell.infinibox.api import InfiniboxAPI
from infinishell.infinibox.exceptions import (
    Forbidden,
    Unauthorized,
    ManagementException,
    CommandFailed,
    ServiceTemporarilyUnavailable,
)
from .timedcache import timedcache
from ... import EntityCommands, MESSAGE
from distutils.version import LooseVersion

log = logging.getLogger(__name__)


def get_dde_url(ddeid, ddeuri):
    return f'../dde/{ddeid}/{ddeuri}'


AnyDdeApiResponse = namedtuple('AnyDdeApiResponse', ('ddeid', 'response'))


class AnyDdeApi:
    def __init__(self, _api: InfiniboxAPI, ddes: Tuple = (3, 1, 2), filter_responsive: bool = True):
        self._api = _api
        self.ddes = ddes
        self.filter_responsive = filter_responsive

    def is_aggregating(self):
        return self._api.is_aggregating()

    def exec(self, f: Callable[[int], Any]) -> AnyDdeApiResponse:

        ddes = self.ddes
        if self.filter_responsive:
            ddes = set(ddes).intersection(responsiveddes(self._api))

        if not ddes:
            raise CommandFailed(f'There is no responsive DDEs left to work with.')

        for ddeid in ddes:
            try:
                return AnyDdeApiResponse(ddeid, f(ddeid))
            except (Unauthorized, Forbidden):
                log.exception(f'Unauthorized/Forbidden on ddeid {ddeid}.')
                raise
            except ManagementException as e:
                if e.http_code == HTTPStatus.CONFLICT and e.error.code == 'DDE_VERSION_OUTDATED':
                    log.exception(f'Ignoring ManagementException(old version) while probing DDE {ddeid}.')
                elif e.http_code in (HTTPStatus.UNAUTHORIZED,
                                     HTTPStatus.FORBIDDEN,
                                     HTTPStatus.CONFLICT,
                                     HTTPStatus.BAD_REQUEST):
                    log.exception(f'ManagementException due to HTTP error response on ddeid {ddeid}')
                    raise
            except ServiceTemporarilyUnavailable as e:
                log.exception(f'DDE {ddeid} is down. Ignoring and moving on.')
            except CommandFailed:
                raise
            except Exception:
                log.exception(
                    f'Ignoring ManagementException while probing DDE {ddeid}.')

        log.error(f'Could not get response from target DDEs {ddes}.')
        raise CommandFailed(f'Could not get response from any DDEs.')

    def _request_method(self, method: Callable, ddeuri: str, *args, **kwargs) -> AnyDdeApiResponse:
        log.debug(f'{method} ( {ddeuri}, {args}, {kwargs} )')
        return self.exec(lambda ddeid: method(get_dde_url(ddeid, ddeuri), *args, **kwargs))

    def getex(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self._request_method(self._api.get, ddeuri, *args, **kwargs)

    def get(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self.getex(ddeuri, *args, **kwargs).response

    def putex(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self._request_method(self._api.put, ddeuri, *args, **kwargs)

    def put(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self.putex(ddeuri, *args, **kwargs).response

    def postex(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self._request_method(self._api.post, ddeuri, *args, **kwargs)

    def post(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self.postex(ddeuri, *args, **kwargs).response

    def deleteex(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self._request_method(self._api.delete, ddeuri, *args, **kwargs)

    def delete(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self.deleteex(ddeuri, *args, **kwargs).response

    def paged_queryex(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self._request_method(self._api.paged_query, ddeuri, *args, **kwargs)

    def paged_query(self, ddeuri, *args, **kwargs) -> AnyDdeApiResponse:
        return self.paged_queryex(ddeuri, *args, **kwargs).response


DEFAULT_DDE_GET_TIMEOUT = 10


def safe_version(s: str) -> LooseVersion:
    return LooseVersion(s or '0')


def is_standby_app_with_old_version(node, expected_iguard_ver: LooseVersion):
    '''
    Old snapshot attached ?
    '''
    return node.dde_instance_id == 3 and \
        safe_version(node.dde_instance_version) < expected_iguard_ver


@timedcache(2 * 60)
def responsiveddes(api) -> List[int]:
    nodes = api.get('integrations/infiniguard/dde_nodes')
    max_iguard_ver = max(safe_version(n.dde_instance_version) for n in nodes)

    VALID_LIFECYCLES = ['ACTIVE', 'FRU_STANDBY']
    ddes = [n.node_id for n in nodes
            if n.state.lifecycle in VALID_LIFECYCLES and
            not is_standby_app_with_old_version(n, max_iguard_ver)]
    return ddes


def ddeget(api, ddeid: int, path: str, timeout=DEFAULT_DDE_GET_TIMEOUT, **kwargs):
    return api.get(get_dde_url(ddeid, path), timeout=timeout, **kwargs)


def ddepost(api, ddeid: int, path: str, data, timeout=DEFAULT_DDE_GET_TIMEOUT, **kwargs):
    return api.post(get_dde_url(ddeid, path), data, timeout=timeout, **kwargs)


def allddeget(api, path: str, timeout=DEFAULT_DDE_GET_TIMEOUT, ddes=(1, 2, 3)) -> Dict[int, Any]:
    responsive = responsiveddes(api)
    unresponsive = set(ddes) - set(responsive)

    executor = Executor()
    futs = {executor.submit(ddeget, api, ddeid, path, timeout): ddeid for ddeid in responsive}
    executor.shutdown(wait=False)

    def futdata(ddeid, future: Future) -> Tuple[int, Any]:
        try:
            return ddeid, future.result()
        except Exception:
            log.exception(f'Could not get {path} from ddeid {ddeid}')
            return ddeid, None

    result = dict(futdata(futs[future], future)
                  for future in concurrent.futures.as_completed(futs, timeout=timeout))
    result.update({ddeid: None for ddeid in unresponsive})
    return result


class DdeEntityCommands(EntityCommands):

    def _post(self, _api, entity, *uri, **kwargs):
        return super()._post(AnyDdeApi(_api), entity, *uri, **kwargs)

    def _postex(self, _api: InfiniboxAPI | AnyDdeApi, entity, *uri, **kwargs) -> AnyDdeApiResponse:
        if not isinstance(_api, AnyDdeApi):
            _api = AnyDdeApi(_api)
        entity_type = kwargs.pop("entity_type", None)
        return _api.postex(self._uri(entity, *uri, entity_type=entity_type),
                           **kwargs)

    def _create(self, _api, returns=MESSAGE, **properties):
        return super()._create(AnyDdeApi(_api), returns, **properties)

    def _delete_entity(self, _api, entity, **kwargs):
        return super()._delete_entity(AnyDdeApi(_api), entity, **kwargs)

    def _put(self, _api, entity, *uri, **kwargs):
        if entity_type := kwargs.get('entity_type'):
            if entity_type.PUT_UNWRAPPED:
                kwargs.update(unwrap_result=False)
        return super()._put(AnyDdeApi(_api), entity, *uri, **kwargs)
