from infinishell.infinibox.renderer.renderer import InfiniboxRenderer, MultiCollectionRenderer
from infinishell.infinibox.renderer.values import (ValueTypeRenderer,
                                                   ValueRenderer)
from typing import Dict
from infinishell.infishell.commands import metadata
from infinishell.infinibox.renderer.json import JSONRenderer, is_json_mode
from infinishell.infishell.renderer import Renderer
from infinishell.infinibox.renderer.csv import CSVRenderer, is_csv_mode
from infinishell.infinibox.renderer.count import is_count_mode, COUNTRenderer


class FieldsValueRenderer(ValueRenderer):
    def __init__(self, value_renderers: Dict[str, ValueTypeRenderer], empty_value="-"):
        self.value_renderers = value_renderers
        super().__init__(empty_value)

    def render(self, field, value, **properties):
        renderer = self.value_renderers.get(field)
        if renderer:
            return str(renderer.render(value, self._empty_value, **properties))
        else:
            return super().render(field, value, **properties)

    def is_left_aligned(self, field):
        renderer = self.value_renderers.get(field)
        return renderer.LEFT_ALIGNED if renderer else \
            super().is_left_aligned(field)


class FieldsRenderer(Renderer):
    def __init__(self, value_renderers: Dict[str, ValueTypeRenderer], general=InfiniboxRenderer):
        self.general = general()
        self.json = JSONRenderer()
        self.csv = CSVRenderer()

        for r, e in [(self.general, '-'), (self.json, ''), (self.csv, '')]:
            r._value_renderer = FieldsValueRenderer(value_renderers, empty_value=e)

    def render(self, obj, stream, color=True, **kwargs):
        general_only = kwargs.pop('general_only', False)

        renderer = None
        if not general_only:
            if is_json_mode():
                renderer = self.json
            elif is_csv_mode():
                renderer = self.csv
            elif is_count_mode():
                renderer = COUNTRenderer()

        if not renderer:
            renderer = self.general

        return renderer.render(obj, stream, color, **kwargs)


def renderfields(value_renderers: Dict[str, ValueTypeRenderer], func=None, **kwargs):
    return metadata(func,
                    renderer=FieldsRenderer(value_renderers),
                    disable_renderer_override=True,
                    render_kwargs=kwargs)


def render_multi_fields(value_renderers: Dict[str, ValueTypeRenderer], func=None, **kwargs):
    return metadata(func,
                    renderer=FieldsRenderer(value_renderers, general=MultiCollectionRenderer),
                    disable_renderer_override=True,
                    render_kwargs=kwargs)


class SuspendDeleteStateRenderer(ValueTypeRenderer):
    def render(self, value, empty_value, **properties):  # @UnusedVariable
        return {True: 'Suspended',
                False: 'As scheduled'
                }.get(value, value)
