from functools import partial

from infinishell.infishell import commandgroup
from infinishell.infishell.exceptions import CommandAborted
from infinishell.infishell.types import Range
from infinishell.infishell.utils import strlist, yesno
from . import ig_command, DDEApp, DDENode
from .. import EntityCommands, render, roles
from ..infiniguard.lib.dde import ddepost, AnyDdeApi
from ..options import QueryOptions, AutoApprove
from infinishell.infishell.shell import get_active_shell
from infinishell.infinibox.commands.utils import is_option_set

QUERY_FIELDS = strlist("dde_instance_name node_id serial status health capacity used_capacity free_capacity")
DETAILED_QUERY_FIELDS = strlist(
    "dde_instance_name node_id serial status health capacity used_capacity "
    "free_capacity reduction_ratio")
SUBSTITUTE = dict(
    dde_instance_name='dde_app',
    node_id='node_id',
    free_capacity='free',
    used_capacity='used'

)


class RebootWait(Range):
    MIN = 1
    MAX = 600


class ShutdownWait(Range):
    MIN = 1
    MAX = 600


def get_dde_map(_api):
    return AnyDdeApi(_api).get('mgmt/dde-app-map')


def get_app_to_node_dict(_api):
    return {mapping.app: mapping.nodeid for mapping in get_dde_map(_api)}


def get_node_to_app_dict(_api):
    return {mapping.nodeid: mapping.app for mapping in get_dde_map(_api)}


def _snap_attached(_api):
    infinisafe_enabled = _api.get('integrations/config/infinisafe/infinisafe.enable')
    return AnyDdeApi(_api).get('cr/ddesnapshots/attachment').attached if infinisafe_enabled else False


def get_app_name(_api, dde_instance_name):
    """
    Return a printable name of the DDE app.
    """
    app_names = {
        'a': 'DDE App A',
        'b': 'DDE App B',
        'standby': 'Standby DDE App (with attached snapshot)' if _snap_attached(_api) else 'Standby DDE App'
    }

    return app_names[dde_instance_name]


@commandgroup(prefix="ddeapp")
class DdeAppCommands(EntityCommands):
    ENTITY_TYPE = DDEApp

    @ig_command(
        description="List all Dedup Engines",
        types=(DDEApp.List,),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS,
                             substitute=SUBSTITUTE, detailed=True)
    )
    @render(
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        substitute=SUBSTITUTE,
        detailed_individually=True
    )
    def query(self, _api, dde=None):
        if not dde:
            dde = [d for d in DDEApp.iter_all(_api)]
        return self._query(_api, dde, _hook=partial(self._query_hook))

    def _query_hook(self, ddes):
        for dde in ddes:
            dde.status = dde.state.lifecycle
            dde.health = dde.state.health
            dde.capacity = dde.state.total_disk_capacity if dde.state.total_disk_capacity > 0 else None
            dde.free_capacity = dde.state.available_disk_capacity if dde.state.available_disk_capacity > 0 else None
            dde.used_capacity = dde.state.used_disk_capacity if dde.state.used_disk_capacity > 0 else None
            dde.reduction_ratio = dde.state.total_reduction_ratio
            yield dde

    @ig_command(
        description="Perform DDE App Reboot",
        types=(DDEApp, RebootWait),
    )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def reboot(self, _api, dde, waittime=None):
        node_id = get_app_to_node_dict(_api)[dde.dde_instance_name]
        app_name = get_app_name(_api, dde.dde_instance_name)
        node_name = f'DDE Node {node_id}'

        if not is_option_set(AutoApprove()):
            confirm_msg = f"{app_name} running on {node_name} will reboot."

            get_active_shell().error(confirm_msg)
            if not yesno():
                raise CommandAborted

        body = dict() if waittime is None else dict(wait_time=waittime)
        ddepost(_api, dde.dde_instance_id, "api/node/reboot", body, params=dict(approved='true'))

        return f"{app_name} will reboot {'now' if waittime is None else f'in {waittime} seconds'}"

    @ig_command(
        description="Perform DDE App Shutdown",
        types=(DDEApp, RebootWait),
    )
    @roles.INFINIDAT.ADMIN
    def shutdown(self, _api, dde, waittime=None):
        app_name = get_app_name(_api, dde.dde_instance_name)

        body = dict() if waittime is None else dict(wait_time=waittime)
        ddepost(_api, dde.dde_instance_id, "api/node/shutdown", body)
        return f"{app_name} will shut down {'now' if waittime is None else f'in {waittime} seconds'}"


NODE_QUERY_FIELDS = strlist("node_id dde_instance_name serial capacity free_capacity status health")
INFINIDAT_NODE_FIELDS = strlist("id node_id dde_instance_name serial capacity free_capacity status health")
NODE_DETAILED_FIELDS = strlist(
    "node_id dde_instance_name serial capacity free_capacity status health color location label")
INFINIDAT_NODE_DETAILED_FIELDS = strlist(
    "id node_id dde_instance_name serial dde_instance_version install_timestamp upgrade_timestamp capacity "
    "free_capacity status health color location label")
NODE_SUBSTITUTE = dict(
    id='internal_node_id',
    dde_instance_name='dde_app_id',
    node_id='node_id',
    location='location',
    capacity='total_capacity',
    free_capacity='available_capacity',
    color='rack_marker_color',
    serial='serial_number',
    dde_instance_version='infiniguard_version'
)


@commandgroup(prefix="ddenode")
class DdeNodeCommands(EntityCommands):
    ENTITY_TYPE = DDENode

    @ig_command(
        description="List all physical DDE nodes",
        types=(DDENode.List,),
        options=QueryOptions(fields=NODE_DETAILED_FIELDS, infinidat_fields=INFINIDAT_NODE_FIELDS,
                             substitute=SUBSTITUTE, detailed=True)
    )
    @render(
        fields=NODE_QUERY_FIELDS,
        detailed_fields=NODE_DETAILED_FIELDS,
        infinidat=dict(
            fields=INFINIDAT_NODE_FIELDS,
            detailed_fields=INFINIDAT_NODE_DETAILED_FIELDS,
            detailed_individually=True,),
        substitute=NODE_SUBSTITUTE,
        detailed_individually=True,
    )
    def query(self, _api, ddenode=None):
        if not ddenode:
            ddenode = [d for d in DDENode.iter_all(_api)]
        return self._query(_api, ddenode, _hook=partial(self._query_hook))

    def _query_hook(self, ddes):
        for dde in ddes:
            dde.status = dde.state.lifecycle
            dde.health = dde.state.health
            dde.capacity = dde.state.total_disk_capacity if dde.state.total_disk_capacity > 0 else None
            dde.free_capacity = dde.state.available_disk_capacity if dde.state.available_disk_capacity > 0 else None
            dde.used_capacity = dde.state.used_disk_capacity if dde.state.used_disk_capacity > 0 else None
            dde.reduction_ratio = dde.state.total_reduction_ratio
            yield dde

    @ig_command(
        description="Perform DDE Node power cycle via lights-out-management",
        types=(DDENode, RebootWait),
    )
    @roles.INFINIDAT.TECHNICIAN
    def powercycle(self, _api, ddenode, waittime=None):
        node_id = ddenode.node_id
        node_name = f'DDE Node {node_id}'

        body = dict(nodeid=node_id) if waittime is None else dict(nodeid=node_id, waittime=waittime)
        AnyDdeApi(_api).post("mgmt/lom-powercycle", body)

        return f"{node_name} will perform a power cycle {'now' if waittime is None else f'in {waittime} seconds'}"

    @ig_command(
        description="Perform DDE Node power off via lights-out-management",
        types=(DDENode, RebootWait),
    )
    @roles.INFINIDAT.ADMIN
    def poweroff(self, _api, ddenode, waittime=None):
        node_id = ddenode.node_id
        node_name = f'DDE Node {node_id}'

        body = dict(nodeid=node_id) if waittime is None else dict(nodeid=node_id, waittime=waittime)
        AnyDdeApi(_api).post("mgmt/lom-poweroff", body)

        return f"{node_name} will power off {'now' if waittime is None else f'in {waittime} seconds'}"
