from __future__ import annotations  # for Union pipe '|'
from infinishell.infishell import commandgroup
from . import ig_command
from collections import OrderedDict
from ..options import QueryOptions
from ...commands import EntityCommands
from .lib.tasksupport import Task
from infinishell.infishell.shell import get_active_shell
from infinishell.infinibox.commands.utils import getch_pause
from .lib.renderer import renderfields
from infinishell.infinibox.renderer.values import ValueTypeRenderer
import time
from infinishell.infinibox.exceptions import APIFailure, Prefixless
from .general import InfiniguardCommands

TASK_FIELDS = OrderedDict(
    id='id',
    type='type',
    created_at='start',
    status='status',
    progress='progress',
    timelimit='time_limit',
)

DETAILED_TASK_FIELDS = OrderedDict(
    id='ID',
    type='Type',
    created_at='Start',
    modified_at='Task Update',
    completed_at='End',
    status='Status',
    progress='Progress',
    timelimit='Time limit',
    description='Description'
)

WATCH_TASK_FIELDS = OrderedDict(
    id='ID',
    type='Type',
    created_at='Start',
    timelimit='Time limit',
    description='Description'
)


class ProgressGaugeRenderer(ValueTypeRenderer):
    def render(self, value, empty_value, **properties):  # @UnusedVariable
        if not isinstance(value, (int, float)):
            return value
        progress = value
        chars = round(progress / 10)
        if chars > 10:
            chars = 10
        filled = '#' * chars
        progressbar = filled.ljust(10)
        return f'[{progressbar}] {round(progress)}%'


class PercentRenderer(ValueTypeRenderer):
    def render(self, value, empty_value, **properties):  # @UnusedVariable
        if not isinstance(value, (int, float)):
            return value
        progress = value
        return f'{round(progress)}%'


class DdeAppRenderer(ValueTypeRenderer):
    def render(self, value, empty_value, **properties):  # @UnusedVariable
        return f'DDE App {value.capitalize()}' \
            if value.lower() in ['a', 'b', 'standby'] else value


class Seconds2MinutesRenderer(ValueTypeRenderer):
    def render(self, value, empty_value, **properties):  # @UnusedVariable
        if isinstance(value, (int, float)):
            minutes = round(value / 60)
            return f'{minutes} min'
        else:
            return value


TASK_RENDERERS = dict(
    controller=DdeAppRenderer(),
    target=DdeAppRenderer(),
    target1=DdeAppRenderer(),
    progress=PercentRenderer(),
    timelimit=Seconds2MinutesRenderer()
)

WATCH_TASK_RENDERERS = TASK_RENDERERS.copy()
WATCH_TASK_RENDERERS['progress'] = ProgressGaugeRenderer()


def render_command_result(obj):
    shell = get_active_shell()
    stream, renderer, color = (shell.output_stream,
                               shell.get_renderer_override(),
                               shell.has_colors())

    spec = shell.current_command.spec
    return spec.renderer.render(obj, stream, renderer=renderer,
                                color=color,
                                **spec.render_kwargs)


class Timeout:
    def __init__(self, start_ms: int, timelimit_sec: int):
        self.start_ms = start_ms
        self.timelimit_sec = timelimit_sec
        self.end_ms = self.start_ms + self.timelimit_sec * 1000

    def expired(self) -> bool:
        t = time.time()
        t_ms = int(round(t * 1000))
        return t_ms > self.end_ms


def cursor_up(lines):
    return f'\033[{lines}A'


class TaskCompletionTimeout(APIFailure, Prefixless):
    pass


class NoRunningTasksFound(APIFailure, Prefixless):
    pass


class TooManyTasksRunning(APIFailure, Prefixless):
    pass


@commandgroup(parent=InfiniguardCommands, prefix="task")
class InfiniguardTask(EntityCommands):
    ENTITY_TYPE = Task

    @ig_command(
        description="Display tasks information",
        options=QueryOptions(fields=TASK_FIELDS.keys(),
                             substitute=TASK_FIELDS,
                             detailed=True),
        types=(Task,)
    )
    @renderfields(
        value_renderers=TASK_RENDERERS,
        fields=TASK_FIELDS.keys(),
        substitute=DETAILED_TASK_FIELDS,
        detailed_fields=DETAILED_TASK_FIELDS.keys(),
        detailed_individually=True)
    def query(self, _api, task=None):
        return self._query(_api, task)
