from . import DDEActiveApp, ig_command
from .nodes import DdeAppCommands
from .lib.dde import allddeget
from .lib.asynctask import AsyncTask
from infinishell.infishell import commandgroup, runtime, get_active_shell
from infinishell.infishell.utils import strlist
from ..types import Size, Entity
from infinishell.infishell.types import Range
from .. import render, set_renderer_query_fields
from infinishell.infishell.exceptions import BadArguments
from infinishell.infinibox.commands import utils
from infinishell.infinibox.commands.options import OutputFormat, Grep, QueryOptions
from itertools import chain
from enum import Enum
from munch import Munch
from collections import OrderedDict
from time import time, sleep

NAS_URI = 'api/nas/app_specific/shares'


class AppSpecificShare(Entity):
    ENTITY_NAME = "Share"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = True
    NAME_FIELD = "name"
    URI = NAS_URI
    FILTERABLE = False

    def get_description(self):
        return "App-specific share"

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        res = allddeget(api, cls.URI)

        unique_shares = {share.name: share for share in
                         chain.from_iterable(share_list for share_list in res.values() if share_list)}
        sorted_shares = [share for _id, share in sorted(unique_shares.items())]

        if kwargs.get("hook"):
            sorted_shares = kwargs["hook"](sorted_shares)
        return sorted_shares


class FileModeFlags(Enum):
    FMF_X_OTH = 1 << 0
    FMF_W_OTH = 1 << 1
    FMF_R_OTH = 1 << 2
    FMF_X_GRP = 1 << 3
    FMF_W_GRP = 1 << 4
    FMF_R_GRP = 1 << 5
    FMF_X_USR = 1 << 6
    FMF_W_USR = 1 << 7
    FMF_R_USR = 1 << 8
    FMF_D_FLG = 1 << 14

    def __str__(self):
        return self.name.split("_")[1].lower()

    def __int__(self):
        return self.value

    def __eq__(self, other):
        return self.value == other

    def __rand__(self, other):
        return self.value & other

    @classmethod
    def find_all(cls, candidate):
        members = cls._member_names_
        return [str(cls[member]) if candidate & cls[member] else '-' for member in members]


class FileAge(Range):
    MIN = 0
    MAX = 50000

    def get_description(self):
        return f"Minimal age of file in days. If specified, only files older than specified, are displayed"


CONTENT_FIELDS = OrderedDict(
    permissions='flags',
    uid='uid',
    gid='gid',
    time='time',
    size='size',
    path='name'
)
SHORT_FIELDS = strlist('size path')
SORTABLE_FIELDS = strlist('path size time')

@commandgroup(prefix="appspecific", parent=DdeAppCommands)
class AppSpecificCommands:
    FILES_URI = 'list_files'

    @ig_command(
        description="View a list of all existing Application Specific shares",
        types=(DDEActiveApp,),
        options=[OutputFormat(), Grep()],
    )
    def query(self, _api, dde):
        uri = f'../dde/{dde.dde_instance_id}/{NAS_URI}'
        res = _api.get(uri)
        shares = [Munch(name=share.name,) for share in res]
        return shares

    def _share_hook(self, fileinst, share):
        fileinst.path = f'{share}/{fileinst.get("filename")}'
        fileinst.permissions = "".join(FileModeFlags.find_all(fileinst.mode)[::-1])
        fileinst.time = int(fileinst.mtime) * 1000
        return fileinst

    def create_sort(self, _api, sort_fields):
        sort = _api.create_sort()
        for field, is_ascending in sort_fields or ():
            sort = sort.sort(field, is_ascending)
        return sort

    @ig_command(
        description="View a list of Application Specific share content",
        types=(DDEActiveApp, AppSpecificShare, FileAge),
        options=QueryOptions(fields=CONTENT_FIELDS.keys(), substitute=CONTENT_FIELDS, sortable=SORTABLE_FIELDS,
                             detailed=True,)
    )
    @render(
        fields=SHORT_FIELDS,
        detailed_fields=CONTENT_FIELDS.keys(),
        substitute=CONTENT_FIELDS,
        detailed_individually=False,
    )
    def content(self, _api, dde, share, age=None):
        set_renderer_query_fields()
        sort_fields = runtime.options.get("sort", None)
        sort = self.create_sort(_api, sort_fields) if sort_fields else None
        share_uri = f'../dde/{dde.dde_instance_id}/{NAS_URI}'
        shares = _api.get(share_uri)
        if share.name not in (s.name for s in shares):
            raise BadArguments(f'Share {share.name} not found on DDE App {dde.dde_instance_name}')
        age_query = f'?mtime=lte:{int(time())-(age*3600*24)}' if age else ''
        files_uri = f'{NAS_URI}/{share.name}/{self.FILES_URI}{age_query}'
        result = []
        can_quit = get_active_shell().is_interactive_terminal()
        with AsyncTask(_api, dde.dde_instance_id, files_uri) as ls:
            while True:
                done = ls.is_task_finished()
                files = ls.get_files()
                if files:
                    result = result+files
                if done and not files:
                    break
                else:
                    if can_quit:
                        utils.getch_pause(1)
                    else:
                        sleep(1)
        files = [self._share_hook(file, share.name) for file in result]
        return sort.sorted_entities(files) if sort else files
