from . import DDEApp, ig_command
from .nodes import DdeAppCommands
from .lib.dde import allddeget, get_dde_url, ddeget
from .lib.context_entity import ContextSensitiveEntity, ContextMixin, filtered
from .lib.asynctask import AsyncTask
from infinishell.infishell import commandgroup, get_active_shell, render
from infinishell.infishell.utils import strlist
from infinishell.infishell.types import Hostname, IPAddress, Range, YesNo, EnumType, Positive, MappedChoiceType, Str
from infinishell.infinibox.commands import utils
from ..types import Size
from infinishell.infishell.exceptions import BadArguments
from enum import Enum
import ipaddress
import capacity
import time
from infinishell.utils import yesno
from munch import Munch

INTERFACE_DETAILS_URI = 'api/network/tools/interface_details'

NUM_BOND_DEVICES = 9


class Reboot:
    NOW = "now"
    DEFER = "defer"


class RebootChoice(MappedChoiceType):
    MAPPING = dict(
        now=Reboot.NOW,
        defer=Reboot.DEFER
    )

    def _get_mapping(self):
        return self.MAPPING


class IfconfigFlags(Enum):
    IFF_UP = 1 << 0
    IFF_BROADCAST = 1 << 1
    IFF_DEBUG = 1 << 2
    IFF_LOOPBACK = 1 << 3
    IFF_POINTOPOINT = 1 << 4
    IFF_NOTRAILERS = 1 << 5
    IFF_RUNNING = 1 << 6
    IFF_NOARP = 1 << 7
    IFF_PROMISC = 1 << 8
    IFF_ALLMULTI = 1 << 9
    IFF_MASTER = 1 << 10
    IFF_SLAVE = 1 << 11
    IFF_MULTICAST = 1 << 12
    IFF_PORTSEL = 1 << 13
    IFF_AUTOMEDIA = 1 << 14
    IFF_DYNAMIC = 1 << 15
    IFF_LOWER_UP = 1 << 16
    IFF_DORMANT = 1 << 17
    IFF_ECHO = 1 << 18

    def __str__(self):
        return " ".join(self.name.split("_")[1:])

    def __int__(self):
        return self.value

    def __eq__(self, other):
        return self.value == other

    def __rand__(self, other):
        return self.value & other

    @classmethod
    def find_all(cls, candidate):
        members = cls._member_names_
        return [str(cls[member]) for member in members if candidate & cls[member]]


class DdeNetworkDevice(ContextSensitiveEntity):
    ENTITY_NAME = "DDE Network Device"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = False
    NAME_FIELD = "name"
    URI = 'api/network/ethernet_ports'
    FILTERABLE = False
    DDE_INSTANCE_ID = None

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    def get_description(self):
        name = self.ENTITY_NAME
        return name if self._get_context(argtype=DDEApp, raw=False) else \
            f"{name}: please specify DDE argument first"

    def _context_hook(self, results, **kwargs):
        filter_intf_type = kwargs.get('filter_intf_type', True)
        dde = self._get_context(argtype=DDEApp, raw=False)
        interface_type = self._get_context(param_name='interface_type')
        if not dde:
            return []

        return [intf for intf in results if (intf.dde_instance_id == dde.dde_instance_id
                                             and (filter_intf_type is False or interface_type is None or
                                                  intf.get('type').lower() == interface_type))]

    @classmethod
    def _entity_hook(cls, intf):
        intf[cls.NAME_FIELD] = intf.device_name
        return intf

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        res = allddeget(api, cls.URI)

        alldde_intfs = []
        for dde_id, intf_list in res.items():
            if intf_list:
                for intf in intf_list:
                    alldde_intfs.append(cls._entity_hook(Munch(dde_instance_id=dde_id, **intf)))

        if kwargs.get("hook"):
            alldde_intfs = kwargs["hook"](alldde_intfs)

        return sorted(alldde_intfs, key=lambda intf: intf.device_name)

    @classmethod
    def get_dde(cls, api, dde: DDEApp, **kwargs):
        intfs = ddeget(api, dde.dde_instance_id, cls.URI, **kwargs)

        if kwargs.get("hook"):
            intfs = kwargs["hook"](intfs)
        return intfs


class DdeNetworkDeviceStaticBonds(DdeNetworkDevice):
    """
    Dde network device that returns the ethernet_port list for 'port' type, and a static list (bondx, x=[0-p)
    for 'bond' type
    """

    @staticmethod
    def _generate_bond_entity(bond_id):
        return Munch(name=f'bond{bond_id}', device_name=f'bond{bond_id}', dde_instance_id=None)

    def _context_hook(self, results, **kwargs):
        filter_intf_type = kwargs.get('filter_intf_type', True)
        interface_type = self._get_context(param_name='interface_type')

        if interface_type == 'bond' and filter_intf_type:
            return [self._generate_bond_entity(bond_id) for bond_id in range(NUM_BOND_DEVICES)]

        else:
            return super()._context_hook(results, **kwargs)

    @classmethod
    def get_by_name(cls, api, name):
        if name.startswith('bond'):
            return [cls._generate_bond_entity(bond_id=name[len('bond'):])]
        else:
            return super().get_by_name(api, name)


class DdeNetworkInterface(ContextSensitiveEntity):
    ENTITY_NAME = "Interface"
    DESCRIPTION = "DDE Interface"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "intfname"
    UNIQUE_NAME = False
    NAME_FIELD = "intfname"
    URI = 'api/network/interfaces'
    MULTI_FIELD_PUT = True
    FILTERABLE = False
    DDE_INSTANCE_ID = None

    def get_description(self):
        name = self.ENTITY_NAME
        return name if self._get_context(argtype=DDEApp, raw=False) else \
            "{name}: please specify DDE first"

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    def _context_hook(self, results, **kwargs):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if not dde:
            return []

        return [intf for intf in results if intf.dde_instance_id == dde.dde_instance_id]

    @classmethod
    def _entity_hook(cls, intf):
        intf[cls.NAME_FIELD] = intf.intfname
        return intf

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        res = allddeget(api, cls.URI)

        alldde_intfs = []
        for dde_id, intf_list in res.items():
            if intf_list:
                for intf in intf_list:
                    alldde_intfs.append(cls._entity_hook(Munch(dde_instance_id=dde_id, **intf)))

        if kwargs.get("hook"):
            alldde_intfs = kwargs["hook"](alldde_intfs)

        return sorted(alldde_intfs, key=lambda intf: intf.intfname)

    @classmethod
    def get_dde(cls, api, dde: DDEApp, **kwargs):
        intfs = ddeget(api, dde.dde_instance_id, cls.URI, **kwargs)

        if kwargs.get("hook"):
            intfs = kwargs["hook"](intfs)
        return intfs


class DdeNetworkVlanInterface(ContextSensitiveEntity):
    ENTITY_NAME = "Device / VLAN Interface"
    DESCRIPTION = "DDE Device name or VLAN Interface"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = False
    NAME_FIELD = "name"
    URI = INTERFACE_DETAILS_URI
    FILTERABLE = False

    def get_description(self):
        name = self.ENTITY_NAME
        return name if self._get_context(argtype=DDEApp, raw=False) else \
            "{name}: please specify DDE first"

    def _context_hook(self, results, **kwargs):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if not dde:
            return []

        return [intf for intf in results if intf.dde_instance_id == dde.dde_instance_id]

    @classmethod
    def _entity_hook(cls, intf):
        intf[cls.NAME_FIELD] = intf.ifname
        return intf

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        res = allddeget(api, cls.URI)

        alldde_intfs = []
        for dde_id, intf_list in res.items():
            if intf_list:
                for intf in intf_list:
                    alldde_intfs.append(cls._entity_hook(Munch(dde_instance_id=dde_id, **intf)))

        if kwargs.get("hook"):
            alldde_intfs = kwargs["hook"](alldde_intfs)

        return sorted(alldde_intfs, key=lambda intf: intf.ifname)


class DDEPort(DdeNetworkDevice):
    def get_description(self):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if dde:
            return "DDE Physical port"
        return "DDE Physical port: please specify DDE first"

    def _context_hook(self, results, **kwargs):
        return [intf for intf in super(DDEPort, self)._context_hook(results) if intf.get('type').lower() == 'port']


class DDEBondMember(DdeNetworkDevice):
    UNIQUE_NAME = True

    def get_description(self):
        return "DDE Physical port for bond member"

    def _context_hook(self, results, **kwargs):
        return [intf for intf in super()._context_hook(results, filter_intf_type=False)
                if intf.get('type').lower() == 'port']


class DDEBondPort(DdeNetworkDevice):
    def get_description(self):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if dde:
            return "DDE Bond port"
        return "DDE Bond port: please specify DDE first"

    def _context_hook(self, results, **kwargs):
        return [intf for intf in super(DDEBondPort, self)._context_hook(results) if intf.get('type').lower() == 'bond']


class IperfVer(EnumType):
    def get_description(self):
        return "IPerf version, either 1 or 3"

    def _get_values(self):
        return [1, 3]

    def _get_actual(self, value):
        return True, int(value)


class Parallel(Positive):
    def get_description(self):
        return "Number of parallel streams"


class DDEIpAddress(IPAddress):
    def get_description(self):
        return "A specific DDE interface (IPv4 address)"


class PayloadSize(Size):
    def get_description(self):
        return "A positive number (up to 2 digits beyond the decimal point)," \
               " optionally followed by a unit; e.g. 1.5KB, 1000. Supported range is 1 - 10000."

    def _verify(self, value):
        error = super(PayloadSize, self)._verify(value)
        if error:
            return error
        if value < 1 or value > 10000:
            return "is out of range"


class Duration(Range):
    MIN = 1
    MAX = 20

    def get_description(self):
        return f"The time duration for running the command, between {self.MIN} and {self.MAX} (sec)"


class IperfDuration(Duration):
    MAX = 30


class IperfBufferSize(Range):
    MIN = 128
    MAX = 1024

    def get_description(self):
        return f"The buffer size for IPerf, between {self.MIN} and {self.MAX} (bytes)"


class ReverseFlag(YesNo, ContextMixin):
    def get_description(self):
        if self._get_context(argtype=IperfVer, raw=False) != 3:
            return "Reverse flag is only supported for iperf v3"
        return "Either yes, or no"

    def get_possible_values(self, _api, starts_with=None):
        if self._get_context(argtype=IperfVer, raw=False) != 3:
            return []
        return super(ReverseFlag, self).get_possible_values(_api, starts_with)


@commandgroup(prefix="network", parent=DdeAppCommands)
class NetworkCommands:
    pass


def async_task_getch(poll_delay: float):
    try:
        utils.getch_pause(poll_delay)
    except KeyboardInterrupt:
        get_active_shell().error(
            "You're going to abort the local command. "
            "Remote command will continue running on the server side until completion.")
        if yesno():
            raise


def _filter_iperf_params(params, context):
    iperf_ver = context(param_name='iperf_ver', raw=False)
    return [arg for arg in params if arg.name != "reverse" or iperf_ver == 3]


@commandgroup(prefix="diagnostics", parent=NetworkCommands)
class NetworkDiagnosticsCommands:
    COMMAND_URI = 'api/network/tools/{}'

    def print_simple_command(self, task: AsyncTask, poll_delay: int = 1):
        ash = get_active_shell()
        poll = async_task_getch if ash.is_interactive_terminal() \
            else time.sleep

        while True:
            out = task.get_lines()
            if task.is_task_finished() and not out:
                break

            for line in out:
                output = ash.error if line.get('stdstream') == 2 \
                    else ash.output
                output(line.get('line'))

            poll(poll_delay)

    @ig_command(
        description="Ping target from specified DDE App",
        types=(DDEApp, Hostname, DDEIpAddress, Duration, PayloadSize, YesNo),
    )
    def ping(self, _api, dde, target, src=None, time=None, payload_size=None, do_not_fragment=None):
        with AsyncTask(_api, dde.dde_instance_id, self.COMMAND_URI.format('ping'), dst=target,
                       src=src.ip if src else None, time=time, payload_size=payload_size,
                       do_not_fragment=do_not_fragment) as ping:
            self.print_simple_command(ping)

    @ig_command(
        description="Trace route to target from specified DDE App",
        types=(DDEApp, Hostname, DDEIpAddress),
    )
    def traceroute(self, _api, dde, target, src=None):
        with AsyncTask(_api, dde.dde_instance_id, self.COMMAND_URI.format('traceroute'), dst=target,
                       src=src.ip if src else None) as trace:
            self.print_simple_command(trace)

    @ig_command(
        description="Dig target from specified DDE App",
        types=(DDEApp, Hostname, Duration),
    )
    def dig(self, _api, dde, target, time=None):
        with AsyncTask(_api, dde.dde_instance_id, self.COMMAND_URI.format('dig'),
                       dst=target, time=time) as dig:
            self.print_simple_command(dig)

    @ig_command(
        description="Run iperf client from DDE to Iperf server",
        types=(DDEApp, Hostname, DDEIpAddress, IperfDuration, IperfBufferSize, IperfVer, Parallel, ReverseFlag)
    )
    @filtered(
        filter_params=_filter_iperf_params
    )
    def iperf(self, _api, dde, target, src=None, time=None, buffer=None, iperf_ver=None, parallel=None, reverse=None):
        if reverse and iperf_ver != 3:
            raise BadArguments("Reverse direction only allowed for Iperf v3")
        with AsyncTask(_api, dde.dde_instance_id, self.COMMAND_URI.format('iperf'), dst=target,
                       src=src.ip if src else None, time=time, buffer_len=buffer, iperf_version=iperf_ver,
                       reverse=reverse, parallel=parallel) as iperf:
            self.print_simple_command(iperf)

    @ig_command(
        description="Calculate MTU between DDE and target",
        types=(DDEApp, Hostname, DDEIpAddress)
    )
    def mtu(self, _api, dde, target, src=None):
        with AsyncTask(_api, dde.dde_instance_id, self.COMMAND_URI.format('mtu'), dst=target,
                       src=src.ip if src else None) as mtu:
            self.print_simple_command(mtu)

    def print_ifconfig(self, intf):
        output_stream = get_active_shell().output_stream
        flags = intf.get('flags', 4099) % (1 << 16)
        stats = intf.get('stats', None)
        flags_verbose = ",".join(IfconfigFlags.find_all(flags))
        hw_addr = intf.get('phys_port_id', intf.hw_addr)
        output_stream.write(f'{intf.ifname + ":": <10} flags: {flags} <{flags_verbose}> mtu={intf.mtu}\n')
        output_stream.write(" " * 11 + f'ether: {hw_addr} txqueuelen {intf.txqlen} (Ethernet)\n')
        output_stream.write(" " * 11 + f'RX packets {stats.rx_packets}  bytes {stats.rx_bytes} '
                                       f'({capacity.capacity.Capacity(stats.rx_bytes * 8)})\n')
        output_stream.write(" " * 11 + f'RX errors {stats.rx_errors}  dropped {stats.rx_dropped}  '
                                       f'overruns {stats.rx_over_errors}  crc {stats.rx_crc_errors}\n')
        output_stream.write(" " * 11 + f'TX packets {stats.tx_packets}  bytes {stats.tx_bytes} '
                                       f'({capacity.capacity.Capacity(stats.tx_bytes * 8)})\n')
        output_stream.write(" " * 11 + f'TX errors {stats.tx_errors}  dropped {stats.tx_dropped}  '
                                       f'carrier {stats.tx_carrier_errors}  aborted {stats.tx_aborted_errors}\n')
        output_stream.write('\n')
        for alias in intf.ip_aliases:
            netmask = ipaddress.IPv4Interface('{}/{}'.format(alias.address, alias.prefixlen)).netmask
            output_stream.write(f'{alias.intfname + ":": <14} flags: {flags} <{flags_verbose}>  mtu={intf.mtu}\n')
            output_stream.write(" " * 15 + f'inet: {alias.address} netmask {netmask} broadcast {alias.broadcast}\n')
            output_stream.write(" " * 15 + f'ether: {hw_addr} txqueuelen {intf.txqlen}  (Ethernet)\n')
            output_stream.write('\n')

    def print_ethtool(self, intf):
        output_stream = get_active_shell().output_stream
        supported_modes = f'\n{" " * 31}'.join([a for a in intf.supported_modes]) if intf.supported_modes else ''
        output_stream.write(f'Settings for {intf.ifname}:\n')
        output_stream.write(" " * 8 + f'Supported ports: [ {intf.supported_ports if intf.supported_ports else ""} ]\n')
        output_stream.write(" " * 8 + f'Supported link modes:  {supported_modes}\n')
        output_stream.write(" " * 8 + f'Speed: {str(intf.speed) + "Mb/s" if intf.speed else "Unknown"}\n')
        output_stream.write(" " * 8 + f'Duplex: {intf.duplex if intf.duplex else "Unknown"}\n')
        output_stream.write(" " * 8 + f'Auto-negotiation: {"on" if intf.autoneg else "off"}\n')
        output_stream.write(" " * 8 + f'Link detected: {"yes" if intf.operstate == "UP" else "no"}\n')
        output_stream.write(" " * 8 + f'Link Failure Count: {intf.carrier_changes}\n')
        output_stream.write('\n')

    def print_bond(self, intf):
        bond_mode = ["load balancing (round-robin)", "fault-tolerance (active-backup)", "", "",
                     "IEEE 802.3ad Dynamic link aggregation"]
        xmit_policy = strlist("POLICY_layer2 POLICY_layer_3_4 POLICY_layer_2_3 POLICY_encapsulated_layer_2_3 "
                              "POLICY_encapsulated_layer_3_4")
        ad_select = strlist("AD_stable AD_bandwidth AD_count")
        b = intf.bond_properties
        ad_info = b.bond_ad_info
        output_stream = get_active_shell().output_stream

        output_stream.write(f'Settings for {intf.ifname}:\n')
        output_stream.write(" " * 8 + f'Bonding Mode: {bond_mode[b.bond_mode]}\n')
        if b.bond_mode == 1:
            output_stream.write(" " * 8 + f'Currently Active Slave: {b.bond_active_slave}\n')
        if b.bond_mode == 4:
            output_stream.write(" " * 8 + f'Transmit Hash Policy: {xmit_policy[b.bond_xmit_hash_policy]} '
                                          f'({b.bond_xmit_hash_policy})\n')
        output_stream.write(" " * 8 + f'MII Status: {"up" if intf.carrier else "down"}\n')
        output_stream.write(" " * 8 + f'MII Polling Interval (ms): {b.bond_miimon}\n')
        output_stream.write(" " * 8 + f'Up Delay (ms): {b.bond_updelay}\n')
        output_stream.write(" " * 8 + f'Down Delay (ms): {b.bond_downdelay}\n')
        if b.bond_mode == 4:
            output_stream.write('\n')
            output_stream.write(" " * 8 + f'802.3ad info:\n')
            output_stream.write(" " * 15 + f'LACP rate: {"fast" if b.bond_ad_lacp_rate else "slow"}\n')
            output_stream.write(" " * 15 + f'Min links: {b.bond_min_links}\n')
            output_stream.write(" " * 15 + f'Aggregator selection policy (ad_select): {ad_select[b.bond_ad_select]}\n')
            output_stream.write(" " * 15 + f'System priority: {b.bond_ad_actor_sys_prio}\n')
            output_stream.write(" " * 15 + f'System MAC address: {intf.hw_addr}\n\n')
            output_stream.write(" " * 8 + f'Active Aggregator Info:\n')
            output_stream.write(" " * 15 + f'Aggregator ID: {ad_info.bond_ad_info_aggregator}\n')
            output_stream.write(" " * 15 + f'Number of ports: {ad_info.bond_ad_info_num_ports}\n')
            output_stream.write(" " * 15 + f'Actor Key: {ad_info.bond_ad_info_actor_key}\n')
            output_stream.write(" " * 15 + f'Partner Key: {ad_info.bond_ad_info_partner_key}\n')
            output_stream.write(" " * 15 + f'Partner Mac Address: {ad_info.bond_ad_info_partner_mac}\n')

    def print_bonded(self, intf):
        output_stream = get_active_shell().output_stream
        output_stream.write('\n')
        output_stream.write(" " * 8 + f'Slave Interface: {intf.ifname}\n')
        output_stream.write(" " * 15 + f'MII Status: {"up" if intf.carrier else "down"}\n')
        output_stream.write(" " * 15 + f'Speed: {str(intf.speed) + "Mb/s" if intf.speed else "Unknown"}\n')
        output_stream.write(" " * 15 + f'Duplex: {intf.duplex if intf.duplex else "Unknown"}\n')
        output_stream.write(" " * 15 + f'Link detected: {"yes" if intf.operstate == "UP" else "no"}\n')
        output_stream.write(" " * 15 + f'Link Failure Count: {intf.carrier_changes}\n')
        output_stream.write(
            " " *
            15 +
            f'Permanent HW addr: {intf.phys_port_id if intf.phys_port_id else intf.hw_addr}\n')

    def _select_interface(self, dde, interfaces):
        return next((i for i in interfaces if i.dde_instance_id == dde.dde_instance_id), None)

    @ig_command(
        description="Ifconfig command",
        types=(DDEApp, DdeNetworkVlanInterface)
    )
    def ifconfig(self, _api, dde, devname=[]):
        interface = self._select_interface(dde, devname)
        uri = f'../dde/{dde.dde_instance_id}/{INTERFACE_DETAILS_URI}{f"?name={interface.name}" if interface else ""}'
        res = _api.get(uri)
        interfaces = [intf for intf in res if not intf.get('backend_port')]
        for intf in interfaces:
            self.print_ifconfig(intf)

    @ig_command(
        description="Ethtool command",
        types=(DDEApp, DDEPort)
    )
    def ethtool(self, _api, dde, devname=[]):
        device = self._select_interface(dde, devname)
        uri = f'../dde/{dde.dde_instance_id}/{INTERFACE_DETAILS_URI}{f"?name={device.name}" if device else ""}'
        res = _api.get(uri)
        interfaces = [intf for intf in res if intf.get('kind') == 'port']
        if len(interfaces) == 0:
            raise BadArguments(f"{device.name} is not a physical interface")
        for intf in interfaces:
            self.print_ethtool(intf)

    @ig_command(
        description="Bonds command",
        types=(DDEApp, DDEBondPort)
    )
    def bonds(self, _api, dde, devname=[]):
        device = self._select_interface(dde, devname)
        uri = f'../dde/{dde.dde_instance_id}/{INTERFACE_DETAILS_URI}'
        res = _api.get(uri)
        interfaces = [intf for intf in res if intf.get('kind') == 'bond' and
                      (device is None or device.name == intf.get('ifname'))]
        if len(interfaces) == 0:
            if device:
                raise BadArguments(f"{device.name} is not a bond interface")
            else:
                get_active_shell().output_stream.write('No bonds found\n')
                return
        for intf in interfaces:
            self.print_bond(intf)
            for bonded_name in intf.bond_properties.bond_ports:
                bonded_port = next((i for i in res if i.get('ifname') == bonded_name), None)
                if bonded_port is not None:
                    self.print_bonded(bonded_port)
            get_active_shell().output_stream.write('\n')

    @ig_command(
        description="Resolv command",
        types=(DDEApp,)
    )
    def resolv(self, _api, dde):
        output_stream = get_active_shell().output_stream
        uri = f'../dde/{dde.dde_instance_id}/{self.COMMAND_URI.format("resolv")}'
        res = _api.get(uri)
        output_stream.write(f'search {" ".join([a for a in res.get("search_domain")])}\n')
        for nameserver in res.get('nameservers'):
            output_stream.write(f'nameserver {nameserver}\n')


class IperfServerPort(Range):
    MIN = 1025
    MAX = 65535

    def get_description(self):
        return "Unprivileged port number"


class IperfBindInterface(Str):
    def get_description(self):
        return 'IP or hostname to bind to. See iperf man pages for more details.'


@commandgroup(prefix="iperf_server", parent=NetworkDiagnosticsCommands)
class IperfServerCommands:
    IPERFSERVERURL = 'api/network/tools/iperf-server/'

    @ig_command(
        description="Start iperf server on DDE",
        types=(DDEApp, IperfVer, IperfServerPort, IperfBindInterface)
    )
    def start(self, _api, dde, version=3, port=None, bind=None):
        data = {}
        if port is not None:
            data['port'] = port
        if bind is not None:
            data['bind'] = bind

        _api.post(
            get_dde_url(dde.dde_instance_id,
                        self.IPERFSERVERURL + f'v{version}/start'),
            data=data)

    @ig_command(
        description="Stop iperf server running on DDE",
        types=(DDEApp, IperfVer)
    )
    def stop(self, _api, dde, version=3):
        _api.post(
            get_dde_url(dde.dde_instance_id,
                        self.IPERFSERVERURL + f'v{version}/stop'))

    @ig_command(
        description="Display iperf server status",
        types=(DDEApp, IperfVer)
    )
    @render(
        detailed_individually=True,
        individually=True)
    def status(self, _api, dde, version=None):
        suffix = f'v{version}/status' if version is not None else 'status'
        return _api.get(get_dde_url(dde.dde_instance_id, self.IPERFSERVERURL + suffix))
