from collections import OrderedDict
from infinishell.infishell.types import IPAddress, Hostname, Str
from .lib.dde import get_dde_url
from .nettools import NetworkCommands
from ..types import Entity
from . import ig_command, DDEApp
from infinishell.infishell import commandgroup, BadArguments
from .. import EntityCommands, render, roles

QUERY_FIELDS = OrderedDict(
    ddeapp='DDE App',
    default_gateway='System Default Gateway',
    dns_servers='DNS Servers List',
    search_domain='DNS Suffix Search List',
    hostname='System Hostname'
)


class Host(Entity):
    ENTITY_NAME = 'Network host'
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    UNIQUE_NAME = True
    SORTABLE = False
    URI = 'api/network/host'
    DDE_INSTANCE_ID = None

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)


@commandgroup(prefix="system", parent=NetworkCommands)
class NetworkSystem(EntityCommands):
    ENTITY_TYPE = Host

    @ig_command(
        description="Display Default Gateway and DNS configuration",
        types=(DDEApp,),
    )
    @render(
        individually=True,
        )
    def query(self, _api, dde):
        uri = f'../dde/{dde.dde_instance_id}/api/network/host'
        response = _api.get(uri)
        primary = secondary = tertiary = ''
        try:
            primary = response.dns_servers[0]
            secondary = response.dns_servers[1]
            tertiary = response.dns_servers[2]
        except IndexError:
            pass
        result = OrderedDict((("DDE App:", dde.dde_instance_name.upper()),
                             ("System Default Gateway:", response.default_gateway),
                              ("Primary DNS:", primary),
                              ("Secondary DNS:", secondary),
                              ("Tertiary DNS:", tertiary),
                              ("DNS Suffix Search List:", response.search_domain),
                              ("System Hostname:", response.hostname)))

        return result

    @ig_command(
        description="Set DNS Servers and DNS search suffixes list",
        types=(DDEApp, IPAddress.List, Str.List),
    )
    @render(
        individually=True,
        )
    @roles.INFINIDAT.ADMIN
    def set_dns(self, _api, dde, dns=None, dns_suffix_list=None):
        uri = f'../dde/{dde.dde_instance_id}/api/network/host'
        if not dns and not dns_suffix_list:
            raise BadArguments('dns or dns_suffix_list argument must be set.')
        data = dict()
        if dns:
            data.update({'dns_servers': [x.ip for x in dns]})
        if dns_suffix_list:
            data.update({'search_domain': dns_suffix_list})
        _api.put(uri, data=data)

    @ig_command(
        description="Set System hostname",
        types=(DDEApp, Hostname),
    )
    @render(
        individually=True,
        )
    @roles.INFINIDAT
    def set_hostname(self, _api, dde, hostname):
        uri = f'../dde/{dde.dde_instance_id}/api/network/host'
        _api.put(uri, data={'hostname': hostname})

