from ...types import Entity
from infinishell.infishell import config
from infinishell.infishell.types import ArgType, DynamicValue
from ....system import system
from ....exceptions import NoSuchEntity
from infinishell.infishell.commands.commands import metadata
from functools import partial


def get_context(context, param_name=None, argtype=None, raw=True):
    if not context:
        return None
    arg = None
    if param_name:
        arg = context.get(param_name, None)
    elif argtype:
        arg = next((a for a in context.values()
                    if (isinstance(a["type"], argtype) or a["type"] == argtype)),
                   None)
    if not arg:
        return None
    if raw:
        return arg["raw"]
    cooked = arg["type"].get_actual(arg["name"], arg["raw"])
    if isinstance(cooked, DynamicValue):
        try:
            _, cooked = cooked.get(system.api, arg["name"])
        except NoSuchEntity:
            return None
    return cooked


def filtered(func=None, filter_params=None):
    """ Command method decorator for defining commands with context-filtered parameters. """

    def context_filter(params, context):
        return filter_params(params, partial(get_context, context=context))

    return metadata(func, param_filter=context_filter)


class ContextMixin(ArgType):
    def _get_context(self, param_name=None, argtype=None, raw=True):
        return get_context(self._context_args, param_name, argtype, raw)


class ContextSensitiveEntity(Entity, ContextMixin):
    COMPLETION_CACHE = False

    def _context_hook(self, results):
        """ Subclasses should implement this method, which is invoked during
        ``get_possible_values()``. The method should filter results according to
        command context variables present. It may use ``_get_context()`` helper function """
        return results

    def _get_possible_values_by_search(self, api, starts_with, type_cls):
        starts_with = starts_with or ""

        def wrapper_hook(results):
            results = self._search_hook(results)
            results = self._context_hook(results)
            return results

        query = api.paged_query(
            "search",
            page_size=config.api_max_page_size,
            params=dict(
                type=type_cls.get_entity_mgmt_name(),
                query=starts_with,
            ),
            hook=wrapper_hook,
        )
        return (entity_info.name for entity_info in query if entity_info and entity_info.name.startswith(starts_with))

    def _get_possible_values_by_query(self, api, starts_with, type_cls):
        type_cls = type_cls or type(self)
        params = {}
        # Create pluck for entity's name field, and any other fields possibly necessary for invoking the default query hook
        pluck = None if not type_cls.PLUCKABLE else api.create_pluck(type_cls.NAME_FIELD,
                                                                     *(type_cls.HOOK_DEPENDENCIES or ()))
        # Create generator - note that this has to be a generator because params may be updated below
        generator = (e for e in type_cls.iter_all(api, params=params, pluck=pluck, page_size=config.api_max_page_size,
                                                  hook=self._context_hook))
        if starts_with:
            if type_cls.FILTERABLE:
                params.update(api.create_filter().like(type_cls.NAME_FIELD, starts_with + "%").to_params())
            else:
                generator = (e for e in generator if e[type_cls.NAME_FIELD].startswith(starts_with))
        # Return only names, not whole entities
        return (e[type_cls.NAME_FIELD] for e in generator)
