from ..types import ComponentNestedEntity
from . import DDENode, ig_command
from .nodes import DdeNodeCommands
from infinishell.infishell import commandgroup
from infinishell.infishell.utils import strlist
from .. import EntityCommands, render
from ..options import QueryOptions
from ...exceptions import ManagementException, NoSuchEntity
from functools import partial
import re


class DDEPort(ComponentNestedEntity):
    CONTAINER_TYPE = DDENode
    CONTAINER_SUBCOMPONENTS_FIELD = "dde_ports"
    SUBCOMPONENT_CONTAINER_ID_FIELD = "dde_node_mgmt_id"
    NAME_PATTERN = re.compile(r"^DDE(\d)P(\d+)$", re.I)
    ENTITY_NAME_FORMATTER = "DDE{}P{:02}"
    ENTITY_NAME_PART_FIELDS = "node_id", "port_num"

    @classmethod
    def get_description(cls):
        return "Name of an existing DDE port"

    @classmethod
    def get_all(cls, api, *args, **kwargs):
        pluck = api.create_pluck(cls.CONTAINER_TYPE.ID_FIELD, cls.CONTAINER_TYPE.NAME_FIELD)
        # Components do not support filtering, so filter on client side
        container_id = kwargs.get("container_id")
        containers = cls.CONTAINER_TYPE.iter_all(api, pluck=pluck)
        if container_id is not None:
            containers = (c for c in containers if c[cls.CONTAINER_TYPE.ID_FIELD] == container_id)
        entities = [cls._entity_hook(cls._attach_container(entity, container))
                    for container in containers
                    for entity in api.get(
                        cls.CONTAINER_TYPE.get_uri(container[cls.CONTAINER_TYPE.ID_FIELD],
                                                   cls.CONTAINER_SUBCOMPONENTS_FIELD))
                    if cls._entity_predicate(entity)]
        if kwargs.get("hook"):
            entities = kwargs["hook"](entities)
        return entities

    @classmethod
    def _attach_container(cls, entity, container):
        entity[cls.CONTAINER_TYPE.NAME_FIELD] = container[cls.CONTAINER_TYPE.NAME_FIELD]
        return entity

    @classmethod
    def _get_entity_by_ids(cls, api, container_id, entity_id):
        entities = []
        for node in cls.CONTAINER_TYPE.iter_all(api):
            if node[cls.CONTAINER_TYPE.NAME_FIELD] != container_id:
                continue
            ents = api.get(cls.CONTAINER_TYPE.get_uri(node.id, cls.CONTAINER_SUBCOMPONENTS_FIELD))
            try:
                entity = [entity for entity in ents if entity.port_num == entity_id][0]
            except Exception:
                entity = None
                raise ManagementException(404, None)
            if entity:
                entity = cls._entity_hook(cls._attach_container(entity, node))
                cls._validate_fetched_entity(entity)
                entities.append(entity)
        return entities

    @classmethod
    def get_by_name(cls, api, name):
        match = cls.NAME_PATTERN.match(name)
        if not match:
            raise NoSuchEntity('"{}" is not a valid {} name'.format(name, cls.get_entity_name()))
        ids = (int(_id) for _id in match.groups())
        try:
            return cls._get_entity_by_ids(api, *ids)
        except ManagementException as e:
            if e.http_code != 404:
                raise
            raise NoSuchEntity('No such {}: "{}"'.format(cls.get_entity_name(), name))


QUERY_FIELDS = strlist("name node_id port_num media_type type state link_state hw_addr")
DETAILED_FIELDS = strlist(
    "id name node_id port_num media_type type state link_state conn_port_comp_id hw_addr connection_speed hw_addr")
SUBSTITUTE = dict(
    node_id='dde_node_id',
    port_num='port_number',
    id='internal_id',
)


@commandgroup(prefix="port", parent=DdeNodeCommands)
class DdePortCommands(EntityCommands):
    ENTITY_TYPE = DDEPort

    @ig_command(
        description="Display DDE ports",
        types=(DDEPort, DDENode,),
        options=QueryOptions(fields=DETAILED_FIELDS,
                             substitute=SUBSTITUTE, detailed=True)
    )
    @render(
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_FIELDS,
        substitute=SUBSTITUTE,
        detailed_individually=True
    )
    def query(self, _api, port=None, dde=None):
        return self._query(_api, port, dde_node_mgmt_id=dde, _hook=partial(self._query_hook))

    def _query_hook(self, ports):
        return ports
