from . import DDEActiveApp, ig_command
from .infinisafe import InfinisafeCommands, infinisafecommandgroup
from infinishell.infishell.exceptions import BadArguments, BadValue
from infinishell.infishell.types import Range, MappedChoiceType, ArgType,\
    Measure
from .lib.tasksupport import report_started_task
from ..options import OutputFormat
from .. import roles
from ...renderer import is_csv_mode, is_json_mode
from collections import OrderedDict
from munch import Munch
import datetime
import re
from infinishell.infinibox.commands.infiniguard.lib.dde import AnyDdeApi
from ..system import system
from infinishell.infinibox.commands.infiniguard.lib.renderer import renderfields, SuspendDeleteStateRenderer
from infinishell.infinibox.renderer.values import MillisecondsTimestampRenderer
import time


def tzparams():
    tzoffset = -(datetime.datetime.now().astimezone().tzinfo.utcoffset(None).total_seconds() // 60)
    tzshift = int(tzoffset / 60)
    tzfrac = 0
    if tzshift != tzoffset / 60:
        offset_frac = abs(tzoffset - tzshift * 60)
        tzfrac = 60 - offset_frac if tzoffset > 0 else offset_frac
        if tzoffset > 0:
            tzshift += 1
    return tzshift, tzfrac


POLICY_URI = "cr/policies/{}/{}"
POLICY_FIELDS = OrderedDict(
    ddeapp='DDE App ID',
    status='Status',
    immutability='Lock State',
    schedule='Schedule',
    retention='Retention',
    suspend_delete='Snapshot Deletion',
    suspend_expires_at='Snapshots deletion is suspended until')


class PolicyStatus(MappedChoiceType):
    def _get_mapping(self):
        return [("Enabled", True), ("Disabled", False)]


class Retention(Range):
    MIN = 1

    def get_description(self):
        _min, _max = self._get_min(), self._get_max()
        return "Snapshot retention, in days, between {} and {}".format(_min, _max)

    def _get_max(self):
        self.MAX = system.api.get('integrations/config/infinisafe/infinisafe.max_snapshots_retention_days')
        return self.MAX

    def verify_raw(self, arg, value):
        try:
            return super().verify_raw(arg, value)
        except BadValue:
            raise BadValue(f'Retention must be a number between {self.MIN} and {self._get_max()}')


class Schedule(ArgType):
    AUTO_COMPLETION_CASE_INSENSITIVE = True
    INVALID_MESSAGE = "is not valid. Expected [list of] whole hours in 24h format hh[:00][,...] or " \
                      "\"Every hour\", got"
    NONROUND_HOUR = "is not a round hour. Expected whole hour(s) 0:00 to 23:00 or 0 to 23, got"
    NOT_AN_HOUR = "is not valid. Use 24-hour format, 00:00 to 23:00 or 0 to 23, " \
                  "separating multiple values by commas. Got"
    NOT_IN_RANGE = "is out of range. Only whole hours 00:00 to 23:00 or 0 to 23 are allowed, got"

    _pattern = re.compile(r"^\s*(every hour|\d{1,2}(:00)?)\s*$", re.I)
    _every = re.compile(r"^\s*every hour\s*$", re.I)
    _non_round = re.compile(r"^\s*\d{1,2}:\d{1,2}\s*$", re.I)

    def __init__(self):
        super(Schedule, self).__init__()
        self.tzshift, self.tzfrac = tzparams()

    def _verify_raw(self, value):
        if not self._pattern.match(value):
            if self._non_round.match(value):
                return self.NONROUND_HOUR
            else:
                return self.NOT_AN_HOUR

    def _get_actual(self, value):
        try:
            if self._every.match(value):
                return True, list(range(24))
            hour = int(value.split(":")[0])
            return True, hour
        except (ValueError, TypeError):
            return False, self.INVALID_MESSAGE

    def _verify(self, value):
        if isinstance(value, list):
            return None
        elif value < 0 or value > 23:
            return self.NOT_IN_RANGE

    @classmethod
    def get_name(cls):
        return "Schedule slot"

    def get_description(self):
        return "A time slot to schedule a daily snapshot 00:00 to 23:00 (hr or hh:00 or \"Every Hour\")"

    def _get_values(self):
        return ["Every hour"] + ["{:02d}:{:02d}".format(i, self.tzfrac) for i in range(24)]

    def get_possible_values(self, _api, starts_with=None):
        return ["Every hour"] + ["{:02d}:{:02d}".format(i, self.tzfrac) for i in range(24)]

    def get_colliding_values(self, api, existing_values, possible_values, current_val):
        if existing_values:
            result = ["{:02d}:{:02d}".format(self._get_actual(v)[1], self.tzfrac) for v in existing_values if
                      not self._every.match(v)]
            return result + ["Every hour"]
        return []


@infinisafecommandgroup(prefix="policy", parent=InfinisafeCommands)
class SnapshotPolicyCommands:
    @ig_command(
        description="Display the InfiniSafe policy",
        types=(DDEActiveApp,),
        options=(OutputFormat(),)
    )
    @renderfields(
        value_renderers=dict(suspend_delete=SuspendDeleteStateRenderer(),
                             suspend_expires_at=MillisecondsTimestampRenderer()),
        individually=True,
        fields=POLICY_FIELDS.keys(),
        substitute=POLICY_FIELDS
    )
    def query(self, _api, dde):
        policy = AnyDdeApi(_api).get(POLICY_URI.format(dde.dde_instance_name, 'system'))
        suspend_policy = AnyDdeApi(_api).get(POLICY_URI.format(dde.dde_instance_name, 'suspend_delete'))
        tzshift, tzfrac = tzparams()
        if len(policy.daily) < 24 or is_json_mode() or is_csv_mode():
            schedule = ["{:02d}:{:02d}".format(idx, round(tzfrac)) for idx in range(24)
                        if idx in [(24 + hr - tzshift) % 24 for hr in policy.daily]]
        else:
            schedule = "Every hour"

        res = Munch(
            ddeapp=dde.dde_instance_name.upper(),
            status='Enabled' if policy.enabled else 'Disabled',
            immutability='Locked' if policy.immutable else 'Unlocked',
            schedule=schedule,
            retention=f'{policy.retention} days',
        )
        if suspend_policy.suspend_delete:
            res.suspend_expires_at = suspend_policy.expires_at
        else:
            res.suspend_delete = suspend_policy.suspend_delete
        return [res]

    @ig_command(
        description="Modify InfiniSafe policy.\nOnly infinidat user can run this command.",
        types=(DDEActiveApp, PolicyStatus, Schedule.List, Retention)
    )
    @roles.INFINIDAT
    def modify(self, _api, dde, status=None, schedule=None, retention=None):
        if not schedule:
            daily = None
        else:
            if len(schedule) == 1 and isinstance(schedule[0], list):
                schedule = schedule[0]
            elif any(isinstance(slot, list) for slot in schedule):
                raise BadArguments('Individual time slots specified in "Every hour" policy')

            tzshift, _ = tzparams()
            daily = sorted((24 + hr + tzshift) % 24 for hr in set(schedule))

        data = {prop: value for prop, value in Munch(enabled=status, retention=retention, daily=daily).items()
                if value is not None}
        if len(data) == 0:
            raise BadArguments("No changes specified. Policy not modified.")

        policy_uri = POLICY_URI.format(dde.dde_instance_name, 'system')
        current_policy = AnyDdeApi(_api).get(policy_uri)
        for k, v in data.items():
            if current_policy.get(k) != v:
                break
        else:
            return 'Same as the current settings. Nothing to change here.'

        AnyDdeApi(_api).put(policy_uri, data=data)
        return "System policy was modified successfully."


SUSPEND_DELETE_FIELDS = OrderedDict(
    suspend_delete='Snapshots Deletion',
    expires_at='Snapshots deletion is suspended until'
)


SUSPEND_DELETE_DATE_UNITS = OrderedDict((unit, cap)
                                        for cap, units in (
                                            ('hours', ('H', 'HOUR', 'HOURS')),
                                            ('days', ('D', 'DAY', 'DAYS')))
                                        for unit in units)


class SuspendDeleteExpiration(Measure):  # DateDuration):
    MIN = '1h'
    MAX = '72h'

    def __init__(self):
        super().__init__()
        self._pattern = re.compile('([1-9][0-9]*)(H|HOUR|HOURS|D|DAY|DAYS)?$', re.IGNORECASE)
        self._now = time.time()

    def _process_actual(self, size: int, unit: str) -> int:
        if not unit:
            return None

        unit = SUSPEND_DELETE_DATE_UNITS[unit.upper()]

        delta = {'hours': 60 * 60, 'days': 24 * 60 * 60}[unit] * size
        ms = int((self._now + delta) * 1000)
        return ms

    def _verify(self, value):
        error = super()._verify(value)
        if error:
            return error
        if value < self.get_actual('min', self.MIN):
            return f'Time duration cannot be less then {self.MIN}'
        if value > self.get_actual('max', self.MAX):
            return f'Time duration cannot be greater then {self.MAX}'

    def _get_units(self):
        return SUSPEND_DELETE_DATE_UNITS.keys()

    def get_description(self):
        return "Time durations between 1 and 72 hours in the form of a positive number," \
               " followed by a unit; e.g. 1H, 10DAYS"

    def _get_error(self):
        return "is not a valid time duration"


@infinisafecommandgroup(prefix="deletion", parent=InfinisafeCommands)
class SuspendDeleteCommands:
    @ig_command(
        description="Display snapshot deletion status.",
        types=(DDEActiveApp,)
    )
    @renderfields(
        value_renderers=dict(suspend_delete=SuspendDeleteStateRenderer(),
                             expires_at=MillisecondsTimestampRenderer()),
        fields=SUSPEND_DELETE_FIELDS.keys(),
        substitute=SUSPEND_DELETE_FIELDS,
        individually=True)
    def query(self, _api, dde):
        suspend_delete_schema = AnyDdeApi(_api).get(POLICY_URI.format(dde.dde_instance_name, 'suspend_delete'))
        return dict(expires_at=suspend_delete_schema.expires_at) if suspend_delete_schema.suspend_delete \
            else dict(suspend_delete=suspend_delete_schema.suspend_delete)

    @ig_command(
        description="Suspend automatic snapshot deletion until a specific time.",
        types=(DDEActiveApp, SuspendDeleteExpiration),
    )
    @roles.INFINIDAT.ADMIN
    def suspend(self, _api, dde, duration):
        policy_uri = POLICY_URI.format(dde.dde_instance_name, 'suspend_delete')
        ddeid, task = AnyDdeApi(_api).putex(policy_uri, data={"expires_at": duration})
        report_started_task(ddeid, task, "Extension of the snapshot\'s expiration time will continue in background")
        success = f"Infinisafe snapshot deletions are suspended on DDE App " \
                  f"{dde.dde_instance_name.upper()}. " \
                  "Frequently monitor system capacity to ensure enough space."
        return success
