from contextlib import contextmanager
from functools import partial
from itertools import chain
from infinishell.infinibox.commands import EntityCommands, roles
from infinishell.infinibox.commands.infiniguard import ig_command, DDEActiveApp
from infinishell.infinibox.commands.options import QueryOptions, AutoApprove
from infinishell.infishell import commandgroup, runtime
from infinishell.infishell.types import EnumType, Str, Bool
from infinishell.infishell.utils import strlist, yesno
from infinishell.infinibox.commands.utils import is_option_set
from .lib.context_entity import ContextSensitiveEntity, filtered
from .lib.dde import allddeget, get_dde_url, ddeget
from infinishell.infishell.shell import get_active_shell
from infinishell.infishell.exceptions import CommandAborted
from infinishell.infinibox.exceptions import ManagementException
from .nodes import DdeAppCommands
from .. import render


QUERY_FIELDS = strlist("name description protocol dedup")
DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist(
    "access_hosts access_users export_path allow_links hidden permissions anongid anonuid commit squash"
)


SUBSTITUTE_FIELDS = dict(
    dedup="deduplication", access_hosts="hosts", access_users="users"
)

NOT_AVAILABLE = "<N/A>"


@contextmanager
def patch_dde_entity(cls, dde: DDEActiveApp):
    original = cls.ENTITY_TYPE.DDE_INSTANCE_ID
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = dde.dde_instance_id
    yield
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = original


NAS_URLS = {
    "nfs": "api/nas/nfs/shares",
    "cifs": "api/nas/cifs/shares",
    "app_specific": "api/nas/app_specific/shares",
}


class ProtocolType(EnumType):
    def _get_values(self):
        return NAS_URLS.keys()


class PermissionType(EnumType):
    def _get_values(self):
        return "ro", "rw"


class SquashType(EnumType):
    def _get_values(self):
        return "root", "none"


class CommonShare(ContextSensitiveEntity):
    ENTITY_NAME = "Share"
    DESCRIPTION = "Network Share"
    ALIAS = "common"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = False
    NAME_FIELD = "name"
    URI = "api/nas/shares"
    MULTI_FIELD_PUT = True
    FILTERABLE = False
    DDE_INSTANCE_ID = None

    def get_description(self):
        name = self.ENTITY_NAME
        return (
            name
            if self._get_context(argtype=DDEActiveApp, raw=False)
            else f"{name}: please specify DDE first"
        )

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    def _context_hook(self, results, **kwargs):
        dde = self._get_context(argtype=DDEActiveApp, raw=False)
        proto = self._get_context(argtype=ProtocolType, raw=False)

        if not dde:
            return []

        return [
            share
            for share in results
            if share.dde_instance_id == dde.dde_instance_id
            and not proto
            or share.protocol == proto
        ]

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        all_shares = list()
        # Iterate all nas types if common, otherwise just single type
        entity_dict = NAS_URLS if cls.ALIAS == "common" else {cls.ALIAS: cls.URI}
        for uri_protocol, uri in entity_dict.items():
            res = allddeget(api, uri, ddes=(1, 2))
            for dde_id, shares in res.items():
                if shares:
                    for share in shares:
                        share.protocol = uri_protocol
                        share.dde_instance_id = dde_id
                    all_shares += shares

        if kwargs.get("hook"):
            all_shares = kwargs["hook"](all_shares)

        return all_shares


class AccessHost(ContextSensitiveEntity):
    ENTITY_NAME = "Host"
    DESCRIPTION = "Access Host"
    PAGED = False
    PLUCKABLE = True
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = False
    NAME_FIELD = "name"
    URI = "api/nfs/shares"
    MULTI_FIELD_PUT = True
    FILTERABLE = False
    DDE_INSTANCE_ID = None
    HOOK_DEPENDENCIES = ("name",)

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    @classmethod
    def _get_host_url(cls, name):
        return f"api/nas/nfs/shares/{name}/share_hosts"

    @classmethod
    def _fetch_entity_by_name(cls, api, name):
        uri = cls._get_host_url(name)
        res = allddeget(api, uri, ddes=[cls.DDE_INSTANCE_ID])
        return res.get(cls.DDE_INSTANCE_ID)


class AccessUser(ContextSensitiveEntity):
    ENTITY_NAME = "User"
    DESCRIPTION = "Access User"
    PAGED = False
    PLUCKABLE = True
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = False
    NAME_FIELD = "name"
    URI = "api/nfs/shares"
    MULTI_FIELD_PUT = True
    FILTERABLE = False
    DDE_INSTANCE_ID = None
    HOOK_DEPENDENCIES = ("name",)

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    @classmethod
    def _get_host_url(cls, name):
        return f"api/nas/cifs/shares/{name}/share_users"

    @classmethod
    def _fetch_entity_by_name(cls, api, name):
        uri = cls._get_host_url(name)
        res = allddeget(api, uri, ddes=[cls.DDE_INSTANCE_ID])
        return res.get(cls.DDE_INSTANCE_ID)


class AppSpecificShare(CommonShare):
    ALIAS = "app_specific"
    URI = NAS_URLS.get("app_specific")
    DESCRIPTION = "App-specific share"


class NfsShare(CommonShare):
    ALIAS = "nfs"
    URI = NAS_URLS.get("nfs")
    DESCRIPTION = "NFS share"


class CifsShare(CommonShare):
    ALIAS = "cifs"
    URI = NAS_URLS.get("cifs")
    DESCRIPTION = "CIFS share"


def get_access_users(_api, dde, item):
    try:
        access_users = ddeget(
            _api,
            dde.dde_instance_id,
            path=AccessUser._get_host_url(item.name),
        )
    except ManagementException as e:
        if 'E1000011' or 'CIFS_SERVER_IS_DISABLED' in str(e):  # CIFS/SMB server is disabled! (E1000011)
            access_users = None
        else:
            raise
    return (
        [f"{user.user}:{user.permissions}" for user in access_users]
        if access_users
        else "All"
    )


def get_access_hosts(_api, dde, item):
    access_hosts = ddeget(
        _api,
        dde.dde_instance_id,
        path=AccessHost._get_host_url(item.name),
    )
    return (
        [f"{host.host}:{host.permissions}" for host in access_hosts]
        if access_hosts
        else "All"
    )

def _filter_anon_params(params, context):
    proto = context(param_name='proto', raw=False)
    # Fields to hide for protocol
    hide_map = {
        "cifs": ["anonuid", "anongid", "squash"],
        "nfs": ["hidden"],
        "app_specific": ["anonuid", "anongid", "squash", "hidden", "permissions"]
    }
    return [arg for arg in params if arg.name not in hide_map.get(proto, [])]


@commandgroup(prefix="share", parent=DdeAppCommands)
class ShareCommands(EntityCommands):
    ENTITY_TYPE = CommonShare

    @staticmethod
    def _filter_by_name(dde, shares):
        return shares and next((share for share in shares), None)

    def _query_hook(self, shares):
        # Inject all missing fields so it shows in output
        for share in shares:
            for attr_name in DETAILED_QUERY_FIELDS:
                if attr_name not in share:
                    share[attr_name] = NOT_AVAILABLE
            yield share

    @ig_command(
        description="List shares",
        types=(DDEActiveApp, ProtocolType, CommonShare),
        options=QueryOptions(
            fields=QUERY_FIELDS, substitute=SUBSTITUTE_FIELDS, detailed=True
        ),
    )
    @render(
        fields=QUERY_FIELDS,
        substitute=SUBSTITUTE_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        detailed_individually=True,
    )
    def query(self, _api, dde, proto=None, name=None):
        share = self._filter_by_name(dde, name)
        with patch_dde_entity(self.__class__, dde):
            # Filter just relevant protocols if proto is specified
            # Otherwise all protocols
            entities = self._query(
                _api, _entity=[[CifsShare, NfsShare, AppSpecificShare]], ALIAS=proto
            )
            res = list(
                chain(
                    *[
                        self._query(
                            _api,
                            _hook=partial(self._query_hook),
                            name=share and share.name,
                            _entity_type=entity,
                            dde_instance_id=dde.dde_instance_id,
                        )
                        for entity in entities
                    ]
                )
            )

            if runtime.options.detailed:
                for item in res:
                    item.access_hosts = item.access_users = NOT_AVAILABLE
                    if item.protocol in ["cifs", "app_specific"]:
                        item.access_users = get_access_users(_api, dde, item)
                    if item.protocol in ["nfs", "app_specific"]:
                        item.access_hosts = get_access_hosts(_api, dde, item)

            return res

    @ig_command(
        description="Create a network share",
        types=(
            DDEActiveApp,  # dde
            ProtocolType,  # proto
            Str,  # name
            Bool,  # deduplication
            Str,  # description
            Bool,  # hidden
            PermissionType,  # permissions
            Str,  # anongid
            Str,  # anonuid
            SquashType,  # squash
        ),
    )
    @filtered(
        filter_params=_filter_anon_params
    )
    @roles.INFINIDAT.ADMIN
    def create(
        self,
        _api,
        dde,
        proto,
        name,
        deduplication=True,
        description=None,
        hidden=None,
        permissions=None,
        anongid=None,
        anonuid=None,
        squash=None,
    ):

        with patch_dde_entity(self.__class__, dde):
            return self._create(
                _api,
                name=name,
                protocol=proto,
                dedup=deduplication,
                description=description,
                hidden=hidden,
                permissions=permissions,
                anongid=anongid,
                anonuid=anonuid,
                squash=squash,
            )

    @ig_command(
        description="Modify a network share",
        types=(
            DDEActiveApp,  # dde
            CommonShare,  # name
            Str,  # description
            Bool,  # hidden
            PermissionType,  # permissions
            Str,  # anongid
            Str,  # anonuid
            SquashType,  # squash
        ),
    )
    @roles.INFINIDAT.ADMIN
    def modify(
        self,
        _api,
        dde,
        name,
        description=None,
        hidden=None,
        permissions=None,
        anongid=None,
        anonuid=None,
        squash=None,
    ):
        share = self._filter_by_name(dde, name)
        with patch_dde_entity(self.__class__, dde):
            restart = None
            if permissions != None:
                restart=True
            return self._update(
                _api,
                share,
                description=description,
                hidden=hidden,
                permissions=permissions,
                anongid=anongid,
                anonuid=anonuid,
                squash=squash,
                restart=restart,
            )

    @ig_command(description="Delete a share", types=(DDEActiveApp, CommonShare))
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, dde, name):
        share = self._filter_by_name(dde, name)
        with patch_dde_entity(self.__class__, dde):
            return self._delete(_api, share)
