from __future__ import annotations  # for Union pipe '|'
from infinishell.infishell import (
    commandgroup,
    render
)
from . import ig_command
from collections import OrderedDict
from munch import Munch
from infinishell.infinibox.commands.system import SystemCommands
from typing import List, Tuple, Dict, Optional
import logging
from itertools import chain
from capacity import Capacity, byte, TB
from .lib.dde import allddeget, AnyDdeApi
from .. import roles
from ..options import Grep, InvertedGrep
from . import DDEActiveApp
from .lib.tasksupport import report_started_task
from infinishell.infishell.exceptions import CommandFailed
from ..system import system as current_system

log = logging.getLogger(__name__)


INFO_FIELDS = OrderedDict(
    system_name='Name',
    model='Model',
    serial='Serial Number',
    iguard_version='Infiniguard Version',
    flash_cache='Available Flash Cache',
    ddenode1_app='DDE node 1 running',
    ddenode2_app='DDE node 2 running',
    ddenode3_app='DDE node 3 running',
    infinisafe_status='Infinisafe Status',
    total_capacity='Total Capacity',
    ddeappa_capacity='DDE A Capacity',
    ddeappb_capacity='DDE B Capacity',
    infinisafe_capacity='Infinisafe Capacity',
    disk_encryption='Disk Encryption',
    ssd_cache_drives='SSD Cache Drives',
    bbu_charge='BBU Charge Percent',
    bbu_protection='Node BBU Protection',
    inactive_drives='Inactive Drives',
    rebuild1='Rebuild 1 In Progress',
    rebuild2='Rebuild 2 In Progress',
    tasks='Tasks In Progress')


class AlreadySwappedError(CommandFailed):
    pass


@commandgroup(prefix="infiniguard")
class InfiniguardCommands:

    def format_flash_cache(self, _api) -> str:
        system_stats = _api.get("system/stats").system_stats
        cache = sum(node.ssd_cache_stats.total_section_capacity if node.ssd_cache_stats else 0
                    for node in system_stats)
        tb = (byte * cache * 68 * 1024) / TB
        return f'{round(tb, 2)} TB'

    def format_tasks(self, _api) -> List[str]:
        tasks = allddeget(_api, 'mgmt/tasks')
        task_descriptions = {task.description
                             for task in chain(*[tlist for tlist in tasks.values() if tlist])
                             if task.status == 'running'}
        return sorted(task_descriptions)

    def format_standby_description(self, _api, cr_enabled):
        standby_msg = 'Standing by, ready for failover'
        if not cr_enabled:
            return standby_msg
        has_attachment = AnyDdeApi(_api).get('cr/ddesnapshots/attachment').attached
        return "Attached snapshot" if has_attachment \
            else standby_msg

    def get_dde_map(self, _api, infinisafe_enabled) -> Dict[int, str]:
        ddemap = AnyDdeApi(_api).get('mgmt/dde-app-map')
        standby_descr = self.format_standby_description(_api, infinisafe_enabled)
        appnames = {'a': 'DDE App A', 'b': 'DDE App B',
                    'standby': standby_descr}
        return {elt.nodeid: appnames[elt.app] for elt in ddemap}

    def get_capacities(self, _api) -> Tuple[Capacity, Capacity, Capacity]:
        report = AnyDdeApi(_api).get("cr/capacityreport")
        appa = report.ddeapps_capacity.app_a.provisioned_dca
        appb = report.ddeapps_capacity.app_b.provisioned_dca
        infinisafe = report.capacity.provisioned
        return appa, appb, infinisafe

    @ig_command(
        description="Display general system information",
        options=(Grep(), InvertedGrep())
    )
    @render(
        individually=True,
        fields=INFO_FIELDS.keys(),
        substitute=INFO_FIELDS
    )
    def info(self, _api):
        system_info_remap = [
            ('system_name', 'name'),
            ('model', 'model_name'),
            ('serial', 'serial'),
            ('total_capacity', 'total_physical_capacity'),
            ('disk_encryption', 'disk_encryption'),
            ('ssd_cache_drives', 'ssd_cache_drives'),
            ('bbu_charge', 'bbu_charge_percent'),
            ('bbu_protection', 'node_bbu_protection'),
            ('inactive_drives', 'inactive_drives'),
            ('rebuild1', 'rebuild_1_in_progress'),
            ('rebuild2', 'rebuild_2_in_progress')]

        sysinfo = SystemCommands().info(_api)
        result = Munch()
        for newname, oldname in system_info_remap:
            if (f := sysinfo.get(oldname)) is not None:
                result[newname] = f

        infinisafe_enabled = _api.get('integrations/config/infinisafe/infinisafe.enable')
        iguard_version = _api.get('integrations/config/general/system_version')
        ddemap = self.get_dde_map(_api, infinisafe_enabled)

        result.update(Munch(
            iguard_version=iguard_version,
            flash_cache=self.format_flash_cache(_api),
            ddenode1_app=ddemap[1],
            ddenode2_app=ddemap[2],
            ddenode3_app=ddemap[3],
            infinisafe_status='Enabled' if infinisafe_enabled else 'Disabled',
            tasks=self.format_tasks(_api)
        ))
        if infinisafe_enabled:
            capacities = self.get_capacities(_api)
            result.update(Munch(ddeappa_capacity=capacities[0],
                                ddeappb_capacity=capacities[1],
                                infinisafe_capacity=capacities[2]))
        return [result]

    @staticmethod
    def deduce_failover_params(ddemap, ddeapp) -> Tuple[int, Dict, str]:
        maplist = sorted([(elt.app, elt.nodeid) for elt in ddemap])
        nodes = tuple(nodeid for _app, nodeid in maplist)
        operations = {
            ((1, 2, 3), 'a'): (2, (2, 1, 3), 'Failover'),
            ((1, 2, 3), 'b'): (1, (1, 2, 3), 'Failover'),
            ((3, 2, 1), 'a'): (2, (2, 3, 1), 'Failback'),
            ((1, 3, 2), 'b'): (1, (1, 3, 2), 'Failback')
        }
        try:
            controller_ddeid, (controller, faulty, replacement), name = operations[(nodes, ddeapp)]
        except KeyError:
            raise AlreadySwappedError(
                f'Failover of DDE App {ddeapp.upper()} is unavailable under the current DDE App mapping.')

        return (controller_ddeid,
                dict(controlNode=controller, faultyNode=faulty, replacementNode=replacement),
                name)

    @ig_command(
        description="Failover DDE App",
        types=(DDEActiveApp,),
    )
    @roles.INFINIDAT.ADMIN
    def failover(self, _api, dde):
        dde = dde.dde_instance_name
        ddemap = AnyDdeApi(_api).get('mgmt/dde-app-map')
        ddeid, start_params, opname = self.deduce_failover_params(ddemap, dde)
        started_task = _api.post(f'../dde/{ddeid}/failover/start', data=start_params)

        report_started_task(ddeid, started_task,
                            f'{opname} of DDE App {dde.upper()}')


class HealthStatus:
    ACTIVE = 'ACTIVE'
    INACTIVE = 'INACTIVE'
    DEGRADED = 'DEGRADED'


def sortdict(dct: Dict):
    return OrderedDict((k, dct[k]) for k in sorted(dct.keys()))


HEALTH_FIELDS = dict(dedupe_status='DEDUPLICATION')


@commandgroup(parent=InfiniguardCommands, prefix="health")
class InfiniguardHealth:
    def get_services(self, _api) -> Dict[str, str]:
        services = _api.get('services')
        services = {service.name + '_status': service.cluster_state
                    for service in services}
        return services

    def get_management(self, _api):
        def get_mgmt_state(nodeid, services) -> Optional[str]:
            for service in services:
                if service.name == 'mgmt':
                    return service.state
            log.error(f'Could not find "mgmt" service in service list of node {nodeid}')
            return None

        node_services = _api.get('components/nodes?fields=id,services')
        mgmt_states = [get_mgmt_state(ns.id, ns.services)
                       for ns in node_services]
        all_mgmts = len(mgmt_states)
        active_mgmts = sum(1 for state in mgmt_states
                           if state == 'ACTIVE')
        if active_mgmts == all_mgmts:
            return HealthStatus.ACTIVE
        elif active_mgmts == 0:
            return HealthStatus.INACTIVE
        else:  # active_mgmts<all_mgmts:
            return HealthStatus.DEGRADED

    def get_deduplication(self, _api):
        nodes = _api.get('integrations/infiniguard/dde_nodes')
        active_nodes = sum(1 for n in nodes
                           if n.state.lifecycle == 'ACTIVE' and n.state.health == 'NORMAL')
        target_active_nodes = sum(1 for n in nodes
                                  if n.state.lifecycle != "FRU_STANDBY" and
                                  n.dde_instance_id < 3)
        if active_nodes == target_active_nodes:
            return HealthStatus.ACTIVE
        elif active_nodes == 0:
            return HealthStatus.INACTIVE
        elif active_nodes < target_active_nodes:
            return HealthStatus.DEGRADED
        else:
            log.error(f'Got {active_nodes} active nodes which is'
                      f' MORE then the expected {target_active_nodes}')
            return 'Error'

    def get_overall_service_status(self, services):
        values = set(val.lower() for val in services.values())
        if len(values) == 0:
            return 'Unknown'
        elif 'inactive' in values:
            return 'Inactive'
        elif 'degraded' in values:
            return 'Degraded'
        else:
            return 'OK'

    @ig_command(
        description="Display services health status",
        options=(Grep(), InvertedGrep())
    )
    @render(individually=True)
    def query(self, _api):
        def format_field(k, v) -> Tuple[str, str]:
            suffix = '_status'
            new_k = HEALTH_FIELDS.get(k)
            if new_k:
                k = new_k
            elif k.endswith(suffix):
                k = k[:-len(suffix)].upper()
            return k, v.capitalize()

        resp = dict(dedupe_status=self.get_deduplication(_api),
                    management_status=self.get_management(_api))
        resp.update(self.get_services(_api))
        resp = dict(format_field(k, v) for k, v in resp.items())
        resp = sortdict(resp)
        if not current_system.user.roles.is_infinidat():
            resp = sortdict({k: v for k, v in resp.items() if k.upper() in ('CORE', 'MANAGEMENT', 'DEDUPLICATION')})
        resp['service_status'] = self.get_overall_service_status(resp)
        resp.move_to_end('service_status', last=False)
        return resp
