from datetime import datetime, timedelta
from functools import partial
from . import DDEApp, DDEActiveApp, ig_command
from .lib.context_entity import ContextSensitiveEntity, DDEBoundEntity
from .nodes import DdeAppCommands
from infinishell.infinibox.exceptions import (ManagementException,
                                              NoSuchEntity)
from infinishell.infishell import commandgroup
from ..options import QueryOptions
from .lib.renderer import render_multi_fields, renderfields
from infinishell.infinibox.renderer.values import ValueTypeRenderer, StorageCapacityRenderer, \
    MillisecondsTimestampRenderer
from .. import set_renderer_query_fields, EntityCommands, render, ENTITY, roles
from collections import OrderedDict
from infinishell.infinibox.commands.infiniguard.lib.dde import allddeget, ddeget, get_dde_url
from munch import Munch
from infinishell.infishell.types import Hostname, IPAddress, EnumType


class ReplicationShare(ContextSensitiveEntity):
    ENTITY_NAME = "Replication share"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = False
    NAME_FIELD = "name"
    URI = 'stats/replication'
    FILTERABLE = False
    DDE_INSTANCE_ID = None

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    def get_description(self):
        name = self.ENTITY_NAME
        return name if self._get_context(argtype=DDEApp, raw=False) else \
            f"{name}: please specify DDE argument first"

    @classmethod
    def _entity_hook(cls, intf):
        intf[cls.NAME_FIELD] = intf.name
        return intf

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        dde_instance_id = cls.DDE_INSTANCE_ID
        alldde_replication_shares = []
        for dde_id, replication_shares_list in allddeget(api, cls.URI, ddes=(1, 2)).items():
            if replication_shares_list:
                for share in replication_shares_list:
                    alldde_replication_shares.append(cls._entity_hook(Munch(dde_instance_id=dde_id, **share)))

        if kwargs.get("hook"):
            alldde_replication_shares = kwargs["hook"](alldde_replication_shares)

        return alldde_replication_shares

    @classmethod
    def get_dde(cls, api, dde: DDEApp, **kwargs):
        replication_shares = ddeget(api, dde.dde_instance_id, cls.URI, timeout=29, **kwargs)
        if kwargs.get("hook"):
            replication_shares = kwargs["hook"](replication_shares)
        return replication_shares

    @classmethod
    def get_by_name(cls, api, name):
        if not (name and name.strip()):
            raise NoSuchEntity(u"No such {}: name is empty".format(cls.get_entity_name()))
        entity = cls._fetch_entity_by_name(api, name)
        if entity is not None:
            cls._validate_fetched_entity(entity)
            return entity
        raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(cls.get_entity_name(), name))


# class ReplicationTarget(DDEBoundEntity):
#     ENTITY_NAME = "Replication target"
#     PAGED = False
#     PLUCKABLE = False
#     SEARCHABLE = False
#     ID_FIELD = "target"
#     UNIQUE_NAME = False
#     NAME_FIELD = "target"
#     URI = 'api/replication/targets'
#     FILTERABLE = False
#     DDE_INSTANCE_ID = None
#
#
class ReplicationRateRenderer(ValueTypeRenderer):
    _rates = ((1000**4, "TB", "TB/sec"),
              (1000**3, "GB", "GB/sec"),
              (1000**2, "MB", "MB/sec"),
              (1000,    "KB", "KB/sec"),
              (1,       "B",   "B/sec"))

    def render(self, value, empty_value, **properties):
        if value == 0 or value is None:
            return "-"
        _, wanted_unit = properties.get("capacity_unit", (None, None))
        for rate, unit, txt in self._rates:
            if (wanted_unit and unit == wanted_unit) or (not wanted_unit and value >= rate):
                return "{:.2f} {}".format(round(value / float(rate)), txt)


class DurationRenderer(ValueTypeRenderer):

    def render(self, value, empty_value, **properties):  # @UnusedVariable
        if value is None:
            return '-'
        r = timedelta(seconds=value)
        return str(r.days) + ' days ' + (datetime.utcfromtimestamp(0) + r).strftime('%H hours %M minutes')


class OptionalMillisecondsTimestampRenderer(MillisecondsTimestampRenderer):

    def render(self, timestamp, empty_value, **properties):
        if timestamp is None:
            return '-'
        elif isinstance(timestamp, str):
            timestamp = int(timestamp)
        return datetime.fromtimestamp(timestamp / 1000.0).strftime("%Y-%m-%d %H:%M:%S")


REPORT_URI = "stats/replication"
REPORT_SEND_FIELDS = OrderedDict(
    name='name',
    replication_target='target',
    replication_state='state',
    replication_type='last job type',
    replication_status='last job status',
    replication_start='last job start time',
    replication_completion='last job end time',
    average_transfer_rate='avg throughput',
    estimated_duration='estimated duration',
    estimated_completion_time='estimated end time',
    remaining='remaining',
    original_data_size='size',
    last_transferred='last transferred file/dir time',
    remaining_data_to_transfer='remaining data to transfer'
)

SHORT_REPORT_SEND_FIELDS = [x for x in REPORT_SEND_FIELDS if x not in ('original_data_size',
                                                                       'last_transferred',
                                                                       'estimated_completion_time',
                                                                       'remaining_data_to_transfer')]

REPORT_RECEIVE_FIELDS = OrderedDict(
    name='name',
    filedir_target='state',
    last_sync_time='last sync time'
)

INFO_SUBSTITUTES = OrderedDict(
    targetName='Target_dedup_engines',
    originalDataSize='Original_data_size',
    actualDataSent='Actual_data_sent',
    averageSendRate='Average_send_rate',
    sourceName='Source_dedup_engines',
    actualDataReceived='Actual_data_received',
    averageReceiveRate='Average_receive_rate'
)
INFO_FIELDS = dict(
    target=list(INFO_SUBSTITUTES)[:4],
    source=list(INFO_SUBSTITUTES)[4:]
)
REPLICATION_INFO_RENDERERS = dict(
    actualDataSent=StorageCapacityRenderer(),
    actualDataReceived=StorageCapacityRenderer(),
    averageSendRate=ReplicationRateRenderer(),
    averageReceiveRate=ReplicationRateRenderer(),
)


@commandgroup(prefix="replication", parent=DdeAppCommands)
class ReplicationCommands(object):
    @ig_command(
        description="Display the DDE APP replication information",
        types=(DDEActiveApp,),
    )
    @render_multi_fields(
        value_renderers=REPLICATION_INFO_RENDERERS,
        fields=list(INFO_SUBSTITUTES),
        detailed_individually=True,
        substitute=INFO_SUBSTITUTES,
        individually=True,
    )
    def info(self, _api, dde):
        data = []
        for host_kind in list(INFO_FIELDS):
            host_info = ddeget(_api, dde.dde_instance_id, f'rest/{host_kind.capitalize()}Stats', unwrap_result=False)
            accumulator = Munch()
            for host in host_info.get(f'{host_kind}Stat', [Munch({fn: "none" if fn == 'hostName' else 0
                                                                  for fn in ['hostName']+INFO_FIELDS[host_kind][1:]})]):
                for field, value in host.items():
                    if field == 'hostName':
                        accumulator[f'{host_kind}Name'] = accumulator.get(f'{host_kind}Name', [])+[value]
                    else:
                        accumulator[field] = accumulator.get(field, 0)+value
            data = data+[Munch(
                fields=INFO_FIELDS[host_kind],
                title=f'{"Send" if host_kind == "target" else "Receive"} settings and cumulative statistics',
                data=[accumulator]
            )]

        control = {
            "allow_json": True,
            "allow_csv": False,
            "additional_fields": [],
            "fields": list(INFO_SUBSTITUTES)
        }
        return {"control": control, "data": data}


@commandgroup(prefix="report", parent=ReplicationCommands)
class ReplicationReportCommands(object):

    @staticmethod
    def _get_replication_shares(_api, dde, sharename=None, filedir_target=False):
        """
        Filter out sending (source) replication shares
        """

        filedir_target = 'Enabled' if filedir_target else 'Disabled'
        # share name list received from autocomplete is not filtered by dde_instance_id.
        if sharename:
            shares = [share for share in sharename if share.dde_instance_id == dde.dde_instance_id]
        else:
            shares = ReplicationShare.get_dde(_api, dde)

        result = [x for x in shares if x.get('filedir_target', '') == filedir_target
                  or x.get('type') == 'REPL_VTL']
        return sorted(result, key=lambda rep_share: rep_share.name.lower())

    @ig_command(
        description="Replication report command for send",
        types=(DDEActiveApp, ReplicationShare,),
        options=QueryOptions(fields=SHORT_REPORT_SEND_FIELDS,
                             substitute=REPORT_SEND_FIELDS,
                             detailed=True, ),
    )
    @renderfields(
        value_renderers=dict(replication_completion=OptionalMillisecondsTimestampRenderer(),
                             replication_start=OptionalMillisecondsTimestampRenderer(),
                             estimated_completion_time=OptionalMillisecondsTimestampRenderer(),
                             estimated_duration=DurationRenderer(),
                             last_transferred=OptionalMillisecondsTimestampRenderer()),
        fields=SHORT_REPORT_SEND_FIELDS,
        detailed_fields=REPORT_SEND_FIELDS.keys(),
        substitute=REPORT_SEND_FIELDS,
        detailed_individually=True,
    )
    def send(self, _api, dde, sharename=None):
        set_renderer_query_fields()
        return self._get_replication_shares(_api, dde, sharename, filedir_target=False)

    @ig_command(
            description="Replication report command for receive",
            types=(DDEActiveApp, ReplicationShare,),
            options=QueryOptions(fields=REPORT_RECEIVE_FIELDS.keys(),
                                 detailed=True, ),
        )
    @renderfields(
        value_renderers=dict(last_sync_time=OptionalMillisecondsTimestampRenderer()),
        fields=REPORT_RECEIVE_FIELDS,
        substitute=REPORT_RECEIVE_FIELDS,
        detailed_individually=True,
    )
    def receive(self, _api, dde, sharename=None):
        set_renderer_query_fields()
        return self._get_replication_shares(_api, dde, sharename, filedir_target=True)
#
#
# APP_ID_TO_NAME = {
#     1: 'a',
#     2: 'b',
#     3: 'standby'
# }
#
#
# class ReplicationSource(DDEBoundEntity):
#     ENTITY_NAME = "Replication source"
#     PAGED = False
#     PLUCKABLE = False
#     SEARCHABLE = False
#     ID_FIELD = "source"
#     UNIQUE_NAME = False
#     NAME_FIELD = "source"
#     URI = 'api/replication/sources'
#     FILTERABLE = False
#     DDE_INSTANCE_ID = None
#
#
# REPLICATION_SRC_FIELDS = OrderedDict(
#     source='source',
#     ddeapp='DDE_App',
# )
#
#
# @commandgroup(prefix="sources", parent=ReplicationCommands)
# class ReplicationSourcesCommands(EntityCommands):
#     ENTITY_TYPE = ReplicationSource
#
#     def _query_hook(self, sources):
#         for src in sources:
#             src.ddeapp = APP_ID_TO_NAME[src.dde_instance_id]
#             yield src
#
#     @ig_command(
#         description="View a list of all replication sources",
#         types=(DDEActiveApp,),
#         options=QueryOptions(
#             fields=list(REPLICATION_SRC_FIELDS.keys())[:1],
#             substitute=REPLICATION_SRC_FIELDS,
#             detailed=True),
#     )
#     @render(
#         fields=list(REPLICATION_SRC_FIELDS.keys())[:1],
#         detailed_fields=REPLICATION_SRC_FIELDS.keys(),
#         substitute=REPLICATION_SRC_FIELDS,
#     )
#     def query(self, _api, dde):
#         return self._query(_api, _hook=partial(self._query_hook),
#                            dde_instance_id=dde.dde_instance_id)
#
#     @ig_command(
#         description="Delete a replication source",
#         types=(DDEActiveApp, ReplicationSource,)
#     )
#     @roles.INFINIDAT.ADMIN
#     def delete(self, _api, dde, source):
#         source = ReplicationSource.filter_by_dde(dde, source)
#         with ReplicationSource.patch_dde_entity(dde):
#             self._delete(_api, source)
#         return f'Replication source host {source.source} was removed'
#
#     @ig_command(
#         description="Add a replication source",
#         types=(DDEActiveApp, Hostname)
#     )
#     @roles.INFINIDAT.ADMIN
#     def add(self, _api, dde, source):
#         with ReplicationSource.patch_dde_entity(dde):
#             self._create(_api, returns=ENTITY, source=source)
#         return f'Replication source host {source} was added'
#
#
# class EncryptType(EnumType):
#     def get_description(self):
#         return "Encryption type: 128-BIT, 256-BIT, TLS or NONE"
#
#     def _get_values(self):
#         return ['128-BIT', '256-BIT', 'TLS', 'NONE']
#
#     def _get_actual(self, value):
#         return True, value
#
#
# REPLICATION_TGT_FIELDS = OrderedDict(
#     target='target_host',
#     source_ip='source_ip',
#     encryption='encryption',
#     encryption_type='encryption_type',
#     state='state',
#     rep_revision='rep_revision',
#     ddeapp='DDE_App',
# )
#
#
# @commandgroup(prefix="targets", parent=ReplicationCommands)
# class ReplicationTargetsCommands(EntityCommands):
#     ENTITY_TYPE = ReplicationTarget
#
#     def _query_hook(self, targets):
#         for tgt in targets:
#             tgt.ddeapp = APP_ID_TO_NAME[tgt.dde_instance_id]
#             yield tgt
#
#     @ig_command(
#         description="View a list of all replication targets",
#         types=(DDEActiveApp,),
#         options=QueryOptions(
#             fields=list(REPLICATION_TGT_FIELDS.keys())[:6],
#             substitute=REPLICATION_TGT_FIELDS,
#             detailed=True),
#     )
#     @render(
#         fields=list(REPLICATION_TGT_FIELDS.keys())[:6],
#         detailed_fields=REPLICATION_TGT_FIELDS.keys(),
#         substitute=REPLICATION_TGT_FIELDS,
#     )
#     def query(self, _api, dde):
#         return self._query(_api, _hook=partial(self._query_hook),
#                            dde_instance_id=dde.dde_instance_id)
#
#     @ig_command(
#         description="Delete a replication target",
#         types=(DDEActiveApp, ReplicationTarget,)
#     )
#     @roles.INFINIDAT.ADMIN
#     def delete(self, _api, dde, target):
#         target = ReplicationTarget.filter_by_dde(dde, target)
#         with ReplicationTarget.patch_dde_entity(dde):
#             self._delete(_api, target)
#         return f"Replication target host {target.target} was removed"
#
#     @ig_command(
#         description="Add a replication target",
#         types=(DDEActiveApp, Hostname, IPAddress, EncryptType)
#     )
#     @roles.INFINIDAT.ADMIN
#     def add(self, _api, dde, target, source_ip, encryption):
#         with ReplicationTarget.patch_dde_entity(dde):
#             create_args = Munch(target=target, source_ip=source_ip.ip, encryption=encryption)
#             self._create(_api, returns=ENTITY, **create_args)
#         return "Replication target was added successfully"
#
#     @ig_command(
#         description="Pause replication services for a target",
#         types=(DDEActiveApp, ReplicationTarget,)
#     )
#     @roles.INFINIDAT.ADMIN
#     def pause(self, _api, dde, target):
#         target = ReplicationTarget.filter_by_dde(dde, target)
#         with ReplicationTarget.patch_dde_entity(dde):
#             self._post(_api, target, 'pause')
#         return f"Replication to {target.target} is now paused"
#
#     @ig_command(
#         description="Resume replication services to a target",
#         types=(DDEActiveApp, ReplicationTarget,)
#     )
#     @roles.INFINIDAT.ADMIN
#     def resume(self, _api, dde, target):
#         target = ReplicationTarget.filter_by_dde(dde, target)
#         with ReplicationTarget.patch_dde_entity(dde):
#             self._post(_api, target, 'resume')
#         return f"Replication to {target.target} is now resumed"
