from infinishell.infinibox.commands.cluster import SORTABLE_FIELDS

from munch import Munch

from infinishell.infishell import commandgroup, command, render
from infinishell.infishell.types import PreferredControllers, YesNo, Str, Int
from infinishell.infishell.utils import strlist, NOTHING
from infinishell.infishell.exceptions import CommandAborted, BadArguments,CommandFailed,BadValue
from infinishell.infinibox.exceptions import NoSuchEntity

from ....utils import input_password
from ..types import (
    OrgUnit,
    SmbGroup,
    SmbUser,
    Hostname,
    SmbPrivileges,
    SmbGroupName,
    SmbActiveDirectoryDomain,
    SmbUserName,
    UnixUser,
    SupplementaryGsids,
    SID,
    PrimaryGsid,
    UID,
    GID,
    )
from ..options import QueryOptions
from .. import EntityCommands, Null, roles, render_collection
from . import ConfigCommandsBase


@commandgroup(parent=ConfigCommandsBase, prefix="nas")
class NasCommands(ConfigCommandsBase):
    pass

@commandgroup(parent=NasCommands, prefix="active_directory")
class ADCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = SmbActiveDirectoryDomain

    @command(
        name="query",
        description="Show Active Directory domain details",
        )
    @render_collection(
        individually=True,
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def details(self, _api):
        domain_result = SmbActiveDirectoryDomain.get_all(_api)
        domain_query = {
            "domain": domain_result.domain,
            "preferred_controllers": domain_result.preferred_ips,
            "org_unit": domain_result.org_unit,
            "preferred_ips": domain_result.preferred_ips,
        }

        return [domain_query]

    @command(
        description="Join an Active Directory domain",
        types=(Str, Str, PreferredControllers.List, Str, OrgUnit),
        )
    @roles.INFINIDAT.ADMIN
    def join(self, _api, domain, username, preferred_controllers, password=None, org_unit=None):
        if password is None:
            password = input_password('Please enter a password for user {}: '.format(username),
            interaction_exc_type=BadArguments, interaction_error="Please specify a password")
             # url# new type
        data = dict(
            domain=domain,
            org_unit=org_unit if org_unit else "",
            username=username,
            password=password,
            preferred_ips=preferred_controllers,)
        result = _api.post("/activedirectory/domains", data)

    @command(
        description="Change Active Directory settings",
        types=(PreferredControllers.List,),
        )
    @roles.INFINIDAT.ADMIN
    def set_preferred_controllers(self, _api, preferred_controllers):
        result = _api.post("/activedirectory/domains/set_preferred_ips", preferred_controllers)

    @command(
        description="Leave the Active Directory domain",
        types=(Str, Str, YesNo)
        )
    @roles.INFINIDAT.ADMIN
    def leave(self, _api, username, password=None, force=None):
        if password is None:
            password = input_password('Please enter a password for user {}: '.format(username),
            interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        data = dict(
            username=username,
            password=password,
        )
        result = _api.post('/activedirectory/domains/leave' + ("?approved=true&force=True" if force else ''), data)


@commandgroup(parent=NasCommands, prefix="smb_users")
class SmbUserCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = SmbUser

    @command(
        description="Create a new SMB local user",
        types=(SmbUserName, Str, YesNo, SmbGroup, SmbGroup.List, SmbPrivileges.List, UID),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, password=None, enabled=True, primary_group=None, groups=None, privileges=["SECHANGENOTIFYPRIVILEGE"], uid=None):
        if password is None:
            password = input_password('Please enter a password for user {}: '.format(name),
            interaction_exc_type=BadArguments, interaction_error="Please specify a password")
             # url# new type
        data = Munch(
            name = name,
            password = password,
            enabled = enabled,
            primary_group_id = primary_group,
            group_ids = groups,
            privileges = privileges,
            uid = uid,
        )
        return self._create(_api, **data)

    @command(
        description="Change the password for an SMB local user",
        types=(SmbUser, Str),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def change_password(self, _api, user, password=None):
        if password is None:
            password = input_password('New password: ',
            interaction_exc_type=BadArguments, interaction_error="Please specify a password")
            if password != input_password(prompt="Re-enter new password: "):
                raise CommandFailed("Passwords differ from one another, please try again")
        if not password.strip():
            raise CommandAborted("Password is empty, aborting")
        self._update(_api, user, password=password)
        return self._result_message(user, "password changed")

    @command(
        description="Change the primary group for an SMB local user",
        types=(SmbUser, SmbGroup),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_primary_group(self, _api, user, primary_group):
        self._update(_api, user, primary_group_id=primary_group)
        return self._result_message(user, "updated")

    @command(
        description="Add the SMB local user as member in the group",
        types=(SmbUser, SmbGroup.List),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def add_group(self, _api, user, group):
        for g in group:
            for ug in user.groups:
                if g.id == ug.id:
                    raise BadValue('SMB user "{}" is already a member of group "{}"'.format(user.name, g.name))
        all_groups = user.groups + group
        self._update(_api, user, group_ids=all_groups)
        return self._result_message(user, "updated")

    @command(
        description="Remove the SMB local user from the group",
        types=(SmbUser, SmbGroup),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def remove_group(self, _api, user, group):
        if group not in user.groups:
            raise BadValue('SMB user "{}" is not a member of non-primary group "{}"'.format(user.name,group.name))
        user.groups.remove(group)
        self._update(_api, user, group_ids=user.groups)
        return self._result_message(user, "updated")
    
    @command(
        description="Change privileges for an SMB local user",
        types=(SmbUser, SmbPrivileges.List),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_privileges(self, _api, user, privileges):
        self._update(_api, user, privileges=privileges)
        return self._result_message(user, "updated")

    QUERY_FIELDS = strlist("name sid enabled primary_group groups privileges uid")
    SORTABLE_FIELDS = strlist("name enabled primary_group uid")
    QUERY_SUBSTITUTES = dict(
        name="user"
        )

    @command(
        description="List existing SMB local users",
        types=(SmbUser, YesNo, SmbGroup, UID),
        options=QueryOptions(fields=QUERY_FIELDS,substitute=QUERY_SUBSTITUTES ,detailed=True, sortable=SORTABLE_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        )

    @roles.INFINIDAT.ADMIN.POOLADMIN
    def query(self, _api, user=None, enabled=None, primary_group=None, uid=None):
        return self._query(_api, user, _hook=self._query_hook, enabled=enabled, primary_group_id=primary_group, uid=uid)

    def _query_hook(self, users):
        for user in users:
            user.primary_group= user.primary_group.name if hasattr(user, "primary_group") else "N/A"
            user.groups = list(map(lambda group: group.name, user.groups))
            yield user

    @command(
        description="Disable an SMB local user",
        types=(SmbUser,),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def disable(self, _api, user):
        if not user.enabled:
            return self._result_message(user, "is already disabled")
        self._update(_api, user, enabled=False)
        return self._result_message(user, "disabled")

    @command(
        description="Enable an SMB local user",
        types=(SmbUser,),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def enable(self, _api, user):
        if user.enabled:
            return self._result_message(user, "is already enabled")
        self._update(_api, user, enabled=True)
        return self._result_message(user, "enabled")

    @command(
        description="Delete an SMB local user",
        types=(SmbUser,),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, user):
        return self._delete(_api, user)
    
    @command(
        description="Set uid for local user",
        types=(SmbUser, UID),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_uid(self, _api, user, uid):
        self._update(_api, user, uid=uid)
        return self._result_message(user, "updated")

    @command(
        description="Clear privileges for an SMB local user",
        types=(SmbUser,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear_privileges(self, _api, user):
        self._update(_api, user, privileges=[])
        return self._result_message(user, "updated")
    
    @command(
        description="Clear uid for an SMB local user",
        types=(SmbUser,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear_uid(self, _api, user):
        self._update(_api, user, uid=Null)
        return self._result_message(user, "updated")
@commandgroup(parent=NasCommands, prefix="smb_groups")
class SmbGroupCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = SmbGroup

    @command(
        description="Create a new SMB local group",
        types=(SmbGroupName, SmbPrivileges.List, SID.List, GID, UID),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, name, privileges=["SECHANGENOTIFYPRIVILEGE"], domain_members=None, gid=None, uid=None):
        data = Munch(
            name = name,
            privileges = privileges,
            domain_members = domain_members,
            gid = gid,
            uid = uid,
        )
        return self._create(_api, **data)


    @command(
        description="Change privileges for an SMB local group",
        types=(SmbGroup, SmbPrivileges.List),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_privileges(self, _api, group, privileges):
        self._update(_api, group, privileges=privileges)
        return self._result_message(group, "updated")


    @command(
        description="Change the AD domain members for an SMB local group",
        types=(SmbGroup, SID.List),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_domain_members(self, _api, group, domain_members):
        self._update(_api, group, domain_members=domain_members)
        return self._result_message(group, "updated")


    @command(
        description="Delete an SMB local group",
        types=(SmbGroup,),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, group):
        return self._delete(_api, group)


    QUERY_FIELDS = strlist("name privileges domain_members sid uid gid")
    SORTABLE_FIELDS = strlist("name uid gid")
    QUERY_SUBSTITUTES = dict(
        name="group"
        )

    @command(
        description="List existing SMB local groups",
        types=(SmbGroup, UID, GID),
        options=QueryOptions(fields=QUERY_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True, sortable=SORTABLE_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def query(self, _api, group=None, uid=None, gid=None):
        return self._query(_api, group, uid=uid, gid=gid)
    
    
    @command(
    description="Set gid for SMB local group",
    types=(SmbGroup, GID),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_gid(self, _api, group, gid):
        self._update(_api, group, gid=gid)
        return self._result_message(group, "updated")
    
    @command(
    description="Set uid for SMB local group",
    types=(SmbGroup, UID),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_uid(self, _api, group, uid):
        self._update(_api, group, uid=uid)
        return self._result_message(group, "updated")

    @command(
        description="Clear privileges for an SMB local group",
        types=(SmbGroup,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear_privileges(self, _api, group):
        self._update(_api, group, privileges=[])
        return self._result_message(group, "updated")
    
    @command(
        description="Clear uid for an SMB local group",
        types=(SmbGroup,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear_uid(self, _api, group):
        self._update(_api, group, uid=Null)
        return self._result_message(group, "updated")
    
    @command(
        description="Clear gid for an SMB local group",
        types=(SmbGroup,),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def clear_gid(self, _api, group):
        self._update(_api, group, gid=Null)
        return self._result_message(group, "updated")

@commandgroup(parent=NasCommands, prefix="nfs_users")
class NFSCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = UnixUser
     
    @command(
        description="Create a new UNIX user",
        types=(Int, Int, Int.List, SID, PrimaryGsid, SupplementaryGsids),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, uid, primary_gid, supplementary_gids=None, sid=None, primary_gsid=None, supplementary_gsids=None):
        self._create(_api, uid=uid, primary_gid=primary_gid, supplementary_gids=supplementary_gids, sid=sid, primary_gsid=primary_gsid, supplementary_gsids=supplementary_gsids)
        return "Unix user '{}' created".format(uid)
    
    @command(
        description="Update a UNIX user",
        types=(Int, Int, Int.List, SID, PrimaryGsid, SupplementaryGsids),
        hidden_substitue=dict(name="UNIX user"),
        )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, uid, primary_gid=None, supplementary_gids=None, sid=None, primary_gsid=None, supplementary_gsids=None):
        try:
            unix_user = UnixUser.get_by_uid(_api, uid)
        except NoSuchEntity as e:
            raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(self.ENTITY_TYPE.get_entity_name(), uid))
        return self._update(_api, unix_user, primary_gid=primary_gid, supplementary_gids=supplementary_gids, sid=sid, primary_gsid=primary_gsid, supplementary_gsids=supplementary_gsids)
    
    @command(
        description="Delete a UNIX user",
        types=(Int,),
        hidden_substitue=dict(name="unix user"),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, uid):
        try:
            unix_user = UnixUser.get_by_uid(_api, uid)
        except NoSuchEntity as e:
            raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(self.ENTITY_TYPE.get_entity_name(), uid))
        self._delete(_api, unix_user)
        return "Unix user '{}' deleted".format(uid)
    
    
    UNIX_USER_QUERY_FIELDS = strlist("uid primary_gid supplementary_gids sid primary_gsid supplementary_gsids")
    UNIX_USER_SORTABLE_FIELDS = strlist("uid primary_gid")
    
    @command(
        description="List existing UNIX users",
        types=(Int, Int,),
        options=QueryOptions(fields=UNIX_USER_QUERY_FIELDS, sortable=UNIX_USER_SORTABLE_FIELDS, detailed=False),
    )
    @render(
        entity=UnixUser,
        fields=UNIX_USER_QUERY_FIELDS,
    )
    @roles.INFINIDAT.ADMIN
    def query(self, _api, uid=None, primary_gid=None):
        return self._query(_api, uid=uid, primary_gid=primary_gid)
    

    @command(
        description="Clear sid and primary_gsid for a UNIX user",
        types=(Int,),
        hidden_substitue=dict(name="UNIX user"),
        )
    @roles.INFINIDAT.ADMIN
    def clear_sid_primary_gsid(self, _api, uid):
        try:
            unix_user = UnixUser.get_by_uid(_api, uid)
        except NoSuchEntity as e:
            raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(self.ENTITY_TYPE.get_entity_name(), uid))
        return self._update(_api, unix_user, sid=Null, primary_gsid=Null)
    
    @command(
        description="Clear supplementary_gsids for a UNIX user",
        types=(Int,),
        hidden_substitue=dict(name="UNIX user"),
        )
    @roles.INFINIDAT.ADMIN
    def clear_supplementary_gsids(self, _api, uid):
        try:
            unix_user = UnixUser.get_by_uid(_api, uid)
        except NoSuchEntity as e:
            raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(self.ENTITY_TYPE.get_entity_name(), uid))
        return self._update(_api, unix_user, supplementary_gsids=Null)