from __future__ import annotations

import signal
import stat
from contextlib import contextmanager
from typing import Iterable, Tuple

from munch import Munch

from infinishell.utils import get_username_password
from . import DDEActiveApp, ig_command, DDEApp
from .lib.asynctask import AsyncTask, DEFAULT_CTRLBRK_CONFIRMATION
from .lib.context_entity import ContextSensitiveEntity
from .lib.renderer import renderfields, Fields, FieldDescr
from .nodes import DdeAppCommands
from .lib.dde import allddeget, get_dde_url, DdeApi
from infinishell.infishell import commandgroup, get_active_shell, CommandFailed
from infinishell.infishell.types import RegexType, Str, Positive
from .. import set_renderer_query_fields, EntityCommands, roles
from infinishell.infinibox.commands.options import OutputFormat, Grep, QueryOptions, InvertedGrep

from ...renderer.renderer import Weight, FieldSpecs, CollectionRenderer
from ...renderer.values import ValueTypeRenderer, MillisecondsTimestampRenderer


class DdeSensitiveEntity(ContextSensitiveEntity):
    DDEAPPTYPE = DDEApp
    DDELIST = (1, 2, 3)

    def get_description(self):
        name = self.ENTITY_NAME
        return name if self._get_context(argtype=self.DDEAPPTYPE, raw=False) else \
            "{name}: please specify DDE first"

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    def _context_hook(self, results, **kwargs):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if not dde:
            return []

        return [obj for obj in results if obj.dde_instance_id == dde.dde_instance_id]

    @classmethod
    def _entity_hook(cls, entity):
        return entity

    @classmethod
    def set_ddeid(cls, entity, ddeid):
        entity['dde_instance_id'] = ddeid

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        result_objects = []
        for dde_id, objectlist in allddeget(api, cls.URI, ddes=cls.DDELIST).items():
            if not objectlist:
                continue
            for o in objectlist:
                cls.set_ddeid(o, dde_id)
                result_objects.append(cls._entity_hook(o))

        if hook := kwargs.get("hook"):
            result_objects = hook(result_objects)

        return result_objects


class AppSpecificShare(DdeSensitiveEntity):
    ENTITY_NAME = "App-specific share"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = True
    NAME_FIELD = "name"
    URI = 'api/nas/app_specific/shares'
    FILTERABLE = False
    DDELIST = (1, 2)


class MTime(RegexType):
    PATTERN = r'[+-]?\d+'

    def _get_pattern(self):
        return self.PATTERN

    def _get_error(self):
        return 'is not a valid UNIX find(1) mtime/atime value'

    def get_description(self):
        return 'UNIX find(1) mtime/atime value, e.g "5", "+7", etc.'


ATime = MTime
User = Str
FilePattern = Str


class FileModeRenderer(ValueTypeRenderer):
    def render(self, value, empty_value, **properties):  # @UnusedVariable
        return stat.filemode(value) if value != empty_value else str(empty_value)


class SecondsTimestampRenderer(MillisecondsTimestampRenderer):
    def render(self, timestamp, empty_value, **properties):
        return super().render(timestamp * 1000, empty_value, **properties)


CONTENT_FIELDS = Fields([
    FieldDescr('mode', len('-rw-r-----'),
               renderer=FileModeRenderer),
    FieldDescr('uid', 5),
    FieldDescr('gid', 5),
    FieldDescr('ctime', FieldSpecs.DATE,
               substitute='Status Change',
               renderer=SecondsTimestampRenderer),
    FieldDescr('mtime', FieldSpecs.DATE,
               substitute='Modified',
               renderer=SecondsTimestampRenderer),
    FieldDescr('size', 10, short_output=True),
    FieldDescr('filename',
               Weight(weight=25, fixed=100, max=1024),
               substitute='name',
               short_output=True),
])


def print_task_lines(task: AsyncTask):
    shell = get_active_shell()
    for fd, line in task.lines:
        _print = shell.output if fd == 1 else shell.error
        _print(line)


def files_or_lines(task: AsyncTask):
    yield from task.files
    if task.result['rc'] != 0:
        print_task_lines(task)
        raise CommandFailed


@contextmanager
def disable_ctrlc():
    original = signal.signal(signal.SIGINT, signal.SIG_IGN)
    try:
        yield
    finally:
        signal.signal(signal.SIGINT, original)


@commandgroup(prefix="appspecific", parent=DdeAppCommands)
class AppSpecificCommands(EntityCommands):
    FILES_URI = 'files'
    ENTITY_TYPE = AppSpecificShare

    @ig_command(
        description="View a list of all existing Application Specific shares",
        types=(DDEActiveApp,),
        options=[OutputFormat(), Grep(), InvertedGrep()],
    )
    def query(self, _api, dde):
        return (Munch(name=share.name) for share in self._query(_api) if dde.dde_instance_id == share.dde_instance_id)

    def _find_args_to_query_params(self, mtime, atime, maxdepth, pattern):
        paramnames = ['mtime', 'atime', 'maxdepth', 'pattern']
        ll = locals()
        return {name: param for name in paramnames if (param := ll[name])}

    @ig_command(
        description="View an Application Specific share content",
        types=(DDEActiveApp, AppSpecificShare, MTime, ATime, Positive, FilePattern),
        options=QueryOptions(fields=CONTENT_FIELDS.fieldnames, detailed=True)
    )
    @renderfields(
        entity='file-list-element',
        value_renderers=CONTENT_FIELDS.renderers,
        general=CollectionRenderer,
        fields=CONTENT_FIELDS.fieldnames,
        detailed_fields=CONTENT_FIELDS.detailedfieldnames,
        substitute=CONTENT_FIELDS.substitute,
        detailed_individually=False
    )
    def find(self, _api, dde, share, mtime=None, atime=None, maxdepth=None, pattern=None):
        result = DdeApi(_api, dde.dde_instance_id).post(
            f'api/nas/app_specific/shares/{share.name}/files',
            params=self._find_args_to_query_params(mtime, atime, maxdepth, pattern))

        set_renderer_query_fields()
        CONTENT_FIELDS.register_specs('file-list-element')
        return files_or_lines(
            AsyncTask(_api, dde.dde_instance_id, result))

    @ig_command(
        description="Delete Application Specific share content",
        types=(DDEActiveApp, AppSpecificShare, Str, FilePattern, MTime, ATime, Positive),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, dde, share, user, pattern, mtime=None, atime=None, maxdepth=None):
        _, password = get_username_password(username=user, prompt=None)
        with disable_ctrlc():
            result = DdeApi(_api, dde.dde_instance_id).delete(
                f'api/nas/app_specific/shares/{share.name}/files',
                data=dict(user=user, password=password),
                params=self._find_args_to_query_params(mtime, atime, maxdepth, pattern))

        print_task_lines(AsyncTask(_api, dde.dde_instance_id, result, DEFAULT_CTRLBRK_CONFIRMATION))
