from infinishell.infinibox.commands import roles
from infinishell.infinibox.commands.infiniguard import ig_command, DDEActiveApp
from infinishell.infinibox.commands.infiniguard.lib.dde import DdeApi
from infinishell.infinibox.commands.infiniguard.lib.renderer import Fields, FieldDescr, renderfields
from infinishell.infinibox.commands.infiniguard.nodes import DdeAppCommands
from infinishell.infinibox.commands.infiniguard.lib.types import Password
from infinishell.infinibox.exceptions import ManagementException
from infinishell.infishell import commandgroup, BadArguments, CommandFailed, BadValue
from infinishell.infishell.types import Str, IPAddress, RegexType, Hostname, Bool, ArgType, List
from infinishell.utils import get_username_password

INFO_FIELDS = Fields([
    FieldDescr('enabled', short_output=True),
    FieldDescr('domain_type', short_output=True, substitute="Domain Type"),
    FieldDescr('domain', short_output=True),
    FieldDescr('workgroup', short_output=True, substitute="Workgroup"),
    FieldDescr('pdc', short_output=True, substitute="Primary Domain Controller"),
    FieldDescr('org', short_output=True, substitute="Organization"),
])

SETTINGS_FIELDS = Fields([
    FieldDescr('server_signing', short_output=True, substitute="server signing"),
    FieldDescr('oplocks', short_output=True, substitute="oplocks")
])


SHARE_ADMINS_FIELDS = Fields([
    FieldDescr('name', short_output=True, substitute="List of Active Directory accounts or groups"),
])


class ADDomainName(RegexType):

    def _get_pattern(self):
        return r"^[a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?(\.[a-zA-Z0-9]([a-zA-Z0-9\-]{0,61}[a-zA-Z0-9])?)*$"

    def _get_error(self):
        return "is not a valid Active Directory Domain Name"

    def get_description(self):
        return "A valid Active Directory Domain Name"


class ADUserName(RegexType):

    def _get_pattern(self):
        return r"^[a-zA-Z0-9_\.\-]{1,256}$"

    def _get_error(self):
        return "is not a valid Active Directory Username"

    def get_description(self):
        return "A valid Active Directory Username"


class ADOrgUnit(RegexType):

    def _get_pattern(self):
        return r"^[\s\S]{1,64}(?:/[\s\S]{1,64})*$"

    def _get_error(self):
        return "is not a valid Active Directory Organizational Unit, e.g.: OU1/OU2"

    def get_description(self):
        return "A valid Active Directory Organizational Unit"


class ADShareAdmin(RegexType):

    def _get_pattern(self):
        return r"^[a-zA-Z0-9]+(?:\.[a-zA-Z0-9]+)?\\[a-zA-Z0-9].*$"

    def _get_error(self):
        return "This is not a valid Active Directory Account/Group."

    def _raise_value_error(self, arg, value, error):
        msg = f"{self._get_error()}\n{self.get_description()}"
        raise BadValue(msg)

    def get_description(self):
        return "Valid Active Directory account or group name is in format: 'CONTOSO.COM\\\Domain Group'" \
               " enclosed by quotes and separated by double backslash"


@commandgroup(prefix="cifs", parent=DdeAppCommands)
class CifsCommands(object):
    URL_PREFIX = 'api/nas/cifs/'

    @ig_command(
        description="Display CIFS configuration info",
        types=(DDEActiveApp,),
    )
    @renderfields(
        value_renderers=INFO_FIELDS.renderers,
        fields=INFO_FIELDS.fieldnames,
        substitute=INFO_FIELDS.substitute,
        detailed_fields=INFO_FIELDS.detailedfieldnames,
        individually=True,
    )
    def info(self, _api, dde):
        return DdeApi(_api, dde.dde_instance_id).get(self.URL_PREFIX + 'info')


@commandgroup(prefix="ads", parent=CifsCommands)
class CifsADSCommands(object):
    URL_PREFIX = 'api/nas/cifs/ads'

    @ig_command(
        description="Join Active Directory",
        types=(DDEActiveApp, ADDomainName, ADUserName, Password, Hostname, ADOrgUnit)
    )
    @roles.INFINIDAT.ADMIN
    def join(self, _api, dde, domain, admin, password=None, pdc=None, org=None):
        if not password:
            _, password = get_username_password(username=admin, prompt=None)
        data = dict(domain=domain, admin=admin, password=password,
                    pdc=pdc if pdc else "", org=org if org else "")
        result = DdeApi(_api, dde.dde_instance_id).post(self.URL_PREFIX, data=data)
        return "Successfully Joined Active Directory."

    @ig_command(
        description="Disjoin Active Directory",
        types=(DDEActiveApp,)
    )
    @roles.INFINIDAT.ADMIN
    def disjoin(self, _api, dde):
        result = DdeApi(_api, dde.dde_instance_id).delete(self.URL_PREFIX)
        return "Successfully Disjoined Active Directory."


@commandgroup(prefix="workgroup", parent=CifsCommands)
class CifsWorkgroupCommands(object):
    URL_PREFIX = 'api/nas/cifs/workgroup'

    @ig_command(
        description="Join Workgroup",
        types=(DDEActiveApp, Str)
    )
    @roles.INFINIDAT.ADMIN
    def join(self, _api, dde, name):
        data = dict(name=name)
        result = DdeApi(_api, dde.dde_instance_id).post(self.URL_PREFIX, data=data)
        return "Successfully Joined Workgroup."

    @ig_command(
        description="Disjoin Workgroup",
        types=(DDEActiveApp,)
    )
    @roles.INFINIDAT.ADMIN
    def disjoin(self, _api, dde):
        result = DdeApi(_api, dde.dde_instance_id).delete(self.URL_PREFIX)
        return "Successfully Disjoined Workgroup."


@commandgroup(prefix="settings", parent=CifsCommands)
class CifsSettingsCommands(object):
    URL_PREFIX = 'api/nas/cifs/settings'

    @ig_command(
        description="Display CIFS settings",
        types=(DDEActiveApp,),
    )
    @renderfields(
        value_renderers=SETTINGS_FIELDS.renderers,
        fields=SETTINGS_FIELDS.fieldnames,
        substitute=SETTINGS_FIELDS.substitute,
        individually=True,
    )
    def query(self, _api, dde):
        return DdeApi(_api, dde.dde_instance_id).get(self.URL_PREFIX)

    @ig_command(
        description="Set CIFS settings",
        types=(DDEActiveApp, Bool, Bool)
    )
    @roles.INFINIDAT.ADMIN
    def set(self, _api, dde, oplocks=None, server_signing=None):
        cur_settings = DdeApi(_api, dde.dde_instance_id).get(self.URL_PREFIX)
        same_value_msg = 'Already set to the same value.'
        if oplocks is not None:
            if cur_settings['oplocks'] == oplocks:
                return same_value_msg
            data = dict(oplocks=oplocks)
        elif server_signing is not None:
            if cur_settings['server_signing'] == server_signing:
                return same_value_msg
            data = dict(server_signing=server_signing)
        else:
            raise BadArguments("Unknown argument")
        result = DdeApi(_api, dde.dde_instance_id).put(self.URL_PREFIX, data=data)
        return f"Successfully Changed."


@commandgroup(prefix="share_admins", parent=CifsCommands)
class CifsShareAdminsCommands(object):
    URL_PREFIX = 'api/nas/cifs/share_admins'

    @ig_command(
        description="List Active Directory share admins",
        types=(DDEActiveApp,),
    )
    @renderfields(
        value_renderers=SHARE_ADMINS_FIELDS.renderers,
        fields=SHARE_ADMINS_FIELDS.fieldnames,
        substitute=SHARE_ADMINS_FIELDS.substitute,
        individually=False,
    )
    def query(self, _api, dde):
        return DdeApi(_api, dde.dde_instance_id).get(self.URL_PREFIX)

    @ig_command(
        description="Add AD accounts/groups to share admins list",
        types=(DDEActiveApp, ADShareAdmin)
    )
    @roles.INFINIDAT.ADMIN
    def add(self, _api, dde, name):
        result = DdeApi(_api, dde.dde_instance_id).post(self.URL_PREFIX, data=dict(name=name))
        return f"Successfully added {name} AD account/group to share admins list"

    @ig_command(
        description="Remove share admin",
        types=(DDEActiveApp, ADShareAdmin)
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, dde, name):
        result = DdeApi(_api, dde.dde_instance_id).delete(self.URL_PREFIX + '/' + name.replace('\\', '/'))
        return f"Successfully deleted {name} AD account/group from share admins list"
