from . import DDEApp, DDEActiveApp, ig_command
from infinishell.infishell import commandgroup, BadArguments
from infinishell.infishell.types import Hostname, IPAddress

from .lib.context_entity import ContextBoundEntity
from .. import EntityCommands, render, ENTITY, roles
from collections import OrderedDict
from ..options import QueryOptions
from .share import ShareCommands, PermissionType
from munch import Munch


class HostAccess(Hostname):
    def _get_patterns(self):
        return (r"^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\."
                r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$",)


class NasShare(ContextBoundEntity):
    ENTITY_NAME = "Share"
    DESCRIPTION = "NFS or App-Specific Share"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = True
    NAME_FIELD = "name"
    CONTEXT_TYPES = [DDEApp, ]
    CONTEXT_PARAMS = []
    URI = "../dde/{arg[DDEApp].actual.dde_instance_id}/api/nas/shares"
    MULTI_FIELD_PUT = True
    FILTERABLE = False

    @classmethod
    def _add_context(cls, results):
        dde = cls.cls_get_context(argtype=DDEApp)
        return [Munch(ddeapp=dde, **entry) for entry in results if entry.protocol != 'cifs']


class NasShareHost(ContextBoundEntity):
    ENTITY_NAME = "Host"
    DESCRIPTION = "NFS Share Host"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "host"
    UNIQUE_NAME = True
    NAME_FIELD = "host"
    CONTEXT_TYPES = [DDEApp, NasShare]
    URI = "../dde/{arg[DDEApp].actual.dde_instance_id}/api/nas/{arg[NasShare].actual.protocol}/shares" \
          "/{arg[NasShare].raw}/share_hosts"
    MULTI_FIELD_PUT = True
    FILTERABLE = False
    HOOK_DEPENDENCIES = ("host",)

    @classmethod
    def _add_context(cls, results):
        dde = cls.cls_get_context(argtype=DDEApp)
        share = cls.cls_get_context(param_name='share')
        return [Munch(ddeapp=dde, share=share, **entry) for entry in results]


SHARE_HOST_FIELDS = OrderedDict(
    ddeapp='DDE App',
    share='Share',
    host='host',
    permissions='permissions',
    effective_perms='effective permissions'
)


@commandgroup(prefix="host_access", parent=ShareCommands)
class HostAccessCommands(EntityCommands):
    ENTITY_TYPE = NasShareHost

    @NasShareHost.with_context
    @ig_command(
        description="List Share Hosts",
        types=(DDEActiveApp, NasShare),
        options=QueryOptions(
            fields=SHARE_HOST_FIELDS.keys(),
            substitute=SHARE_HOST_FIELDS,
            detailed=True,
        ),
    )
    @render(
        fields=list(SHARE_HOST_FIELDS.keys())[2:],
        detailed_fields=SHARE_HOST_FIELDS.keys(),
        substitute=SHARE_HOST_FIELDS,
        detailed_individually=True,
    )
    def query(self, _api, dde, share):
        return self._query(_api)

    @NasShareHost.with_context
    @ig_command(
        description="Add Share Host",
        types=(DDEActiveApp, NasShare, HostAccess, PermissionType,)
    )
    @roles.INFINIDAT.ADMIN
    def add(self, _api, dde, share, host, permissions=None):
        create_args = Munch(host=host)
        if permissions and share.protocol == 'nfs':
            create_args["permissions"] = permissions
        elif permissions and share.protocol == 'app_specific':
            raise BadArguments("argument 'permissions' is not allowed argument for application specific shares.")
        self._create(_api, returns=ENTITY, **create_args)
        return f"Host '{host}' was successfully added to the {share.protocol} share '{share.name}'s access list"

    @NasShareHost.with_context
    @ig_command(
        description="Delete Share Host",
        types=(DDEActiveApp, NasShare, NasShareHost,)
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, dde, share, host):
        self._delete(_api, host)
        return f"Host {host.host} was successfully removed from the share '{share.name}'s access list"
