import re

from ...types import Entity
from infinishell.infishell import config
from infinishell.infishell.types import ArgType, DynamicValue, List, MappedEnumType, TypeChoicesHelpMixin
from infinishell.infishell.commands.commands import ContextWrapper
from ....system import system
from ....exceptions import NoSuchEntity
from .dde import get_dde_url, allddeget
from .. import DDEApp
from infinishell.infishell.commands.commands import metadata
from functools import partial
from contextlib import contextmanager
from munch import Munch
import time
import inspect
from decorator import decorator


def get_context(context, param_name=None, argtype=None, raw=True):
    if not context:
        return None
    arg = None
    if param_name:
        arg = context.get(param_name, None)
    elif argtype:
        arg = next((a for a in context.values()
                    if (isinstance(a["type"], argtype) or a["type"] == argtype)),
                   None)
    if not arg:
        return None
    if raw:
        return arg["raw"]
    if "cooked" in arg:  # shortcut if the context has already been preprocessed
        return arg["cooked"]
    wrapped_name = ContextWrapper(arg["name"], context)  # support context-sensitive fetches downstream
    cooked = arg["type"].get_actual(wrapped_name, arg["raw"])
    if isinstance(cooked, DynamicValue):
        try:
            _, cooked = cooked.get(system.api, wrapped_name)
        except NoSuchEntity:
            cooked = None
    arg["cooked"] = cooked  # cache for potential multiple access
    return cooked


def filtered(func=None, filter_params=None):
    """ Command method decorator for defining commands with context-filtered parameters. """

    def context_filter(params, context):
        return filter_params(params, partial(get_context, context=context))

    return metadata(func, param_filter=context_filter)


class ContextMixin(ArgType):
    __CONTEXT = None
    __CACHE = {}  # can be set to None to disable caching of processed context
    __TIMEOUT = 2

    def _get_context(self, param_name=None, argtype=None, raw=True):
        return self.cls_get_context(self._context_args, param_name, argtype, raw)

    @classmethod
    def cls_get_context(cls, context=None, param_name=None, argtype=None, raw=True):
        tl = time.time()-cls.__TIMEOUT
        rawval = None
        if not raw and cls.__CACHE is not None:
            rawval = get_context(context or cls.__CONTEXT, param_name, argtype)
            key = param_name or argtype.__name__
            if key in cls.__CACHE and cls.__CACHE[key].raw == rawval and cls.__CACHE[key].t > tl:
                return cls.__CACHE[key].value
        res = get_context(context or cls.__CONTEXT, param_name, argtype, raw)
        if not raw and cls.__CACHE is not None:
            cls.__CACHE[param_name or argtype.__name__] = Munch(t=time.time(), value=res, raw=rawval)
        return res

    @classmethod
    @contextmanager
    def cls_set_context(cls, context):
        original = cls.__CONTEXT
        cls.__CONTEXT = context
        try:
            yield
        finally:
            cls.__CONTEXT = original


class ContextSensitiveEntity(Entity, ContextMixin):
    COMPLETION_CACHE = False
    CONTEXT_SENSITIVE_CACHE = False

    @classmethod
    def context_changed(cls, _api, partial_value, context_args):
        """ Subclasses should override this method, which is invoked by rlcontext
        when ``COMPLETION_CACHE`` is False and ``CONTEXT_SENSITIVE_CACHE`` is True.
        Should verify whether the context-relevant variables have changed and
        store the new state, then return True to invalidate the cache or False to use it"""
        return True

    def _context_hook(self, results):
        """ Subclasses should implement this method, which is invoked during
        ``get_possible_values()``. The method should filter results according to
        command context variables present. It may use ``_get_context()`` helper function """
        return results

    @classmethod
    def _fetch(cls, api, arg, value):
        with cls.cls_set_context(arg.context if hasattr(arg, "context") else None):
            return super(ContextSensitiveEntity, cls)._fetch(api, arg, value)

    def _get_possible_values_by_search(self, api, starts_with, type_cls):
        starts_with = starts_with or ""

        def wrapper_hook(results):
            results = self._search_hook(results)
            results = self._context_hook(results)
            return results

        query = api.paged_query(
            "search",
            page_size=config.api_max_page_size,
            params=dict(
                type=type_cls.get_entity_mgmt_name(),
                query=starts_with,
            ),
            hook=wrapper_hook,
        )
        return (entity_info.name for entity_info in query if entity_info and entity_info.name.startswith(starts_with))

    def _get_possible_values_by_query(self, api, starts_with, type_cls):
        type_cls = type_cls or type(self)
        params = {}
        # Create pluck for entity's name field, and any other fields possibly necessary for invoking the default query hook
        pluck = None if not type_cls.PLUCKABLE else api.create_pluck(type_cls.NAME_FIELD,
                                                                     *(type_cls.HOOK_DEPENDENCIES or ()))
        # Create generator - note that this has to be a generator because params may be updated below
        generator = (e for e in type_cls.iter_all(api, params=params, pluck=pluck, page_size=config.api_max_page_size,
                                                  hook=self._context_hook))
        if starts_with:
            if type_cls.FILTERABLE:
                params.update(api.create_filter().like(type_cls.NAME_FIELD, starts_with + "%").to_params())
            else:
                generator = (e for e in generator if e[type_cls.NAME_FIELD].startswith(starts_with))
        # Return only names, not whole entities
        return (e[type_cls.NAME_FIELD] for e in generator)


class ContextBoundEntity(ContextSensitiveEntity, TypeChoicesHelpMixin):
    CONTEXT_SENSITIVE_CACHE = True
    CONTEXT_TYPES = []  # override in children with list of types to specify the argtype(s) the entity is bound to
    CONTEXT_PARAMS = []  # override in children with list of strings to specify param names the entity is bound to
    URI = ''  # override in children with format string containing {param[name].raw} or {arg[ArgType].actual} entries
    _PARAM_MATCH = re.compile(r'param\[([a-z_]+[a-z0-9_]*)]\.(raw|actual)', re.IGNORECASE)
    _ARG_MATCH = re.compile(r'arg\[([a-z_]+[a-z0-9_]*)]\.(raw|actual)', re.IGNORECASE)
    __RAW_CACHE = {}  # for context_changed to work
    SHOW_MISSING_CONTEXT = False  # show missing argTypes or params in description when invoked from auto-complete

    def should_render_choices(self):  # TypeChoicesMixin Hack: recognize and clear context when help command is called
        self._context_args = None
        return False

    def get_type_name_parts(self):  # Placeholder, not actually used, but required for TypeChoicesMixin hack.
        return (self.get_name(),)

    def get_all_possible_values(self, _api, starts_with=None, context_args=None):
        with self.cls_set_context(context_args):
            return super(ContextBoundEntity, self).get_all_possible_values(_api, starts_with, context_args)

    @classmethod
    def context_satisfied(cls, context=None):
        """ Returns True if all context dependencies are satisfied.
        :param context: dict in the form of ``{"param": {name: "param", type: ArgType, raw: str(value), cooked: Any}}"""
        return all([cls.cls_get_context(context, param_name=param) is not None for param in cls.CONTEXT_PARAMS] +
                   [cls.cls_get_context(context, argtype=arg) is not None for arg in cls.CONTEXT_TYPES])

    @classmethod
    def get_missing_context(cls, context=None):
        """ Returns list of missing context dependencies.
        :param context: dict in the form of ``{"param": {name: "param", type: ArgType, raw: str(value), cooked: Any}}"""
        return ([param for param in cls.CONTEXT_PARAMS if cls.cls_get_context(context, param_name=param) is None] +
                [arg.__name__ for arg in cls.CONTEXT_TYPES if cls.cls_get_context(context, argtype=arg) is None])

    @classmethod
    def get_arg(cls, name):
        """ Returns argType from CONTEXT_TYPES by name
        :param name: argType name"""
        return next(iter([arg for arg in cls.CONTEXT_TYPES if arg.__name__ == name]), None)

    @classmethod
    def context_changed(cls, _api, partial_value, context_args):
        """ Helper function for command line autocompletion. Returns True if relevant arguments have changed"""
        result = False
        for param in cls.CONTEXT_PARAMS:
            if cls.__RAW_CACHE.get(param) != get_context(context_args, param_name=param):
                result = True
                cls.__RAW_CACHE[param] = get_context(context_args, param_name=param)
        return result

    def get_description(self):
        missing = self.get_missing_context(self._context_args)
        if len(missing) == 0 or self._context_args is None:  # it is None only in help display, so return description
            return super(ContextBoundEntity, self).get_description()
        ctx = f"required arguments ({', '.join(missing)})" if self.SHOW_MISSING_CONTEXT else "other required arguments"
        return f"{self.get_entity_name()}: Please specify {ctx} first"

    @classmethod
    def get_uri(cls, *subresource):
        """ Returns correct URI enriched by context arguments/params or None if context is not satisfied """
        if not cls.context_satisfied():
            return None
        param = {}
        for par, kind in cls._PARAM_MATCH.findall(cls.URI):
            if par not in param:
                param[par] = Munch()
            param[par][kind] = cls.cls_get_context(param_name=par, raw=(kind == 'raw'))
        arg = {}
        for typename, kind in cls._ARG_MATCH.findall(cls.URI):
            if typename not in arg:
                arg[typename] = Munch()
            arg[typename][kind] = cls.cls_get_context(argtype=cls.get_arg(typename), raw=(kind == 'raw'))

        uri = cls.URI.format(param=param, arg=arg)
        if not subresource:
            return uri
        return "{}/{}".format(uri, "/".join(str(s) for s in subresource))

    @classmethod
    def _add_context(cls, results):
        """ Subclasses can implement this method, which is invoked during
        ``get_all()``. The method can add context variables to results or filter results
        based on context.
        It may use ``cls_get_context()`` helper function """
        return results

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        if not cls.context_satisfied():
            return []
        return cls._add_context(super(ContextBoundEntity, cls).get_all(api, *_args, **kwargs))

    @staticmethod
    def _get_raw_param_value(name, typedef, cooked):
        """ Try and regenerate original raw string from "cooked" parameter value.
        Uses infinishell heuristics in reverse.
        :param name: Argument name as defined in command signature. Required for name/id heuristic
        :param typedef: Argument type as defined in command decorator. Instance is expected, not class.
        :param cooked: Value received by the command function, as processed by infinishell.

        CAVEAT: list support is currently limited to lists of simple types.
        """
        if isinstance(typedef, Entity):  # obtain entity name or id field value using infinishell heuristics
            accessor = getattr(typedef, "ID_FIELD" if name.endswith("_id") else "NAME_FIELD")
            if hasattr(typedef, "UNIQUE_NAME") and typedef.UNIQUE_NAME is False:
                return str(cooked[0][accessor]) if len(cooked) else ""  # get the name of the first object
            else:
                return str(cooked[accessor])
        elif isinstance(typedef, MappedEnumType):
            mapping = typedef._get_mapping()  # obtain and reverse enum mapping
            mapping = {value: key for key, value in (mapping.items() if isinstance(mapping, dict) else mapping)}
            return mapping.get(cooked, str(cooked))
        elif isinstance(typedef, List):  # assume list of string-like entities. Don't dig further. #caveat
            return ",".join([str(item) for item in cooked])
        else:
            return str(cooked)

    @classmethod
    def with_context(cls, func):
        """ Use as an outermost decorator for EntityCommands commands when ENTITY_TYPE is context-bound.
        It will automatically deduce context variables (function parameters, their types, values and raw values)
        and wrap the command call in context manager that will ensure the entity has correct path and settings.
        :param func: Infinishell command function, an EntityCommands class member
        """
        @decorator  # ensure the wrapped function has the correct signature: infinishell relies on ``inspect``
        def wrapper(f, *args, **kw):
            arg_insts = [argtype() for argtype in f._command_metadata.arg_types]  # extract command ArgTypes
            f_args = inspect.getfullargspec(f).args  # extract command's parameter names
            argspec = zip(f_args[2:], args[2:], arg_insts)  # mesh all those with cooked values, except for self & _api
            context = {argname: {"name": argname,  # generate artificial context dict
                                 "type": type_inst,  # instance of ArgType is required for this, not the class
                                 "cooked": val,
                                 "raw": cls._get_raw_param_value(argname, type_inst, val)}  # regenerate raw string too
                       for argname, val, type_inst in argspec
                       if argname in cls.CONTEXT_PARAMS  # limit context to parameter names that are in CONTEXT_PARAMS
                       or isinstance(type_inst, tuple(cls.CONTEXT_TYPES))}  # or argtypes in CONTEXT_TYPES
            with cls.cls_set_context(context):  # set the entity's context and call the command function
                return f(*args, **kw)

        return wrapper(func)


class DDEBoundEntity(ContextSensitiveEntity):
    """Entities that only exist on a specific DDE and can be accessed only through specific DDE Api
    """
    CONTEXT_SENSITIVE_CACHE = True
    _CONTEXT_DDE = None
    DDE_INSTANCE_ID = None
    DEFAULT_DDES = (1, 2)

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_url(cls.DDE_INSTANCE_ID, path)

    def get_description(self):
        name = self.ENTITY_NAME
        return name if self._get_context(argtype=DDEApp, raw=True) else \
            f"{name}: please specify DDE argument first"

    def _context_hook(self, results, **kwargs):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if not dde:
            return results
        return [tgt for tgt in results if tgt.dde_instance_id == dde.dde_instance_id]

    @classmethod
    @contextmanager
    def patch_dde_entity(cls, dde):
        original = cls.DDE_INSTANCE_ID
        cls.DDE_INSTANCE_ID = dde.dde_instance_id
        try:
            yield
        finally:
            cls.DDE_INSTANCE_ID = original

    @staticmethod
    def filter_by_dde(dde, entities):
        return entities and next((ent for ent in entities if
                                 ent.dde_instance_id == dde.dde_instance_id), None)

    def _get_possible_values_by_search(self, api, starts_with, type_cls):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if dde:
            with self.patch_dde_entity(dde):
                return super(DDEBoundEntity, self)._get_possible_values_by_search(api, starts_with, type_cls)
        else:
            return super(DDEBoundEntity, self)._get_possible_values_by_search(api, starts_with, type_cls)

    def _get_possible_values_by_query(self, api, starts_with, type_cls):
        dde = self._get_context(argtype=DDEApp, raw=False)
        if dde:
            with self.patch_dde_entity(dde):
                return super(DDEBoundEntity, self)._get_possible_values_by_query(api, starts_with, type_cls)
        else:
            return super(DDEBoundEntity, self)._get_possible_values_by_query(api, starts_with, type_cls)

    # def _filtered(_self, func):
    #     # @wraps(func)
    #     def wrapper(self, _api, *args, **kwargs):
    #         print(args, kwargs.values())
    #         ddeinst = next((param for param in itertools.chain(args, kwargs.values()) if isinstance(param, DDEApp)), None)
    #         patchable = hasattr(self.__class__, 'ENTITY_TYPE') and self.__class__.ENTITY_TYPE
    #         filterable, fidx = next(((param, idx) for idx, param in itertools.chain(enumerate(args), kwargs.items())
    #                                 if isinstance(param, list) and isinstance(next(iter(param)), _self)), (None, -1))
    #         if filterable:
    #             filtered_var = next((f for f in filterable if f.dde_instance_id == ddeinst.dde_instance_id))
    #             if isinstance(fidx, int) and fidx < len(args):
    #                 args[fidx] = filtered_var
    #             else:
    #                 kwargs[fidx] = filtered_var
    #         if patchable:
    #             with _self._patch_dde_entity(patchable, ddeinst):
    #                 func(self, *args, **kwargs)
    #         else:
    #             func(self, _api, *args, **kwargs)
    #
    #     # wrapper._command_metadata = func._command_metadata
    #     return wrapper
    #
    # filtered = classmethod(_filtered)

    @classmethod
    def context_changed(cls, _api, partial_value, context_args):
        dde = get_context(context_args, argtype=DDEApp, raw=True)
        result = cls._CONTEXT_DDE != dde
        cls._CONTEXT_DDE = dde
        return result

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        alldde_entities = []
        inst_id = cls.DDE_INSTANCE_ID or \
                  (cls.cls_get_context(argtype=DDEApp, raw=False) or Munch(dde_instance_id=None)).dde_instance_id
        ddes = cls.DEFAULT_DDES if not inst_id else (inst_id, )
        for dde_id, entity_list in allddeget(api, cls.URI, ddes=ddes).items():
            if entity_list:
                for entity in entity_list:
                    alldde_entities.append(Munch(dde_instance_id=dde_id, **entity))

        if kwargs.get("hook"):
            alldde_entities = kwargs["hook"](alldde_entities)
        return alldde_entities

    @classmethod
    def get_by_name(cls, api, name):
        if not (name and name.strip()):
            raise NoSuchEntity(u"No such {}: name is empty".format(cls.get_entity_name()))
        entity = cls._fetch_entity_by_name(api, name)
        if entity is not None:
            cls._validate_fetched_entity(entity)
            return entity
        raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(cls.get_entity_name(), name))


@contextmanager
def patch_dde_entity(cls, dde: DDEApp):
    original = cls.ENTITY_TYPE.DDE_INSTANCE_ID
    cls.ENTITY_TYPE.DDE_INSTANCE_ID = dde.dde_instance_id
    try:
        yield
    finally:
        cls.ENTITY_TYPE.DDE_INSTANCE_ID = original
