from . import DDENode, ig_command
from .nodes import DdeNodeCommands
from .lib.context_entity import ContextSensitiveEntity, get_context
from infinishell.infishell import commandgroup
from infinishell.infishell.utils import strlist
from .. import EntityCommands, render
from ..options import QueryOptions
from ...exceptions import NoSuchEntity
from functools import partial
from contextlib import contextmanager
from munch import Munch


def get_dde_node_url(ddemgmtid, resource_url):
    return f'integrations/infiniguard/dde_nodes/{ddemgmtid}/{resource_url}'


class DDENodePort(ContextSensitiveEntity):
    DDE_MGMT_ID = None
    URI = "dde_ports"
    ENTITY_NAME = "DDE port"
    NAME_FIELD = "conn_port_comp_id"
    CASE_INSENSITIVE_NAME = True
    CONTEXT_SENSITIVE_CACHE = True
    _CONTEXT_DDE = None
    FILTERABLE = False
    PLUCKABLE = False
    SEARCHABLE = False
    UNIQUE_NAME = True
    PAGED = False

    @classmethod
    def get_uri(cls, *subresource):
        path = super().get_uri(*subresource)
        return get_dde_node_url(cls.DDE_MGMT_ID, path)

    def _context_hook(self, results, **kwargs):
        dde = self._get_context(argtype=DDENode, raw=False)
        if not dde:
            return results
        return [tgt for tgt in results if tgt.dde_node_mgmt_id == dde.id]

    @classmethod
    @contextmanager
    def patch_dde_entity(cls, dde):
        original = cls.DDE_MGMT_ID
        cls.DDE_MGMT_ID = dde.id
        try:
            yield
        finally:
            cls.DDE_MGMT_ID = original

    def _get_possible_values_by_search(self, api, starts_with, type_cls):
        dde = self._get_context(argtype=DDENode, raw=False)
        if dde:
            with self.patch_dde_entity(dde):
                return super(DDENodePort, self)._get_possible_values_by_search(api, starts_with, type_cls)
        else:
            return super(DDENodePort, self)._get_possible_values_by_search(api, starts_with, type_cls)

    def _get_possible_values_by_query(self, api, starts_with, type_cls):
        dde = self._get_context(argtype=DDENode, raw=False)
        if dde:
            with self.patch_dde_entity(dde):
                return super(DDENodePort, self)._get_possible_values_by_query(api, starts_with, type_cls)
        else:
            return super(DDENodePort, self)._get_possible_values_by_query(api, starts_with, type_cls)

    @classmethod
    def context_changed(cls, _api, partial_value, context_args):
        dde = get_context(context_args, argtype=DDENode, raw=True)
        result = cls._CONTEXT_DDE != dde
        cls._CONTEXT_DDE = dde
        return result

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        alldde_entities = []
        pluck = api.create_pluck(DDENode.ID_FIELD, DDENode.NAME_FIELD)
        ddes = DDENode.iter_all(api, pluck=pluck)
        for dde in ddes:
            with cls.patch_dde_entity(dde):
                entity_list = cls.query(api)
                for entity in entity_list:
                    alldde_entities.append(Munch(node_id=dde.node_id, **entity))

        if kwargs.get("hook"):
            alldde_entities = kwargs["hook"](alldde_entities)
        return alldde_entities

    @classmethod
    def get_by_name(cls, api, name):
        if not (name and name.strip()):
            raise NoSuchEntity(u"No such {}: name is empty".format(cls.get_entity_name()))
        entity = cls._fetch_entity_by_name(api, name)
        if entity is not None:
            cls._validate_fetched_entity(entity)
            return entity
        raise NoSuchEntity(u'Command rejected: No such {}: "{}"'.format(cls.get_entity_name(), name))


QUERY_FIELDS = strlist("conn_port_comp_id node_id port_num media_type type state link_state hw_addr connection_speed")
DETAILED_FIELDS = strlist(
    "id conn_port_comp_id node_id port_num label media_type type state link_state hw_addr connection_speed \
    created_at updated_at")
SUBSTITUTE = dict(
    node_id='dde_node_id',
    port_num='port_number',
    id='internal_id',
    conn_port_comp_id='name',
    created_at='created',
    updated_at='updated'
)


@commandgroup(prefix="port", parent=DdeNodeCommands)
class DdePortCommands(EntityCommands):
    ENTITY_TYPE = DDENodePort

    @ig_command(
        description="Display DDE ports",
        types=(DDENode, DDENodePort,),
        options=QueryOptions(fields=DETAILED_FIELDS,
                             substitute=SUBSTITUTE, detailed=True)
    )
    @render(
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_FIELDS,
        substitute=SUBSTITUTE,
        detailed_individually=True
    )
    def query(self, _api, dde=None, port=None):
        return self._query(_api, port, dde_node_mgmt_id=dde, _hook=partial(self._query_hook))

    def _query_hook(self, ports):
        return ports
