from . import DDEApp, DDEActiveApp, ig_command
from infinishell.infishell import commandgroup, get_active_shell
from infinishell.infishell.types import Str, ChoiceType, RegexType
from .lib.types import Password
from .nodes import DdeAppCommands
from .cifs import CifsCommands
from .lib.context_entity import ContextBoundEntity, filtered
from .. import set_renderer_query_fields, EntityCommands, render, ENTITY, roles
from collections import OrderedDict
from functools import partial
from ..options import QueryOptions
from .share import ShareCommands, PermissionType
from infinishell.infinibox.exceptions import ManagementException, NotFound
from infinishell.infishell.exceptions import BadArguments
from munch import Munch
from infinishell.utils import NamedOrderedDict, get_username_password
import re


class UserAccessibleShare(ContextBoundEntity):
    ENTITY_NAME = "Share"
    DESCRIPTION = "CIFS or App-Specific Share"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "name"
    UNIQUE_NAME = True
    NAME_FIELD = "name"
    CONTEXT_TYPES = [DDEApp, ]
    CONTEXT_PARAMS = []
    URI = "../dde/{arg[DDEApp].actual.dde_instance_id}/api/nas/shares"
    MULTI_FIELD_PUT = True
    FILTERABLE = False

    @classmethod
    def _add_context(cls, results):
        dde = cls.cls_get_context(argtype=DDEApp)
        return [Munch(ddeapp=dde,
                      _user_path="nas/cifs/workgroup_users" if entry.protocol == 'cifs' else "backup_app_user",  # magic
                      **entry) for entry in results if entry.protocol != 'nfs']


class ShareAccessUser(ContextBoundEntity):
    ENTITY_NAME = "User"
    DESCRIPTION = "CIFS Workgroup or Backup User"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "user"
    UNIQUE_NAME = True
    NAME_FIELD = "user"
    CONTEXT_TYPES = [DDEApp, UserAccessibleShare]
    URI = "../dde/{arg[DDEApp].actual.dde_instance_id}/api/nas/{arg[UserAccessibleShare].actual.protocol}/shares" \
          "/{arg[UserAccessibleShare].raw}/share_users"
    MULTI_FIELD_PUT = True
    FILTERABLE = False
    HOOK_DEPENDENCIES = ("user",)
    _ERR_RE = re.compile(r'E1000011|E1000062|CIFS_SERVER_IS_DISABLED', re.IGNORECASE)

    @classmethod
    def _add_context(cls, results):
        dde = cls.cls_get_context(argtype=DDEApp)
        share = cls.cls_get_context(param_name='share')
        return [Munch(ddeapp=dde, share=share, **entry) for entry in results]

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        try:
            return super(ShareAccessUser, cls).get_all(api, *_args, **kwargs)
        except ManagementException as e:
            if cls._ERR_RE.search(str(e)):
                return []
            raise

    def get_description(self):
        missing = self.get_missing_context(self._context_args)
        if len(missing) == 0:  # it is None only in help display, so return description
            share = self.cls_get_context(self._context_args, param_name='share', raw=False)
            if share and hasattr(share, "protocol"):
                return f"{'CIFS Workgroup' if share.protocol == 'cifs' else 'Backup'} User"
        return super(ShareAccessUser, self).get_description()


class AccessListUser(ShareAccessUser):
    ENTITY_NAME = "User"
    DESCRIPTION = "CIFS Workgroup or Backup User"
    URI = "../dde/{arg[DDEApp].actual.dde_instance_id}/api/{arg[UserAccessibleShare].actual._user_path}"
    ID_FIELD = "username"
    NAME_FIELD = "username"
    HOOK_DEPENDENCIES = ("username",)


class BackupAppUser(ContextBoundEntity):
    ENTITY_NAME = "Backup app user"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    ID_FIELD = "username"
    UNIQUE_NAME = True
    NAME_FIELD = "username"
    CONTEXT_TYPES = [DDEApp, ]
    URI = "../dde/{arg[DDEApp].actual.dde_instance_id}/api/backup_app_user"
    FILTERABLE = False

    @classmethod
    def _add_context(cls, results):
        dde = cls.cls_get_context(argtype=DDEApp)
        return [Munch(ddeapp=dde, **entry) for entry in results]


class CifsWorkgroupUser(ContextBoundEntity):
    ENTITY_NAME = "wg_user"
    DESCRIPTION = "Workgroup User"
    PAGED = False
    PLUCKABLE = False
    SEARCHABLE = False
    FILTERABLE = False
    ID_FIELD = "username"
    UNIQUE_NAME = True
    NAME_FIELD = "username"
    CONTEXT_TYPES = [DDEApp, ]
    URI = "../dde/{arg[DDEApp].actual.dde_instance_id}/api/nas/cifs/workgroup_users"
    MULTI_FIELD_PUT = True

    @classmethod
    def _add_context(cls, results):
        dde = cls.cls_get_context(argtype=DDEApp)
        return [Munch(ddeapp=dde, **entry) for entry in results]


BACKUP_USER_FIELDS = OrderedDict(
    ddeapp='DDE App',
    username='username',
    description='description',
)

APP_ID_TO_NAME = {
    1: 'a',
    2: 'b',
    3: 'standby'
}


@commandgroup(prefix="backupuser", parent=DdeAppCommands)
class BackupUserCommands(EntityCommands):
    ENTITY_TYPE = BackupAppUser

    @BackupAppUser.with_context
    @ig_command(
        description="Display all users associated with Application Specific shares, OST Storage Servers, "
                    "and PTT Backup applications.",
        types=(DDEActiveApp,),
        options=QueryOptions(fields=BACKUP_USER_FIELDS.keys(),
                             substitute=BACKUP_USER_FIELDS,
                             detailed=True, ),
    )
    @render(
        fields=list(BACKUP_USER_FIELDS.keys())[1:],
        detailed_fields=BACKUP_USER_FIELDS.keys(),
        substitute=BACKUP_USER_FIELDS,
    )
    def query(self, _api, dde):
        return self._query(_api)

    @BackupAppUser.with_context
    @ig_command(
        description="Add a backup application user",
        types=(DDEActiveApp, Str, Password, Str)
    )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, dde, name, password, description=None):
        if not password:
            _, password = get_username_password(username=name, prompt=None)
        create_args = Munch(username=name, password=password, description=description)
        self._create(_api, returns=ENTITY, **create_args)
        return f"Backup application user '{name}' was created successfully"

    @BackupAppUser.with_context
    @ig_command(
        description="Edit a backup application user",
        types=(DDEActiveApp, BackupAppUser, Password, Str)
    )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, dde, name, password=None, description=None):
        update_args = Munch()
        if password is not None:
            update_args.update(password=password)
        if description is not None:
            update_args.update(description=description)
        if not update_args:
            raise BadArguments('No changes specified. User not modified.')
        self._update(_api, name, _multi_field_put=True, **update_args)
        return f"Backup application user '{name.username}' was modified successfully"

    @BackupAppUser.with_context
    @ig_command(
        description="Delete a backup application user",
        types=(DDEActiveApp, BackupAppUser)
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, dde, name):
        self._delete(_api, name)
        return f"Backup application user '{name.username}' was deleted successfully"


SHARE_USER_FIELDS = OrderedDict(
    ddeapp='DDE App',
    share='Share',
    user='username',
    permissions='permissions',
    effective_perms='effective permissions'
)


def ensure_cifs(_api, dde):
    try:
        status = _api.get(f"../dde/{dde.dde_instance_id}/api/nas/cifs/info")
    except NotFound as n:
        raise ManagementException(404, Munch(code="CIFS_UNKNOWN", message="CIFS/SMB status not available"))
    if status.domain_type == "Workgroup":
        if not status.enabled:
            raise ManagementException(400, Munch(code="CIFS_DISABLED",
                                                 message="CIFS server is disabled."))
        return True
    elif status.domain_type == "Active Directory":
        raise ManagementException(400, Munch(code="CIFS_ACTIVE_DIRECTORY",
                                             message="This server is a member of Active Directory. "
                                                     "Use Microsoft management console to control user access."))
    else:
        raise ManagementException(400, Munch(code="CIFS_NO_WORKGROUP",
                                             message="This server is not joined to a workgroup."))


def filter_grant_params(params, context):
    share = context(argtype=UserAccessibleShare, raw=False)
    if share and share.protocol == 'app_specific':
        return [arg for arg in params if arg.name != 'permissions']
    return params


@commandgroup(prefix="user_access", parent=ShareCommands)
class UserAccessCommands(EntityCommands):
    ENTITY_TYPE = ShareAccessUser

    @ShareAccessUser.with_context
    @ig_command(
        description="List Share Users",
        types=(DDEActiveApp, UserAccessibleShare),
        options=QueryOptions(
            fields=SHARE_USER_FIELDS.keys(),
            substitute=SHARE_USER_FIELDS,
            detailed=True,
        ),
    )
    @render(
        fields=list(SHARE_USER_FIELDS.keys())[2:],
        detailed_fields=SHARE_USER_FIELDS.keys(),
        substitute=SHARE_USER_FIELDS,
        detailed_individually=True,
    )
    def query(self, _api, dde, share):
        if share.protocol == "cifs":
            ensure_cifs(_api, dde)
            result = [u for u in self._query(_api)]
            if not result:
                result = "All share users allowed to access this share\nAccess rights = rw\nEffective rights = rw"
            return result

    @ShareAccessUser.with_context
    @ig_command(
        description="Remove a workgroup user's right to access the specified CIFS or App-Specific share. \n \n"
                    "NOTE: When the last workgroup user is deleted from a CIFS share, the share access list of the "
                    "specified share is empty, which means the share now allows read-write access to all workgroup "
                    "users. \nIt is worth noting that the effective access rights depend on share settings (rw/ro), "
                    "so rw access is not guaranteed. The effective access rights depend on the share access mode. \n"
                    "If the share is read-only, all users can have read-only access regardless of their settings.",
        types=(DDEActiveApp, UserAccessibleShare, ShareAccessUser)
    )
    @roles.INFINIDAT.ADMIN
    def revoke(self, _api, dde, share, user):
        if share.protocol == "cifs":
            ensure_cifs(_api, dde)
        self._delete(_api, user)
        return f"Access rights for user '{user.user}' were revoked successfully"

    @ShareAccessUser.with_context
    @filtered(filter_params=filter_grant_params)
    @ig_command(
        description="Grant an existing workgroup user the right to access the specified CIFS or App-Specific share.",
        types=(DDEActiveApp, UserAccessibleShare, AccessListUser, PermissionType)
    )
    @roles.INFINIDAT.ADMIN
    def grant(self, _api, dde, share, user, permissions=None):
        create_args = Munch(user=user.username)
        if share.protocol == "cifs":
            ensure_cifs(_api, dde)
            if permissions is None:
                permissions = "rw"  # default for cifs, irrelevant for app-specific
            create_args["permissions"] = permissions
        else:
            if permissions == "ro":
                get_active_shell().warning("User permissions for app-specific share are always 'rw'")
        result = self._create(_api, returns=ENTITY, **create_args)
        return f"Access rights ({result.effective_perms}) to user '{user.username}' were granted successfully"


WORKGROUP_USER_FIELDS = OrderedDict(
    ddeapp='DDE App',
    username='username',
    description='description',
    role='role',
    admin='admin'
)


class CifsUserRole(ChoiceType):
    ROLES = NamedOrderedDict.constants("wguser", "wgadmin")

    def _get_values(self):
        return self.ROLES.keys()


class CifsUserName(RegexType):
    def _get_pattern(self):
        return r'^[a-zA-Z0-9_]+$'

    def _get_error(self):
        return "must contain one or more alphanumeric characters or '_'"

    def get_description(self):
        return "Workgroup Username"


class CifsUserDescription(CifsUserName):
    def get_description(self):
        return "Workgroup User Description"


@commandgroup(prefix="user", parent=CifsCommands)
class WorkgroupUserCommands(EntityCommands):
    ENTITY_TYPE = CifsWorkgroupUser

    def _query_hook(self, users, role):
        return [user for user in users if role is None or user.role == role]

    @CifsWorkgroupUser.with_context
    @ig_command(
        description="Display a List of Workgroup Users",
        types=(DDEActiveApp, CifsWorkgroupUser, CifsUserRole),
        options=QueryOptions(
            fields=WORKGROUP_USER_FIELDS.keys(),
            substitute=WORKGROUP_USER_FIELDS,
            detailed=True,
        ),
    )
    @render(
        fields=list(WORKGROUP_USER_FIELDS.keys())[1:],
        detailed_fields=WORKGROUP_USER_FIELDS.keys(),
        substitute=WORKGROUP_USER_FIELDS,
        detailed_individually=True,
    )
    def query(self, _api, dde, username=None, role=None):
        ensure_cifs(_api, dde)
        if username:
            set_renderer_query_fields()
            return [username] if role is None or username.role == role else []
        return self._query(_api, _hook=partial(self._query_hook, role=role))

    @CifsWorkgroupUser.with_context
    @ig_command(
        description="Add a workgroup user account",
        types=(DDEActiveApp, CifsUserName, Password, CifsUserRole, CifsUserDescription),
    )
    @roles.INFINIDAT.ADMIN
    def add(self, _api, dde, username, password, role, description=None):
        ensure_cifs(_api, dde)
        create_args = Munch(username=username, password=password, role=role, description=description)
        self._create(_api, returns=ENTITY, **create_args)
        return f"User account '{username}' was successfully added to the workgroup with role '{role}'"

    @CifsWorkgroupUser.with_context
    @ig_command(
        description="Modify a workgroup user account",
        types=(DDEActiveApp, CifsWorkgroupUser, Password, CifsUserDescription),
    )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, dde, username, password=None, description=None):
        if password is not None or description is not None:
            modify_args = Munch()
            if password is not None:
                modify_args["password"] = password
            if description is not None:
                modify_args["description"] = description
            self._update(_api, username, returns=ENTITY, **modify_args)
            return f"User account '{username.username}' was successfully modified"
        else:
            raise BadArguments("No changes specified. User not modified.")

    @CifsWorkgroupUser.with_context
    @ig_command(
        description="Delete a workgroup user account",
        types=(DDEActiveApp, CifsWorkgroupUser),
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, dde, username):
        ensure_cifs(_api, dde)
        self._delete(_api, username)
        return f"User account '{username.username}' was deleted successfully"
