from infinishell.infinibox.features import feature_command
from ..types import Entity
from ...api import QuerySort
from infinishell.infishell import BadValue
from logging import getLogger

log = getLogger(__name__)


def ig_command(**dekwargs):
    def ig_decorator(func, *args, **kwargs):
        dekwargs.update({'feature': 'iguard', 'status': True})
        return feature_command(**dekwargs)(func, *args, **kwargs)

    return ig_decorator


class DDEApp(Entity):
    SEARCHABLE = False
    PAGED = True
    PLUCKABLE = True
    ID_FIELD = "dde_instance_id"
    NAME_FIELD = "dde_instance_name"
    HOOK_DEPENDENCIES = ("dde_instance_id", "state")
    URI = "integrations/infiniguard/dde_nodes"
    QUERY_SORT = QuerySort().asc("dde_instance_id")

    @staticmethod
    def get_enabled_apps():
        return ["a", "b", "standby"]

    @classmethod
    def get_entity_name(cls):
        return "DDE App"

    @classmethod
    def get_description(cls):
        return "DDE App online"

    @classmethod
    def _query_hook(cls, nodes):
        def update_node(node):
            node[cls.NAME_FIELD] = ["a", "b", "standby"][node[cls.ID_FIELD] - 1]
            return node

        def _enabled(node):
            return node[cls.NAME_FIELD] in cls.get_enabled_apps() and \
                node.state.lifecycle in ['ACTIVE', 'FRU_STANDBY']

        return [updated_node for node in nodes
                if _enabled(updated_node := update_node(node))]

    @classmethod
    def _fetch_entity_by_name(cls, api, name):
        try:
            index = cls.get_enabled_apps().index(name.lower())
        except ValueError:
            log.exception(f'Cannot find App named {name}')
            return None
        _id = index + 1
        if not _id:
            return None
        fltr = api.create_filter().eq(cls.ID_FIELD, _id) if cls.FILTERABLE else None
        entities = cls.iter_all(api, filter=fltr)
        if entities is not None:
            if cls.CASE_INSENSITIVE_NAME:
                lower_name = name.lower()

                def predicate(entity):
                    entity_name = entity[cls.NAME_FIELD]
                    return entity_name and entity_name.lower() == lower_name
            else:
                def predicate(entity):
                    return str(entity[cls.NAME_FIELD]) == name
            if cls.UNIQUE_NAME:
                for entity in entities:
                    if predicate(entity):
                        return entity
            else:
                entities = [entity for entity in entities if predicate(entity)]
                if entities:
                    return entities
        return None


class DDEActiveApp(DDEApp):
    @staticmethod
    def get_enabled_apps():
        return ["a", "b"]

    @classmethod
    def get_entity_name(cls):
        return "active DDE App"

    @classmethod
    def get_description(cls):
        return "active DDE App online"


class DDENode(Entity):
    SEARCHABLE = False
    PAGED = True
    PLUCKABLE = True
    ID_FIELD = "id"
    NAME_FIELD = "node_id"
    HOOK_DEPENDENCIES = ("dde_instance_id",)
    URI = "integrations/infiniguard/dde_nodes"
    QUERY_SORT = QuerySort().asc("node_id")
    MIN = 1
    MAX = 3

    @classmethod
    def get_description(cls):
        return "Represents a physical DDE node with possible IDs (1, 2, or 3)"

    @classmethod
    def _query_hook(cls, nodes):
        for node in nodes:
            node["dde_instance_name"] = ["a", "b", "standby"][node["dde_instance_id"] - 1]
        return nodes

    def _verify_raw(self, value):
        message = f'DDE Node ID must be a number between {self.MIN} and {self.MAX}'
        if not value.isdecimal():
            return message

        if not int(value) in range(self.MIN, self.MAX + 1):
            return message

    def verify_raw(self, arg, value):
        error_message = self._verify_raw(value)
        if error_message:
            raise BadValue(error_message)
