import re
from functools import partial
from infinishell.infinibox.commands import EntityCommands, roles, Null
from infinishell.infinibox.commands.infiniguard import ig_command, DDEApp
from infinishell.infinibox.commands.options import QueryOptions
from infinishell.infishell import commandgroup
from infinishell.infishell.types import EnumType, Netmask, IPAddress, MappedEnumType, Range
from infinishell.infishell.utils import strlist
from infinishell.infinibox.commands.infiniguard.lib.types import IPv4AddressNullable
from .lib.context_entity import filtered, patch_dde_entity
from .nettools import NetworkCommands, DdeNetworkDevice, DDEBondMember, DdeNetworkInterface, DdeNetworkDeviceStaticBonds
from .. import render

QUERY_FIELDS = strlist("devname intfname vlan_id ext_dde_ip ip_address netmask gateway nat traffic_type")
SUBSTITUTE_FIELDS = dict(
    devname='Device Name',
    intfname='Interface',
    ext_dde_ip='DDE External IP',
    ip_address='Interface IP',
    nat='NAT IP',
    ddeapp='DDE App',
    vlan_id='VLAN ID'
)
DETAILED_QUERY_FIELDS = strlist(
    "ddeapp devname intfname vlan_id ext_dde_ip ip_address netmask gateway nat traffic_type")


APP_ID_TO_NAME = {
    1: 'a',
    2: 'b',
    3: 'standby'
}


class InterfaceType(EnumType):
    def _get_values(self):
        return 'port', 'bond'


class Mode(MappedEnumType):
    def _get_mapping(self):
        return {
            'Round Robin': 'RR',
            'Active Backup': 'AB',
            'LACP': 'LACP'
        }


class YesChoice(MappedEnumType):
    def _verify_raw(self, value):
        if not value.upper() == 'YES':
            return "can only accept the value 'YES'"

    def _get_mapping(self):
        return {'YES': True}


class MTU(Range):
    MIN = 576
    MAX = 9000


class VLAN(Range):
    MIN = 2
    MAX = 4094


class TrafficType(EnumType):
    def _get_values(self):
        return 'any', 'data', 'replication'


def _filter_bond_params(params, context):
    interface_type = context(param_name='interface_type', raw=False)

    if interface_type != 'bond':
        return [arg for arg in params if arg.name not in ('mode', 'members')]
    else:
        return params


@commandgroup(prefix="interface", parent=NetworkCommands)
class NetworkInterfaceCommands(EntityCommands):
    ENTITY_TYPE = DdeNetworkInterface

    @staticmethod
    def _filter_by_dde(dde, devices):
        """
        Even though the autocomplete is performed after filtering on the dde instance,
        the selected devname from all ddes are passed to the command as a list,
        and must be filtered again for the devname from the correct DDE.

        In case of bond devices, we use a static list and therefore don't need to filter. In this case the device's
        dde instance with be None.
        """
        return devices and next((device for device in devices if
                                 device.dde_instance_id is None or device.dde_instance_id == dde.dde_instance_id), None)

    @ig_command(
        description="Create an interface on a port or bond device",
        types=(DDEApp, InterfaceType, DdeNetworkDeviceStaticBonds, IPAddress, Netmask, IPAddress,
               DDEBondMember.List, Mode, VLAN, TrafficType.List, IPAddress, MTU)
    )
    @filtered(
        filter_params=_filter_bond_params
    )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, dde, interface_type, devname, ip_address, netmask, gateway, members=None, mode=None,
               vlan=None, traffic_type=None, nat=None, mtu=None):
        devname = self._filter_by_dde(dde, devname)
        with patch_dde_entity(self.__class__, dde):
            return self._create(_api,
                                devname=devname.name,
                                interface_type=interface_type,
                                ip_address=ip_address.ip,
                                netmask=netmask.netmask,
                                gateway=gateway.ip,
                                members=members and [member.name for member in members],
                                mode=mode,
                                vlan=vlan,
                                traffic_type=traffic_type,
                                nat=nat and nat.ip,
                                mtu=mtu,
                                )

    @ig_command(
        description="Delete an interface on a port or bond device",
        types=(DDEApp, DdeNetworkInterface)
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, dde, interface):
        interface = self._filter_by_dde(dde, interface)
        with patch_dde_entity(self.__class__, dde):
            self._delete(_api, interface)

    @ig_command(
        description="Modify NAT of an interface",
        types=(DDEApp, DdeNetworkInterface, IPv4AddressNullable)
    )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, dde, interface, nat):
        interface = self._filter_by_dde(dde, interface)
        with patch_dde_entity(self.__class__, dde):
            self._update(_api,
                         interface,
                         nat=Null if nat == 'none' or nat == '' else nat)

    @ig_command(
        description="Set an interface to be the external IP of the DDE",
        types=(DDEApp, DdeNetworkInterface)
    )
    @roles.INFINIDAT.ADMIN
    def set_ext_dde_ip(self, _api, dde, interface):
        interface = self._filter_by_dde(dde, interface)
        with patch_dde_entity(self.__class__, dde):
            self._update(_api,
                         interface,
                         ext_dde_ip=True)

    def _query_hook(self, intfs):
        for intf in intfs:
            intf.ddeapp = APP_ID_TO_NAME[intf.dde_instance_id]
            if 'nat' not in intf:
                intf.nat = ''

            vlan_id = None
            # Match with VLAN interface - p4p2.104:1
            match = re.match(r'^(\w+)\.(?P<vlan>\d+)\:(\d+)$', intf.intfname)
            if match:
                vlan_id = match.group('vlan')
            intf.vlan_id = vlan_id
            yield intf

    @ig_command(
        description="List interfaces",
        types=(DDEApp, DdeNetworkInterface, DdeNetworkDevice),
        options=QueryOptions(fields=QUERY_FIELDS, substitute=SUBSTITUTE_FIELDS, detailed=True)
    )
    @render(fields=QUERY_FIELDS,
            substitute=SUBSTITUTE_FIELDS,
            detailed_fields=DETAILED_QUERY_FIELDS,
            detailed_individually=True)
    def query(self, _api, dde, interface=None, devname=None):
        interface = self._filter_by_dde(dde, interface)
        device = self._filter_by_dde(dde, devname)
        with patch_dde_entity(self.__class__, dde):
            return self._query(_api, _hook=partial(self._query_hook),
                               intfname=interface and interface.intfname,
                               devname=device and device.name,
                               dde_instance_id=dde.dde_instance_id)
