from infinishell.infinibox.commands.types import Entity
from infinishell.infinibox.commands.infiniguard.lib.dde import allddeget, ddeget
import re
from infinishell.infinibox.exceptions import NoSuchEntity, ManagementException
from infinishell.infishell.options import FlagType
from typing import Dict, List, Optional
from infinishell.infishell.shell import get_active_shell
from infinishell.infinibox.commands.utils import is_option_set
from itertools import chain


class Task(Entity):
    DDEURI = 'mgmt/tasks'
    SEARCHABLE = False
    PAGED = False
    PLUCKABLE = False
    FILTERABLE = True
    SORTABLE = True
    ID_FIELD = 'taskid'
    NAME_FIELD = "id"
    NAME_PATTERN = re.compile(r"^([abs])(\d+)$", re.I)
    RUNNING_STATUS = 'running'

    @classmethod
    def get_uri(cls, *subresource):
        raise NotImplementedError

    @classmethod
    def get_all(cls, api, *_args, **kwargs):
        res: Dict[int, Optional[List]] = allddeget(api, cls.DDEURI)

        unique_tasks = {(task['controller'], task[cls.ID_FIELD]): cls._entity_hook(task)
                        for task in
                        chain.from_iterable(task_list
                                            for task_list in res.values() if task_list)}
        tasks = [task for _id, task in sorted(unique_tasks.items())]

        if kwargs.get("hook"):
            tasks = kwargs["hook"](tasks)
        return tasks

    @classmethod
    def _entity_hook(cls, task):
        task[cls.NAME_FIELD] = cls.build_name(task)
        return task

    @classmethod
    def iter_all(cls, api, *args, **kwargs):
        return iter(cls.get_all(api, *args, **kwargs))

    @classmethod
    def get_by_name(cls, api, name):
        match = cls.NAME_PATTERN.match(name)
        if not match:
            raise NoSuchEntity('"{}" is not a valid {} name'.format(name, cls.get_entity_name()))
        appletter, taskid = match.groups()
        try:
            return cls._get_entity_by_ids(api, appletter, taskid)
        except ManagementException as e:
            if e.http_code != 404:
                raise
            raise NoSuchEntity('No such {}: "{}"'.format(cls.get_entity_name(), name))

    DDEAPPLETTERS = ' ABS'

    @classmethod
    def _get_entity_by_ids(cls, api, appletter, taskid):
        ddeid = cls.DDEAPPLETTERS.index(appletter.upper())
        entity = ddeget(api, ddeid, f'{cls.DDEURI}/{taskid}')
        entity = cls._entity_hook(entity)
        cls._validate_fetched_entity(entity)
        return entity

    @classmethod
    def build_name(cls, task):
        prefix = task['controller'][0].upper()
        suffix = task[cls.ID_FIELD]
        return f'{prefix}{suffix}'

    @classmethod
    def ddeapp(cls, ddeid):
        return cls.DDEAPPLETTERS[ddeid]


class NoninteractiveOption(FlagType):
    NAMES = ("ni", "noninteractive")

    def is_hidden_from_help(self):
        return False

    def get_description(self):
        return "Non-blocking mode, start the command and return immediately"


def report_started_task(ddeid, started_task, description):
    started_task['controller'] = Task.ddeapp(ddeid)
    watch_command = f'infiniguard.task.query'
    out = f'{description}. To monitor the progress run command "{watch_command}"\n'
    get_active_shell().output(out)
