from functools import partial
from infinishell.infinibox.features import feature_status
from munch import Munch
from infinishell.infishell import commandgroup
from ..types import (
    AuthUrlLink,
    IdpName,
    Idps,
    Issuer,
    MappableUserRole,
    SSOCertificate,
    UserGroup,
    UserGroupName,
)
from .. import EntityCommands, LazyIdNameMapping
from . import ConfigCommandsBase
from infinishell.infinibox.commands.options import QueryOptions

from infinishell.infishell import commandgroup
from infinishell.infishell.commands.commands import command, render
from infinishell.infishell.exceptions import BadArguments
from infinishell.infishell.utils import strlist
from infinishell.infishell.types import (
    YesNo,
)
from .. import roles, EntityCommands


@commandgroup(parent=ConfigCommandsBase, prefix="sso", predicate=feature_status(feature="sso", enabled=True))
class SSOCommands(ConfigCommandsBase):
    pass


@commandgroup(parent=SSOCommands, prefix="idps", predicate=feature_status(feature="sso", enabled=True))
class SSOIdpsCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = Idps

    @command(
        description="Define an Identity Provider (IdP) that stores and manages users' digital identities.",
        types=(Issuer, IdpName, AuthUrlLink, YesNo, YesNo, YesNo, SSOCertificate),
    )
    @roles.INFINIDAT.ADMIN
    def define(self, _api, issuer, name, sign_on_url, enabled=None, signed_response=None, signed_assertion=None, signing_certificate=None):
        data = Munch(
            name=name,
            issuer=issuer,
            sign_on_url=sign_on_url,
            enabled=enabled,
            signed_response=signed_response,
            signed_assertion=signed_assertion,
            signing_certificate=signing_certificate,
        )
        data = {key: value for key, value in data.items() if value is not None}
        return self._create(_api, **data)
        

    @command(
        description="Modify the details of an Identity Provider (IdP) that stores and manages users' digital identities.",
        types=(Idps, IdpName, AuthUrlLink, YesNo, YesNo, YesNo, SSOCertificate),
    )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, idp, name=None, sign_on_url=None, enabled=None, signed_response=None, signed_assertion=None, signing_certificate=None):
        if all(arg is None for arg in (name, sign_on_url, enabled, signed_response, signed_assertion, signing_certificate)):
            raise BadArguments("At least one argument must be specified")
        data = Munch(
            name=name,
            sign_on_url=sign_on_url,
            enabled=enabled,
            signed_response=signed_response,
            signed_assertion=signed_assertion,
            signing_certificate=signing_certificate,
        )
        data = {key: value for key, value in data.items() if value is not None}
        return self._update(_api, idp, _multi_field_put=True, **data)
        

    @command(
        description="Remove an Identity Provider (IdP) so that it no can longer store and manage users' digital identities.",
        types=(Idps,),
    )
    @roles.INFINIDAT.ADMIN
    def remove(self, _api, idp):
        return self._delete(_api, idp)
        

    QUERY_FIELDS = strlist("""name enabled issuer sign_on_url signed_response signed_assertion signing_certificate_serial signing_certificate_expiry""")
    SORTABLE_FIELDS = strlist("name enabled issuer sign_on_url signed_response signed_assertion")
    DETAILED_FIELDS = QUERY_FIELDS + strlist("signing_certificate")

    @command(
        description="List the Identity Providers (IdP) that store and manage users' digital identities.",
        types=(Idps, Issuer, AuthUrlLink, YesNo, YesNo, YesNo),
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=QUERY_FIELDS,
                             sortable=SORTABLE_FIELDS,  detailed=True),
    )
    @render(
        fields=QUERY_FIELDS,
        detailed_individually=True,
        detailed_fields=DETAILED_FIELDS,
    )
    def query(self, _api, idp=None, issuer=None, sign_on_url=None, enabled=None, signed_response=None, signed_assertion=None):
        return self._query(_api, idp, issuer=issuer, sign_on_url=sign_on_url, enabled=enabled, signed_response=signed_response, signed_assertion=signed_assertion)


@commandgroup(parent=SSOCommands, prefix="user_group", predicate=feature_status(feature="sso", enabled=True))
class SSOIdpsCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = UserGroup

    @command(
        description="Create a user group for an Identity Provider (IdP) to manage the digital identities of users that have the same role.",
        types=(Idps, UserGroupName, MappableUserRole),
    )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, idp, name, role):
        data = Munch(
            name=name,
            role=role,
            type="Sso",
            idp_id=idp.id,
        )
        data = {key: value for key, value in data.items() if value is not None}
        return self._create(_api, **data)

    @command(
        description="Modify a user group used by an Identity Provider (IdP) to manage the digital identities of users that have the same role.",
        types=(UserGroup, UserGroupName, MappableUserRole),
    )
    @roles.INFINIDAT.ADMIN
    def modify(self, _api, group, name=None, role=None):
        if all(arg is None for arg in (name, role)):
            raise BadArguments("At least one argument must be specified")
        data = Munch(
            name=name,
            role=role,
        )
        data = {key: value for key, value in data.items() if value is not None}
        return self._update(_api, group, **data)

    @command(
        description="Delete a user group used by an Identity Provider (IdP) to manage the digital identities of users that have the same role.",
        types=(UserGroup,),
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, group):
        return self._delete(_api, group)
        

    QUERY_FIELDS = strlist("""
        name role idp
        """)
    SORTABLE_FIELDS = strlist("name role")
    
    @command(
        description="List the user groups used by Identity Providers (IdPs) to manage the digital identities of users that have the same role.",
        types=(UserGroup, MappableUserRole),
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=QUERY_FIELDS,
                             sortable=SORTABLE_FIELDS,  detailed=True),
    )
    @render(
        fields=QUERY_FIELDS,
        detailed_individually=True,
        detailed_fields=QUERY_FIELDS,
    )
    def query(self, _api, group=None, role=None):
        hook = partial(self._query_hook,
                       idp_name_by_id=LazyIdNameMapping(Idps, _api),
                       )
        return self._query(_api, group, role=role, _hook=hook)

    def _query_hook(self, groups, idp_name_by_id):
        for g in groups:
            g.idp = idp_name_by_id[g.idp_id]
        return groups
