
from collections import OrderedDict
from infinishell.infinibox.commands import EntityCommands, roles
from infinishell.infinibox.commands.options import QueryOptions
from infinishell.infinibox.features import feature_command
from infinishell.infishell import commandgroup, command
from infinishell.infishell.commands.commands import render
from infinishell.infishell.types import YesNo
from infinishell.infishell.utils.python import strlist
from .types import SSAExpress, Gigabyte, SSDDrives


@commandgroup(prefix="ssa_express")
class SSAExpress(EntityCommands):
    ENTITY_TYPE = SSAExpress

    @feature_command(
        description="For SSA Express, report the status, total capacity, free capacity, used capacity, and the number of active entities",
        types=(),
        feature='ssa_express',
        status=True,
    )
    @render(
        individually=True,
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN.READONLY
    def info(self, _api,):
        ssa_express_info = _api.get('system/ssa_express_info')
        entity_count = _api.get('system/entity_counts')
        
        result = OrderedDict((
            ("State", ssa_express_info.ssa_express_status),
            ("Total SSA EXPRESS Capacity", ssa_express_info.total_ssa_express_capacity),
            ("Free SSA Express Capacity", ssa_express_info.free_ssa_express_capacity),
            ("Used SSA Express Capacity", ssa_express_info.used_ssa_express_capacity),
            ("Entity Count", entity_count.ssa_express_volumes + entity_count.ssa_express_filesystems),
            ))
        return result

    @feature_command(
        description="Set capacity values for SSA Express and SSD Cache. Ensure that the capacity values all match a table row in the ssa_express.capacity_bundles CLI feature_command output",
        types=(Gigabyte, Gigabyte, SSDDrives, YesNo, YesNo),
        feature='ssa_express',
        status=True,
    )
    @roles.INFINIDAT
    def set_capacity(self, _api, total_ssa_express_capacity, total_ssd_drives_capacity, total_ssd_drives, force_ssa_express_unsupported_capacity=False, force_ssa_express_without_free_list=False):
        data = dict(
            total_ssa_express_capacity_gb=total_ssa_express_capacity,
            total_ssd_drives_capacity_gb=total_ssd_drives_capacity,
            total_ssd_drives=total_ssd_drives,
        )
        force_string_field = ''
        if force_ssa_express_unsupported_capacity:
            force_string_field += "force_ssa_express_unsupported_capacity=True&"
        if force_ssa_express_without_free_list:
            force_string_field += "force_ssa_express_without_free_list=True&"
        _api.post('system/ssa_express/capacity?' + force_string_field, data)
        return f"SSA Express capacity was set to {total_ssa_express_capacity} GB"

    QUERY_FIELDS = strlist("total_ssd_drives total_ssd_drives_capacity_gb total_ssd_cache_capacity_gb total_ssa_express_capacity_gb")

    @feature_command(
        description="Before running the ssa_express.set_capacity command, check which bundle sizes are supported",
        options=QueryOptions(fields=QUERY_FIELDS),
        feature='ssa_express',
        status=True,
    )
    @render(
        fields=QUERY_FIELDS,
        entity=SSAExpress,
        )
    @roles.INFINIDAT
    def capacity_bundles(self, _api,):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI.format("supported_capacities")
        return self._query(_api,)
        