from datetime import datetime
from functools import partial
import textwrap
import pytz
from infinishell.main import get_version_mapping
from infinishell.utils import yesno
from munch import Munch
from infinishell.infinibox.commands.options import QueryOptions
from infinishell.infinibox.commands.types import DateDuration

from infinishell.infishell import commandgroup
from infinishell.infishell.commands.commands import command, render
from infinishell.infishell.shell import get_active_shell
from infinishell.infishell.exceptions import BadArguments, CommandAborted, CommandFailed
from infinishell.infishell.shell import get_active_shell
from infinishell.infishell.utils import strlist
from infinishell.infishell.types import (
    YesNo
)
from . import render_collection, roles, EntityCommands


from .config import ConfigCommandsBase
from .types import ConsistencyGroup, FilesystemOrSnapshot, Name, ScheduleDelete, ScheduleDisable, ScheduleEnable, ScheduleTimeDelta, SnapPolicy, SnapPolicySchedules, ScheduleType, DayOfTheWeek, Hour, Minute, VolumeOrSnapshot, SnapshotPolicyAssigntment, Suffix


SNAP_POLICY_QUERY_FIELDS = strlist(
    "name suffix")

SNAP_POLICY_SORTABLE_FIELDS = strlist("name suffix")

SNAP_POLICY_DETAILED_FIELDS = strlist("name suffix")

SNAP_POLICY_SUBSTITUTES_FIELDS = dict(
    name="policy",
)


SNAP_POLICY_SCHEDULES_QUERY_FIELDS = strlist(
    "snapshot_policy name type lock_snapshots interval day_of_week hour minute retention enabled")

SNAP_POLICY_SCHEDULES_SORTABLE_FIELDS = strlist("name lock_snapshots")

SNAP_POLICY_SCHEDULES_SUBSTITUTES_FIELDS = dict(
    name="schedule_name",
    type="schedule_type",
    lock_snapshots="lock_state",
    day_of_week="day",
    enabled="status",
)


SNAP_POLICY_ASSINGMENT_QUERY_FIELDS = strlist(
    "snapshot_policy_name assigned_entity_type assigned_entity_name")

SNAP_POLICY_ASSINGMENT_SUBSTITUTES_FIELDS = dict(
    snapshot_policy_name="snapshot_policy",
    assigned_entity_name="name",
)

@commandgroup(parent=ConfigCommandsBase, prefix="snap_policy")
class SnapPolicyCommands(EntityCommands, ConfigCommandsBase):
    ENTITY_TYPE = SnapPolicy

    @command(
        description="Create a policy that schedules automatic snapshots and appends a suffix to the names of these snapshots.",
        types=(Name, Suffix),
    )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, name, suffix):
        return self._create(_api, name=name, suffix=suffix)

    @command(
        description="Rename a snapshot policy",
        types=(SnapPolicy, Name,),
        hidden_substitue=dict(name="user"),
    )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, policy, new_name):
        return self._rename(_api, policy, new_name)

    @command(
        description="Delete a snapshot policy and all of its schedules.",
        types=(SnapPolicy,),
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, policy):
        return self._delete(_api, policy)

    @command(
        description="List the existing snapshot policies",
        types=(SnapPolicy,),
        options=QueryOptions(
            fields=SNAP_POLICY_QUERY_FIELDS, sortable=SNAP_POLICY_SORTABLE_FIELDS, substitute=SNAP_POLICY_SUBSTITUTES_FIELDS, detailed=True),
    )
    @render(
        fields=SNAP_POLICY_QUERY_FIELDS,
        detailed_fields=SNAP_POLICY_DETAILED_FIELDS,
        substitute=SNAP_POLICY_SUBSTITUTES_FIELDS,
        detailed_individually=True,
    )
    def query(self, _api, policy=None):
        return self._query(_api, policy)

@commandgroup(parent=ConfigCommandsBase, prefix="snap_policy")
class SnapPolicyCommands(EntityCommands, ConfigCommandsBase):
    ENTITY_TYPE = SnapshotPolicyAssigntment

    @command(
        description="Assign a dataset to a snapshot policy.",
        types=(SnapPolicy, VolumeOrSnapshot,
               FilesystemOrSnapshot, ConsistencyGroup),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def assign(self, _api, policy, vol=None, fs=None, cg=None):
        if not vol and not fs and not cg:
            raise BadArguments("vol or fs or cg parameter must be used")
        if sum(x is not None for x in [vol, fs, cg]) > 1:
            raise BadArguments(
                "Only one parameter from vol, fs, cg is allowed")
        entities = vol or fs or cg
        entity_id = vol.id if vol else fs.id if fs else cg.id if cg else None
        entity_type = "Volume" if vol else "Filesystem" if fs else "Cg" if cg else None
        data = dict(assigned_entity_id=entity_id, assigned_entity_type=entity_type)
        _api.post(self.ENTITY_TYPE.URI_TEMPLATE.format(
            policy.id), data)
        if entity_type == "Cg":
            entity_type = "Consistency group"
        return "{} '{}' was assigned to policy '{}'".format(entity_type, entities.name, policy.name)
    
    @command(
        description="Unassign a dataset from a snapshot policy.",
        types=(SnapPolicy, VolumeOrSnapshot,
               FilesystemOrSnapshot, ConsistencyGroup),
    )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def unassign(self, _api, policy, vol=None, fs=None, cg=None):
        if not vol and not fs and not cg:
            raise BadArguments("vol or fs or cg parameter must be used")
        if sum(x is not None for x in [vol, fs, cg]) > 1 and policy:
            raise BadArguments("Only one parameter from vol, fs, cg is allowed when policy selected")
        entity_id = vol.id if vol else fs.id if fs else cg.id if cg else None
        entity_name = vol.name if vol else fs.name if fs else cg.name if cg else None
        entity_type = "Volume" if vol else "Filesystem" if fs else "Cg" if cg else None
        data = dict(assigned_entity_id=entity_id)
        _api.post(self.ENTITY_TYPE.URI_TEMPLATE_DELETE.format(
            policy.id, entity_id), data)
        if entity_type == "Cg":
            entity_type = "Consistency group"
        return ("{} '{}' was unassigned from policy '{}'".format(entity_type, entity_name, policy.name))  
    
    @command(
        description="List the datasets (volumes/filesystems/consistency groups/snapshots) that use this snapshot policy.",
        types=(SnapPolicy,),
        options=QueryOptions(
            fields=SNAP_POLICY_ASSINGMENT_QUERY_FIELDS, substitute=SNAP_POLICY_ASSINGMENT_SUBSTITUTES_FIELDS, detailed=True),
    )
    @render(
        fields=SNAP_POLICY_ASSINGMENT_QUERY_FIELDS,
        detailed_fields=SNAP_POLICY_ASSINGMENT_QUERY_FIELDS,
        substitute=SNAP_POLICY_ASSINGMENT_SUBSTITUTES_FIELDS,
        detailed_individually=True,
        entity=SnapshotPolicyAssigntment,
    )
    def assignment_query(self, _api, policy):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TMPLATE_QUERY.format(policy.id)
        return self._query(_api)


@commandgroup(prefix="schedule", parent=SnapPolicyCommands)
class FilesystemTreeQCommands(EntityCommands):
    ENTITY_TYPE = SnapPolicySchedules


    def _get_day_of_week(self, day, next_day):
        days = ["SUNDAY", "MONDAY", "TUESDAY", "WEDNESDAY", "THURSDAY", "FRIDAY", "SATURDAY"]
        if day == "DAILY":
            return day
        day_index = days.index(day.upper())
        if next_day:
            return days[(day_index + 1) % 7]
        else:
            return days[(day_index - 1) % 7]
    
    def _get_time_of_day(self, ibox_timestamp, client_time_of_day_in_seconds, client_day_of_week, format_to_client):
        DAY_IN_SECONDS = 86400
        day_of_week = None
        
        # Get the local date
        current_time = datetime.now()
        # Get the date on ibox
        ibox_time = datetime.utcfromtimestamp(ibox_timestamp / 1000)

        # Calculate the hour difference
        hour_difference_in_seconds = abs(ibox_time.hour - current_time.hour) * 3600
        
        if format_to_client:
            if current_time > ibox_time:
                time_of_day_in_secodns = client_time_of_day_in_seconds + hour_difference_in_seconds
            else:
                time_of_day_in_secodns = client_time_of_day_in_seconds - hour_difference_in_seconds
        else:    
            if current_time > ibox_time:
                time_of_day_in_secodns = client_time_of_day_in_seconds - hour_difference_in_seconds
            else:
                time_of_day_in_secodns = client_time_of_day_in_seconds + hour_difference_in_seconds
        
        
        if time_of_day_in_secodns < 0:
            time_of_day_in_secodns = time_of_day_in_secodns + DAY_IN_SECONDS
            day_of_week = self._get_day_of_week(client_day_of_week, False)
        elif time_of_day_in_secodns > DAY_IN_SECONDS:
            time_of_day_in_secodns = time_of_day_in_secodns - DAY_IN_SECONDS
            day_of_week = self._get_day_of_week(client_day_of_week, True)
        return int(time_of_day_in_secodns), day_of_week if day_of_week else client_day_of_week

    @command(
        description="Add a schedule to a snapshot policy",
        types=(SnapPolicy, Name, ScheduleType, DateDuration, ScheduleTimeDelta,
               DayOfTheWeek, Hour, Minute, YesNo, YesNo),
    )
    @roles.INFINIDAT.ADMIN
    def add(self, _api, policy, name, type, retention, interval=None, day=None, hour=None, minute=None, enabled=None, lock=None):
        version_map = get_version_mapping()
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(policy.id)
        system_utc = _api.get("system/localtime/utc_time")
        formatted_time_of_day = None

        if version_map == 'infinibox' and type == "CLOCK":
            config = _api.get('config')
            timezone = pytz.timezone(config.mgmt['environment.timezone'])
            system_utc = _api.get("system/localtime/utc_time")
            utcnow = datetime.utcnow()
            tz_system_datetime = utcnow.replace(tzinfo=pytz.utc).astimezone(timezone).strftime("%a %b %d %Y %H:%M:%S")
            prompt_parts = (
                textwrap.fill(
                    "Warning: The time parameters in this command use your system's internal time zone which is {}.").format(timezone),
                textwrap.fill(
                "Your system's current time is {}.").format(tz_system_datetime),
                textwrap.fill(
                "Do you wish to continue?.")
            )
            if not yesno("\n\n".join(prompt_parts), always_prompt=True):
                    raise CommandAborted("Command aborted by user")
        if lock is True:
            get_active_shell().warning("Locked snapshots cannot be modified or deleted during their retention.")
        if type == "PERIODIC" and interval is None:
            raise BadArguments(
                "An Interval value is required for schedules that are Periodic type.")
        if type == "CLOCK" and day is None:
            raise BadArguments(
                "A Day value is required for schedules that are Clock type.")
        if type == "CLOCK" and hour is None:
            raise BadArguments(
                "An Hour value is required for schedules that are Clock type.")
        if type == "CLOCK" and minute is None:
            raise BadArguments(
                "A Minute value is required for schedules that are Clock type.")
        time_of_day = None
        if hour is not None and minute is not None:
            time_of_day = int((hour + (minute / 60)) * 3600)
            formatted_time_of_day, day = self._get_time_of_day(system_utc, time_of_day, day, False)
            if day == "DAILY":
                day = "ALL"
        data = Munch(
            enabled=enabled,
            type=type,
            name=name,
            interval=interval,
            day_of_week=day,
            time_of_day=formatted_time_of_day,
            retention=retention['seconds'],
            lock_snapshots=lock,
        )
        data = {key: value for key, value in data.items() if value is not None}
        self._create(_api, **data)
        return "A {} type schedule named '{}' has been added to policy '{}'.".format(type, name, policy.name)

    @command(
        description="Delete a schedule from snapshot policy",
        types=(SnapPolicy, ScheduleDelete,),
    )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, policy, schedule):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(policy.id)
        schedule_entity = self.ENTITY_TYPE.get_by_name(_api, schedule)
        return self._delete(_api, schedule_entity)
    
    @command(
        description="Disable an existing snapshot policy schedule.",
        types=(SnapPolicy, ScheduleDisable,),
    )
    @roles.INFINIDAT.ADMIN
    def disable(self, _api, policy, schedule):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(policy.id)
        schedule_entity = self.ENTITY_TYPE.get_by_name(_api, schedule)
        if schedule_entity.enabled is False:
            raise CommandFailed("Schedule '{}' is already disabled.".format(schedule_entity.name))
        _api.post(self.ENTITY_TYPE.URI_TEMPLATE_DISABLE.format(policy.id, schedule_entity.id))
        return "Schedule '{}' was disabled.".format(schedule_entity.name)
        
    @command(
        description="Enable an existing snapshot policy schedule.",
        types=(SnapPolicy, ScheduleEnable,),
    )
    @roles.INFINIDAT.ADMIN
    def enable(self, _api, policy, schedule):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(policy.id)
        schedule_entity = self.ENTITY_TYPE.get_by_name(_api, schedule)
        if schedule_entity.enabled:
            raise CommandFailed("Schedule '{}' is already enabled.".format(schedule_entity.name))
        _api.post(self.ENTITY_TYPE.URI_TEMPLATE_ENABLE.format(policy.id, schedule_entity.id))
        return "Schedule '{}' was enabled.".format(schedule_entity.name)

    @command(
        description="List the existing snapshot policies schedules",
        types=(SnapPolicy, YesNo),
        options=QueryOptions(
            fields=SNAP_POLICY_SCHEDULES_QUERY_FIELDS, sortable=SNAP_POLICY_SCHEDULES_SORTABLE_FIELDS, substitute=SNAP_POLICY_SCHEDULES_SUBSTITUTES_FIELDS, detailed=True),
    )
    @render(
        fields=SNAP_POLICY_SCHEDULES_QUERY_FIELDS,
        detailed_fields=SNAP_POLICY_SCHEDULES_QUERY_FIELDS,
        substitute=SNAP_POLICY_SCHEDULES_SUBSTITUTES_FIELDS,
        detailed_individually=True,
    )
    def query(self, _api, policy, lock=None):
        self.ENTITY_TYPE.URI = self.ENTITY_TYPE.URI_TEMPLATE.format(policy.id)
        system_utc = _api.get("system/localtime/utc_time")
        return self._query(_api, lock_snapshots=lock, _hook=partial(self._query_hook, policy=policy, system_utc=system_utc))

    def _query_hook(self, entities, policy=None, system_utc=None):
        results = []
        for e in entities:
            e.snapshot_policy = policy.name
            e.lock_snapshots = 'Locked' if e.lock_snapshots else 'Unlocked'
            e.enabled = "ENABLED" if e.enabled else "DISABLED"
            if e.time_of_day and e.day_of_week:
                e.time_of_day, day_of_week = self._get_time_of_day(system_utc, e.time_of_day, e.day_of_week, True)
                e.day_of_week = day_of_week
            if e.time_of_day == 0:
                e.hour = 0
                e.minute = 0
            elif e.time_of_day:
                e.hour = e.time_of_day // 3600
                e.minute = (e.time_of_day % 3600)//60
            else:
                e.hour = None
                e.minute = None
            results.append(e)
        return results